/*------------------------------------------------------------------------*/
/*                                                                        */
/*                         AMULET3 Generalised Memory                     */
/*                                                                        */
/*                                             Created: JDG 11/12/96      */
/*                                       Last Modified: JDG  14/7/97      */
/*                                                                        */
/*HEALTH WARNING >>>SWI FILE HANDLING ADDED<<<   Modified: JVW  22/10/97  */
/*------------------------------------------------------------------------*/
/*                                                                        */
/*   This version comprises a non-latching memory stage followed by a     */
/*     pipeline latch.                                                    */
/*                                                                        */
/*------------------------------------------------------------------------*/
/* Timing parameters                                                      */

MEM_delay : var(int).

/*------------------------------------------------------------------------*/

dual_port     : var(bool).              /* Flag set if memory dual ported */
mem_addr_mask : var(int).      /* Mask indicating number of memory blocks */
/*    0 => 1 block,  1 => 2 blocks,  3 => 4 blocks, ...,  15 => 16 blocks */

/*------------------------------------------------------------------------*/

freadfile(filename     : val(string),
          memory_array : var(int^40000)):
          expr(void) =
  (
  handle : var(bfile).
  number : var(int).

  i       : var(int).
  B_count : var(int).
  inch    : var(char).

  handle := bopenin(filename);
        i := 0;
   number := 0;
  B_count := 0;

  while (!eof(handle)) do
    (
    inch := fgetc(handle);
    number := number or ((cast(int, inch)) <<(8*(B_count)));
    B_count := (B_count + 1) and 16#00000003;
    memory_array[i] := number;
    if B_count = 0 then
    (
     i := i + 1;
     number := 0
    )
    );

  printstr("Memory loaded\n");
i:=0;
while i < 10 do
(
 printhex(memory_array[i]);
 printstr("\n");
 i := i + 1
)
  ).

/*------------------------------------------------------------------------*/

Memory(Iaddr : var(chan(Addr_bus_type)),
       op_code : var(chan(Data_bus_type)),
       Daddr   : var(chan(Addr_bus_type)),
       Dout    : var(chan(Data_bus_type))):
       expr(void) =
  (
  memory_array : var(int^40000).
  i            : var(int).
  block_lock   : var(bool^16).

  Memory_cycle_count : var(int).                /* Analysis purposes only */
  Last_cycle_count   : var(int).                /* Analysis purposes only */
  Last_time          : var(int).                /* Analysis purposes only */

  Mem_status: expr(void) =                           /* Analysis printout */
    (
    printstr("{{Time is "); printint(now); printstr(" (elapsed ");
    printint(now - Last_time); printstr(")}}\n");
    printstr("{{Memory cycle count is "); printint(Memory_cycle_count);
    printstr(" (elapsed "); printint(Memory_cycle_count - Last_cycle_count);
    printstr(")}}\n");
    Last_time := now;
    Last_cycle_count := Memory_cycle_count
    ).


  Mem_port(Addr : var(chan(Addr_bus_type)),
           Data : var(chan(Data_bus_type)),
           Name : val(string)):
           expr(void) =
    (
    bit_lines : var(chan(Data_bus_type)).
                                    /* Input to sense amps. (latch stage) */
    Qcol       : var(int).
    Aborted    : var(bool).
    block_addr : var(int).

    data_value : var(Data_bus_type).
    mem_addr   : var(int).
    byte_sel   : var(int).

/*----------------------------------------------------------------------*/
/* Following for SWI handling */
    L_handles    : var(bfile^100).
    nxt_hand     : var(int).
    file_name    : var(int^16).
    Rbase        : var(int).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

do_SWI(SWI_type:val(int)) : expr(void) =
  (
    Byte_sel     : var(int).  
    Word_sel     : var(int).  
    B_count      : var(int).  

  (case SWI_type                                    /* Pick up SWI number */

    when 16#61 => /* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>  Read clock to R0 */
      memory_array[Rbase+0] := now         /* LARD time returned in  Reg0 */

    when 16#63 =>  /* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Read time to R0 */
      memory_array[Rbase+0] := now         /* LARD time returned in  Reg0 */

    when 16#66 =>  /* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Open File */
    (
    inch         : var(int).  
    j            : var(int).  
    first_ch     : var(bool).
    fnam_str     : var(string).

    newstringvar(fnam_str);


 /* Reg0 has file name pointer */
    Byte_sel := memory_array[Rbase+0] and 16#00000003;
    Word_sel := memory_array[Rbase+0] >> 2;
    j        := 0;
    B_count  := 0;
    first_ch := true;
    repeat
      (
      inch := (memory_array[Word_sel] >> (Byte_sel * 8)) and 16#000000FF;
      if first_ch then
         fnam_str := c2s(cast(char,inch))
      else
         fnam_str := fnam_str@c2s(cast(char,inch));

      if (B_count = 0) then                         /* First byte in word */
        file_name[j] := inch << 24
      else
        file_name[j] := file_name[j] or (inch << (8 * (3 - B_count)));
      Byte_sel := (Byte_sel + 1) and 16#00000003;
      if (Byte_sel = 0) then Word_sel := Word_sel + 1;
      B_count  := (B_count + 1)  and 16#00000003;
      if (B_count = 0) then j := j + 1
      ) until (inch = 0);

    (case memory_array[Rbase+1] /* Reg1 has file access info */

      when 16#00000000 =>                        /* "r" open for reading  */
        L_handles[nxt_hand] := bopenin(fnam_str)
       
      when 16#00000001 =>               /* "rb" open for reading (binary) */
        L_handles[nxt_hand] := bopenin(fnam_str)

      when 16#00000002 =>             /* "r+" open for update (read+write)*/
        L_handles[nxt_hand] := bopeninout(fnam_str)

      when 16#00000003 =>            /* "r+b" open for update (read+write)*/
        L_handles[nxt_hand] := bopeninout(fnam_str)

      when 16#00000004 =>                       /* "w" create for writing */
       (
        L_handles[nxt_hand] := bopenout(fnam_str);
        ftrunc(L_handles[nxt_hand])
	)
      when 16#00000005 =>                       /* "wb" create for update */
       (
        L_handles[nxt_hand] := bopenout(fnam_str);
        ftrunc(L_handles[nxt_hand])
	)
      when 16#00000006 =>                       /* "w+" create for update */
        (
        L_handles[nxt_hand] := bopeninout(fnam_str);
        ftrunc(L_handles[nxt_hand])
	 )
      when 16#00000007 =>                     /* "w+b"  create for update */
        (
        L_handles[nxt_hand] := bopeninout(fnam_str);
        ftrunc(L_handles[nxt_hand])
	 )

      when 16#00000008  =>          /* "a" append open for writing at eof */
        (
        L_handles[nxt_hand] := bopenout(fnam_str);
        fseek(L_handles[nxt_hand],flen(L_handles[nxt_hand]))
	 )

      when 16#00000009  =>         /* "ab" append open for writing at eof */
        (
        L_handles[nxt_hand] := bopenout(fnam_str);
        fseek(L_handles[nxt_hand],flen(L_handles[nxt_hand]))
	 )

      when 16#0000000A  =>        /* "a+"  append open for writing at eof */
        (
        L_handles[nxt_hand] := bopeninout(fnam_str);
        fseek(L_handles[nxt_hand],flen(L_handles[nxt_hand]))
	 )

      when 16#0000000B  =>        /* "a+b" append open for writing at eof */
        (
        L_handles[nxt_hand] := bopeninout(fnam_str);
        fseek(L_handles[nxt_hand],flen(L_handles[nxt_hand]))
	 )

       default
         (
         printstr("\n");
         printint(memory_array[Rbase+1]);
         printstr(" : Unrecognised file access\n")
         )
    );                                                     /* End of case */

    memory_array[Rbase+0] := nxt_hand; /* L_handles index returned in  Reg0 */
    nxt_hand := nxt_hand +1;
    endstringvar(fnam_str)
    )      /* end of when 16#66 */


    when 16#68 => /* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> Close file with handle R0 */
      close(L_handles[memory_array[Rbase+0]])


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

    when 16#69 => /* Write R2 bytes from [R1] to file R0, R0 returns status */
      (
      outch        : var(int).  

      B_count  := memory_array[Rbase + 2];
      Byte_sel := memory_array[Rbase + 1] and 16#03;
      Word_sel := memory_array[Rbase + 1] >> 2;
      while B_count > 0 do
        (
        outch := ((memory_array[Word_sel] >> (Byte_sel * 8)) and 16#FF);
        fputc(L_handles[memory_array[Rbase+0]], cast(char, outch));
        Byte_sel := ((Byte_sel + 1) and 16#00000003);
        if (Byte_sel = 0) then Word_sel := Word_sel + 1;
        B_count := B_count - 1
        );
      memory_array[Rbase+0] :=  B_count          /* Return residual count */
                          /* what about return of STATUS ??  99999999999  */
      )

    when 16#6A => /* Read R2 bytes from file R0 to  [R1] R0 returns # unread  */
      (
      inch         : var(int).  

      B_count := memory_array[Rbase + 2];
      Byte_sel := memory_array[Rbase + 1] and 16#03;
      Word_sel := memory_array[Rbase + 1] >> 2;
      while ((B_count > 0) && !eof(L_handles[memory_array[Rbase+0]])) do
        (
        inch := cast(int, fgetc(L_handles[memory_array[Rbase+0]]));
        inch := (inch << (Byte_sel * 8));
        memory_array[Word_sel] := memory_array[Word_sel] or inch;
        Byte_sel := ((Byte_sel + 1) and 16#03);
        if (Byte_sel = 0) then Word_sel := Word_sel + 1;
        B_count := B_count - 1
        );
      memory_array[Rbase+0] :=  B_count          /* Return residual count */
    /* early EOF tested 9999 What else ????  */
      )

    when 16#70 => /* Install handler - Sometime 9999  */
      (
      printstr(" : SWI call to Install Handler\n")
      )

    default printstr("Bollards!\n")

  )
  ).

/*>>>>>>>>>>>>>>>>>>>>>>> END of SWI handling <<<<<<<<<<<<<<<<<<<<<<<<<<*/



/*----------------------------------------------------------------------*/

    Aborted := false;
    Rbase := 16#10;     /* Base of dumped reg set */
    nxt_hand := 0;      /* L_handles index */
    init_chan(bit_lines, Name@"bitlines");

    Latch(bit_lines, Data, 0, false, Name@"latch") |

    forever
      (
      Addr?                                           /* Wait for request */
        (
        Memory_cycle_count := Memory_cycle_count + 1;

        mem_addr := (?Addr)->addr>>2;
        byte_sel := ((?Addr)->addr) and 3;

        block_addr := (mem_addr >> 2) and mem_addr_mask;  /* 4 words/line */
        if (!dual_port) then
          (
          wait_until(!block_lock[block_addr]);  /* Wait if block conflict */
          block_lock[block_addr] := true          /* Signal block entered */
          );
/*
printstr("Memory invoked, address "); printhex((?Addr)->addr);
printstr(" at time "); printint(now); printstr("\n");
*/
        wait_for(MEM_delay);                              /* Memory delay */

        data_value->extra := (?Addr)->extra;     /* Auxilliary data field */

        if (Aborted && (Qcol = (?Addr)->Qcol)) then
          data_value->abort := true          /* true _following_ an abort */
                             /* Prevent further action until colour match */
        else
         (
          Qcol    := (?Addr)->Qcol;               /* Adopt current colour */

          data_value->data  := 0;                    /* Just some default */

          if ((?Addr)->addr = Tube_addr) then
            (                                                  /* for SWI */
            data_value->abort := false;
            if (((?Addr)->Dout and 16#7F) = 4) then
              (Mem_status; printstr("Stopping\n"); Run := false)
            else
              (
              if (((?Addr)->Dout and 16#7F) = 5) then Mem_status
              else ASCII_out((?Addr)->Dout)
              )
            )
/*Here's the SWI call*/
        else    /* Not Tube */
          (
          if (((?Addr)->addr = SWI_addr) && ((?Addr)->Write)) then
             (
              data_value->abort := false;
              do_SWI((?Addr)->Dout)
             )
            
/*End of SWI call insert*/
          else    /* Not SWI */
            (
            if ((mem_addr < 0) || (mem_addr >= 40000)) then/* Memory management*/
              data_value->abort := true
            else
              (
              data_value->abort := false;
              if ((?Addr)->Read) then data_value->data := memory_array[mem_addr];
              if ((?Addr)->Write) then
                (
                if ((?Addr)->Size = 4) then                 /* Word write */
                  memory_array[mem_addr] := (?Addr)->Dout
                else  /*not word size*/
                    (
                     mask : var(int).
                     (case (?Addr)->Size
                       when 1 =>
                         (case byte_sel
                           when 0 => mask := 16#000000FF
                           when 1 => mask := 16#0000FF00
                           when 2 => mask := 16#00FF0000
                           when 3 => mask := 16#FF000000
                         )
                       when 2 =>
                         (case (byte_sel and 2)
                           when 0 => mask := 16#0000FFFF
                           when 2 => mask := 16#FFFF0000
                         )
                       default              /* NB size 4 already trapped */
                         (
                         printint((?Addr)->Size);
                         printstr(" is not a legal transfer size\n")
                         )
                     );                       /* End of "case(?Addr)" */
                  memory_array[mem_addr] := (memory_array[mem_addr] and not(mask))
                                         or ((?Addr)->Dout and mask)
                    )   /* else not word size */
                  )     /* if (?Addr)->Write */
              )         /* Not Memory Management */
            )           /* Not a SWI call*/
          );            /* Not a Tube address*/
          Aborted := data_value->abort
         );            /* if Aborted && .... else */

        bit_lines!data_value;

        block_lock[block_addr] := false            /* Signal block exited */
        )                                    /*  (?Addr)..Acknowledge input */
      )                                            /* forever */
    ).


  freadfile("programme", memory_array);                 /* Initialisation */
  (i : var(int). forseq i in 0 to 15 do block_lock[i] := false); /* Init. */
  Memory_cycle_count := 0;
  Last_time          := 0;
  Last_cycle_count   := 0;

  (Mem_port(Iaddr, op_code, "Imem_") | Mem_port(Daddr, Dout, "Dmem_"))
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
