/*------------------------------------------------------------------------*/
/*                                                                        */
/*                         AMULET3 Generalised Memory                     */
/*                                                                        */
/*                                             Created: JDG 11/12/96      */
/*                                       Last Modified: JDG  14/7/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/*   This version comprises a non-latching memory stage followed by a     */
/*     pipeline latch.                                                    */
/*                                                                        */
/*------------------------------------------------------------------------*/
/* Timing parameters                                                      */

MEM_delay : var(int).

/*------------------------------------------------------------------------*/

dual_port     : var(bool).              /* Flag set if memory dual ported */
mem_addr_mask : var(int).      /* Mask indicating number of memory blocks */
/*    0 => 1 block,  1 => 2 blocks,  3 => 4 blocks, ...,  15 => 16 blocks */

/*------------------------------------------------------------------------*/

freadfile(filename     : val(string),
          memory_array : var(int^40000)):
          expr(void) =
  (
  handle : var(tfile).
  number : var(int).

  i : var(int).

  handle := topenin(filename);
  i := 0;
  while (!eof(handle)) do
    (
    number := freadint(handle);
    memory_array[i] := number;
    i := i + 1
    );

  printstr("Memory loaded\n")
  ).

/*------------------------------------------------------------------------*/

Memory(Iaddr : var(chan(Addr_bus_type)),
       op_code : var(chan(Data_bus_type)),
       Daddr   : var(chan(Addr_bus_type)),
       Dout    : var(chan(Data_bus_type))):
       expr(void) =
  (
  memory_array : var(int^40000).
  i            : var(int).
  block_lock   : var(bool^16).

  Memory_cycle_count : var(int).                /* Analysis purposes only */
  Last_cycle_count   : var(int).                /* Analysis purposes only */
  Last_time          : var(int).                /* Analysis purposes only */

  Mem_status: expr(void) =                           /* Analysis printout */
    (
    printstr("{{Time is "); printint(now); printstr(" (elapsed ");
    printint(now - Last_time); printstr(")}}\n");
    printstr("{{Memory cycle count is "); printint(Memory_cycle_count);
    printstr(" (elapsed "); printint(Memory_cycle_count - Last_cycle_count);
    printstr(")}}\n");
    Last_time := now;
    Last_cycle_count := Memory_cycle_count
    ).


  Mem_port(Addr : var(chan(Addr_bus_type)),
           Data : var(chan(Data_bus_type)),
           Name : val(string)):
           expr(void) =
    (
    bit_lines : var(chan(Data_bus_type)).
                                    /* Input to sense amps. (latch stage) */
    Qcol       : var(int).
    Aborted    : var(bool).
    block_addr : var(int).

    data_value : var(Data_bus_type).
    mem_addr   : var(int).
    byte_sel   : var(int).

    Aborted := false;

    init_chan(bit_lines, Name@"bitlines");

    Latch(bit_lines, Data, 0, false, Name@"latch") |

    forever
      (
      Addr?                                           /* Wait for request */
        (
        Memory_cycle_count := Memory_cycle_count + 1;

        mem_addr := (?Addr)->addr>>2;
        byte_sel := ((?Addr)->addr) and 3;

        block_addr := (mem_addr >> 2) and mem_addr_mask;  /* 4 words/line */
        if (!dual_port) then
          (
          wait_until(!block_lock[block_addr]);  /* Wait if block conflict */
          block_lock[block_addr] := true          /* Signal block entered */
          );
/*
printstr("Memory invoked, address "); printhex((?Addr)->addr);
printstr(" at time "); printint(now); printstr("\n");
*/
        wait_for(MEM_delay);                              /* Memory delay */

        data_value->extra := (?Addr)->extra;     /* Auxilliary data field */

        if (Aborted && (Qcol = (?Addr)->Qcol)) then
          data_value->abort := true          /* true _following_ an abort */
                             /* Prevent further action until colour match */
        else
          (
          Qcol    := (?Addr)->Qcol;               /* Adopt current colour */

          data_value->data  := 0;                    /* Just some default */

          if ((?Addr)->addr = Tube_addr) then
            (                                                  /* for SWI */
            data_value->abort := false;
            if (((?Addr)->Dout and 16#7F) = 4) then
              (Mem_status; printstr("Stopping\n"); Run := false)
            else
              (
              if (((?Addr)->Dout and 16#7F) = 5) then Mem_status
              else ASCII_out((?Addr)->Dout)
              )
            )
          else
            (
            if ((mem_addr < 0) || (mem_addr >= 40000)) then/* Memory management*/
              data_value->abort := true
            else
              (
              data_value->abort := false;
              if ((?Addr)->Read) then data_value->data := memory_array[mem_addr];
              if ((?Addr)->Write) then
                (
                if ((?Addr)->Size = 4) then                 /* Word write */
                  memory_array[mem_addr] := (?Addr)->Dout
                else
                  (
                  mask : var(int).
                  (case (?Addr)->Size
                    when 1 =>
                      (case byte_sel
                        when 0 => mask := 16#000000FF
                        when 1 => mask := 16#0000FF00
                        when 2 => mask := 16#00FF0000
                        when 3 => mask := 16#FF000000
                      )
                    when 2 =>
                      (case (byte_sel and 2)
                        when 0 => mask := 16#0000FFFF
                        when 2 => mask := 16#FFFF0000
                      )
                    default                  /* NB size 4 already trapped */
                      (
                      printint((?Addr)->Size);
                      printstr(" is not a legal transfer size\n")
                      )
                  );                                     /* End of "case" */
                  memory_array[mem_addr] := (memory_array[mem_addr] and not(mask))
                                         or ((?Addr)->Dout and mask)
                  )
                )
              )
            );
          Aborted := data_value->abort
          );

        bit_lines!data_value;

        block_lock[block_addr] := false            /* Signal block exited */
        )                                            /* Acknowledge input */
      )
    ).


  freadfile("programme", memory_array);                 /* Initialisation */
  (i : var(int). forseq i in 0 to 15 do block_lock[i] := false); /* Init. */
  Memory_cycle_count := 0;
  Last_time          := 0;
  Last_cycle_count   := 0;

  (Mem_port(Iaddr, op_code, "Imem_") | Mem_port(Daddr, Dout, "Dmem_"))
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
