/*------------------------------------------------------------------------*/
/*                                                                        */
/*                           AMULET3 Thumb decoder                        */
/*                                                                        */
/*                                             Created: JVW  22/1/97      */
/*                                            Modified: JDG   6/3/97      */
/*                                            Modified: JVW  19/3/97      */
/*                                            Modified: JDG  28/4/97      */
/* (MOVN/CMP S-bit set)                                                   */
/*------------------------------------------------------------------------*/
/* By the pricking of our Thumbs !!                                       */

Thumb_Ex(ThumbI  : val(int),
         Th_LSA  : var(int),
         Th_special : var(bool),
         ArmI    : var(int),
         PC     : var(int)):
         expr(void)=
  (

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* The following is a Field Move procedure.  */

  FM(Mask       : val(int),
     Shift_L    : val(bool),
     Shift_dist : val(int)):
     expr(void)=
    (
    Al_reg : var(int).
    Al_reg := ThumbI and Mask;
    if Shift_L then Al_reg := Al_reg << Shift_dist
    else            Al_reg := Al_reg >> Shift_dist;
    ArmI := ArmI or Al_reg
    ).                                    /* End of Field Move definition */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* The following is a Field Move procedure.  */
/* But this one inverts the selected field   */

  FMI(Mask       : val(int),
      Shift_L    : val(bool),
      Shift_dist : val(int)):
      expr(void)=
    (
    Al_reg : var(int).

    Al_reg := ThumbI and Mask;
    Al_reg := Al_reg xor Mask;
    if Shift_L then Al_reg := Al_reg << Shift_dist
    else            Al_reg := Al_reg >> Shift_dist;
    ArmI := ArmI or Al_reg
    ).                          /* End of Inverting Field Move definition */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* The following is a Field Insert  procedure.  */
/* This one inserts supplied (Mask) field   */

  FI(In_field   : val(int),
     Shift_L    : val(bool),
     Shift_dist : val(int)):
     expr(void)=
    (
    Al_reg : var(int).
    Al_reg := In_field;
    if Shift_L then Al_reg := Al_reg << Shift_dist
    else            Al_reg := Al_reg >> Shift_dist;
    ArmI := ArmI or Al_reg
    ).                                  /* End of Field Insert definition */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* The following expands Thumb encoding to Arm format                     */
/* Assumption here is that Thumb instruction is in ls 16 bits of input field */

F1orF2Exp: expr(void)=
  (
  if ((ThumbI and 16#00001800) != 16#00001800) then/* F1 Move shifted reg.*/
    (
    ArmI := 16#E1B00000;
    FM(16#00001800, false,  10#6);                             /* Op field */
    FM(16#000007C0,  true,  10#1);                              /* Offset5 */
    FM(16#00000038, false,  10#3);                                   /* Rs */
    FM(16#00000007,  true, 10#12)                                    /* Rd */
    )                                         /* End of Format 1 expansion */
  else
    (                                                        /* F2 Add/Sub */
    ArmI := 16#E0100000;
    FM(16#00000400,  true, 10#15);                              /* Imm Bit */
    FM(16#00000200,  true, 10#13);                        /* Op field <22> */
    FMI(16#00000200,  true, 10#14);                       /* Op field <23> */
    FM(16#000001C0, false,  10#6);                           /* Offset3/Rn */
    FM(16#00000038,  true, 10#13);                                   /* Rs */
    FM(16#00000007,  true, 10#12)                                    /* Rd */
    )                                         /* End of Format 2 expansion */
  ).                                                   /* End of F1orF2Exp */

/*=========================================================================*/
                /* F3 Mov/comp/add/sub imm. */

F3Exp: expr(void)=
  (
  ArmI := 16#E2100000;
  (case (ThumbI and 16#00001800)
    when 16#00000000 => FI(16#00D00000, true, 10#1)        /* F3(00) - MOV */
    when 16#00000800 => FI(16#00A00000, true, 10#1)        /* F3(01) - CMP */
    when 16#00001000 => FI(16#00400000, true, 10#1)        /* F3(10) - ADD */
    when 16#00001800 => FI(16#00200000, true, 10#1)        /* F3(11) - SUB */
  );

  FM(16#00000700,  true,  10#8);                         /* Rd - operand 1 */
  FM(16#00000700,  true,  10#4);                       /* Rd - destination */
  FM(16#000000FF,  true,  10#0)                                 /* Offset8 */
  ).                                                       /* End of F3Exp */

/*=========================================================================*/

/* >>>>>>>>>>>>>>>>>>> F4 EXPANSION <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
F4Exp: expr(void)=
  (

  F4Exp_arithmetic: expr(void) =
    (
    FM(16#000003C0,  true, 10#15);           /* Move op. code (okay here) */
    FM(16#00000038, false,  10#3);                            /* Rs field */
    FM(16#00000007,  true, 10#12);                    /* Rd field as dest */
    FM(16#00000007,  true, 10#16)                     /* Rd field as Op-1 */
    ).

  F4Exp_shift: expr(void) =
    (
    FI(16#01A00010,  true,  0);             /* F4 LSL -> MOV & Imm. shift */
    FM(16#00000100, false,  2);                             /* Shift code */
    FM(16#00000040, false,  1);
    FM(16#00000038,  true,  10#5);                           /* Rs field  */
    FM(16#00000007,  true,  10#0);                        /* Rd as source */
    FM(16#00000007,  true, 10#12)                    /* Rd as destination */
    ).

  F4Exp_negate: expr(void) =
    (
    FI(16#02000000,  true,  10#0);          /* F4 NEG -> RSBS Rd, Rs, #0  */
    FI(16#00300000,  true,  10#1);                          /* RSB OpCode */
    FM(16#00000038,  true, 10#13);                     /* Rn field source */
    FM(16#00000007,  true, 10#12)                      /* F4 NEG Rd field */
    ).

  F4Exp_multiply: expr(void) =
    (
    FI(16#00100000,  true,  10#0);                       /* Op.=0 A=0 S=1 */
    FI(16#00000090,  true,  10#0);            /* Bits 7->4 = 1001 for MUL */
    FM(16#00000038, false,  10#3);                /* F4 MUL Rs as Source. */
    FM(16#00000007,  true, 10#16);                  /* F4 MUL Rd as Dest. */
    FM(16#00000007,  true,  10#8)                  /* F4 MUL Rd as Source */
    ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  ArmI := 16#E0100000;
  (case (ThumbI and 16#000003C0)
    when 16#00000000 => F4Exp_arithmetic                           /* AND */
    when 16#00000040 => F4Exp_arithmetic                           /* EOR */
    when 16#00000080 => F4Exp_shift                                /* LSL */
    when 16#000000C0 => F4Exp_shift                                /* LSR */
    when 16#00000100 => F4Exp_shift                                /* ASR */
    when 16#00000140 => F4Exp_arithmetic                           /* ADC */
    when 16#00000180 => F4Exp_arithmetic                           /* SBC */
    when 16#000001C0 => F4Exp_shift                                /* ROR */
    when 16#00000200 => F4Exp_arithmetic                           /* TST */
    when 16#00000240 => F4Exp_negate                               /* NEG */
    when 16#00000280 => F4Exp_arithmetic                           /* CMP */
    when 16#000002C0 => F4Exp_arithmetic                           /* CMN */
    when 16#00000300 => F4Exp_arithmetic                           /* ORR */
    when 16#00000340 => F4Exp_multiply                             /* MUL */
    when 16#00000380 => F4Exp_arithmetic                           /* BIC */
    when 16#000003C0 => F4Exp_arithmetic                           /* MVN */
  )
  ).                                                       /* End of F4Exp*/

/*========================================================================*/
/*      F5 Expansion    */

F5Exp: expr(void)=
  (
  if ((ThumbI and 16#00000300) != 16#00000300) then  /* NOT the "BX" case */
    (
    ArmI := 16#E0000000;
    FM(16#00000080,  true, 10#12);                              /* H1 bit */
    FM(16#00000080,  true,  10#8);                              /* H1 bit */
    FM(16#00000040, false,  10#3);                              /* H2 bit */
    FM(16#00000038, false,  10#3);                            /* Rs field */
    FM(16#00000007,  true, 10#12);                    /* Rd field - Dest. */
    FM(16#00000007,  true, 10#16);                /* Rd field - First Op. */

    (case (ThumbI and 16#00000300)
      when 16#00000000 => FI(16#00400000, true, 10#1)/* Op Code 0100 - ADD */
      when 16#00000100 => FI(16#00A80000, true, 10#1)/* Op Code 1010 - CMP - Set CCs */
      when 16#00000200 => FI(16#00D00000, true, 10#1)/* Op Code 1101 - MOV */
      default printstr("Failed Format 5 Decode")
    )
    )                                               /* End of NON-BX cases */
  else
    (                                              /*  This is the BX case */
    ArmI := 16#E12FFF10;
    FM(16#00000078, false,  10#3)                    /*  H2 bit & Rs field */
    )                                                    /* End of BX case */
  ).                                                       /* End of F5Exp */

/*=========================================================================*/
/*      F6 Expansion    */

F6Exp: expr(void)=
  (
  ArmI := 16#E59F0000;
  FM(16#00000700,  true,  10#4);                               /* Rd field */
  FM(16#000000FF,  true,  10#2)                             /* Word8 field */
  ).                                                /* End of F6 expansion */

/*=========================================================================*/
/*      F7 Expansion    */

F7Exp: expr(void)=
  (
  ArmI := 16#E7800000;
  FM(16#00000800,  true,  10#9);                                  /* L bit */
  FM(16#00000400,  true, 10#12);                                  /* B bit */
  FM(16#000001C0, false,  10#6);                        /* Ro - Offset reg */
  FM(16#00000038,  true, 10#13);                               /* Rb field */
  FM(16#00000007,  true, 10#12)                                /* Rd field */
  ).                                                /* End of F7 expansion */

/*=========================================================================*/
/*      F8 Expansion    */

F8Exp: expr(void)=
  (
  ArmI := 16#E1800090;
  (case (ThumbI and 16#00000C00)
    when 16#00000000 => FI(16#00000020, true, 0)            /* L=0 S=0 H=1 */
    when 16#00000400 => FI(16#00100040, true, 0)            /* L=1 S=1 H=0 */
    when 16#00000800 => FI(16#00100020, true, 0)            /* L=1 S=0 H=1 */
    when 16#00000C00 => FI(16#00100060, true, 0)            /* L=1 S=1 H=1 */
  );
  FM(16#000001C0, false,  10#6);                        /* Ro - Offset reg */
  FM(16#00000038,  true, 10#13);                               /* Rb field */
  FM(16#00000007,  true, 10#12)                                /* Rd field */
  ).                                                /* End of F8 expansion */

/*=========================================================================*/
/*      F9 Expansion    */

F9Exp: expr(void)=
  (
  ArmI := 16#E5800000;
  FM(16#00001000,  true, 10#10);                                  /* B bit */
  FM(16#00000800,  true,  10#9);                                  /* L bit */
  if ((ThumbI and 16#00001000) = 0) then                       /* I.E. B=0 */
    FM(16#000007C0, false,  10#4)                               /* Offset5 */
  else                                                         /* I.E. B=1 */
    FM(16#000007C0, false,  10#6);                              /* Offset5 */
  FM(16#00000038,  true, 10#13);                               /* Rn field */
  FM(16#00000007,  true, 10#12)                                /* Rd field */
  ).                                                /* End of F9 expansion */

/*=========================================================================*/
/*      F10 Expansion    */

F10Exp: expr(void)=
  (
  ArmI := 16#E1C000B0;
  FM(16#00000800,  true,  10#9);                                  /* L bit */
  FM(16#00000600, false,  10#1);                      /* m.s. Offset5 bits */
  FM(16#000001C0, false,  10#5);                      /* l.s. Offset5 bits */
  FM(16#00000038,  true, 10#13);                               /* Rn field */
  FM(16#00000007,  true, 10#12)                                /* Rd field */
  ).                                               /* End of F10 expansion */

/*=========================================================================*/
/*      F11 Expansion    */

F11Exp: expr(void)=
  (
  ArmI := 16#E58D0000;
  FM(16#00000800,  true,  10#9);                                  /* L bit */
  FM(16#00000700,  true,  10#4);                               /* Rd field */
  FM(16#000000FF,  true,  10#2)                    /* Word8 shifted left 2 */
  ).                                               /* End of F11 expansion */

/*=========================================================================*/
/*      F12 Expansion    */

F12Exp: expr(void)=
  (
  ArmI := 16#E28D0F00;
  FMI(16#00000800,  true,  10#6);                                /* SP bit */
  FM(16#00000700,  true,  10#4);                               /* Rd field */
  FM(16#000000FF,  true,  10#0);           /* Word8 (shift is in op. code) */

  if ((ThumbI and 16#00000800) = 16#00000000) then               /* PC Add */
    PC := PC and 16#FFFFFFFD
/* This Zeros bit-1 of PC for this Add to PC instruction only JVW 26/3/97  */

  ).                                               /* End of F12 expansion */

/*=========================================================================*/
/*      F13 Expansion    */

F13Exp: expr(void)=
  (
  ArmI := 16#E20DDF00;

  if ((ThumbI and 16#00000080) = 0) then                       /* I.E. S=0 */
    FI(16#00400000,  true,  10#1)                        /* Op Field field */
  else                                                         /* I.E. S=1 */
    FI(16#00200000,  true,  10#1);                       /* Op Field field */
  FM(16#0000007F,  true,  10#0)                            /* SWord7 field */
  ).                                               /* End of F13 expansion */

/*=========================================================================*/
/*      F14 Expansion    */

F14Exp: expr(void)=
 (
  ArmI := 16#E80D0000;
  (case (ThumbI and 16#00000900)                               /* Op field */
    when 16#00000000 => FI(16#01200000,  true,  10#0)           /* L=0 R=0 */
    when 16#00000100 => FI(16#01204000,  true,  10#0)           /* L=0 R=1 */
    when 16#00000800 => FI(16#00B00000,  true,  10#0)           /* L=1 R=0 */
    when 16#00000900 => FI(16#00B08000,  true,  10#0)           /* L=1 R=1 */
  );                             /* NB S bit clear in last case (correct?) */
  FM(16#000000FF,  true,  10#0)                             /* Rlist field */
 ).                                                /* End of F14 expansion */

/*=========================================================================*/
/*      F15 Expansion (Multple Load/Store)   */

F15Exp: expr(void)=
 (
  ArmI := 16#E8A00000;
  FM(16#00000800,  true,  10#9);                                 /* L-bit  */
  FM(16#00000700,  true,  10#8);                               /* Rb field */
  FM(16#000000FF,  true,  10#0);                            /* Rlist field */
  if (((ThumbI and 16#F800) = 16#C800)        /* If LDM & reg. in list ... */
  && (((1 << ((ThumbI >> 8) and 7)) and ThumbI) != 0)) then
    ArmI := ArmI and 16#FFDFFFFF                /* ... clear writeback bit */
 ).                                                /* End of F15 expansion */

/*=========================================================================*/
/*      F16/17 Expansion    */

F1617Exp: expr(void)=
  (
  if ((ThumbI and 16#00000F00) != 16#00000F00) then /* F16 -  NOT SWI Case */
    (
    ArmI := 16#0A000000;
    FM(16#00000F00,  true, 10#20);                    /* Cond. Code  Field */
    FM(16#000000FF, false,  10#1);                   /* SOffset field  >>1 */
    Th_LSA := (ThumbI and 16#00000001);            /* Least sig Inst addr. */

    if ((ThumbI and 16#00000080) = 16#00000080) then     /* Neg - sign ext */
      FI(16#00FFFF80,  true,  10#0)                        /* Neg sign ext */
    )
  else
    (                                                    /*  SWI case - F17*/
    ArmI := 16#EF000000;
    FM(16#000000FF,  true,  10#0)                         /* SOffset field */
    )
  ).                                           /* End of F16/F17 expansion */

/*=========================================================================*/
/*      F18 Expansion    */

F18Exp: expr(void)=
  (
  ArmI := 16#EA000000;
  FM(16#000007FE, false,  10#1);                     /* SOffset field  >>1 */
  Th_LSA := (ThumbI and 16#00000001);               /* Half word alignment */
  if ((ThumbI and 16#00000400) = 16#00000400) then       /* Neg - sign ext */
    FI(16#00FFFC00,  true,  10#0)                          /* Neg sign ext */
 ).                                                /* End of F18 expansion */

/*=========================================================================*/
/*                       F19 Expansion EXPERIMENTAL 999                */

F19Exp: expr(void)=
  (
  Th_special := true;
  ArmI := 16#E28EE000;
  FMI(16#00000800,  true,  10#5);                          /* H-bit select */
                                                        /* NOT-H in bit 16 */
  FM(16#00000800,  true,  10#1);                            /* H in bit-12 */
  FM(16#000007FF, false,  10#1);     /* 11-bit offset  (>>1 for JDG logic) */
  Th_LSA := (ThumbI and 16#00000001)                /* Half word alignment */
  ).                                               /* End of F19 expansion */

/*=========================================================================*/
/*                       FUndef Expansion                                  */

FUndef: expr(void)=
  (
  ArmI := 16#E6000010;                           /* Undefined Instruction  */
/*  printstr(" $ Undefined Format \n"); printhex(ThumbI); printstr("\n")   */
   null
   ).                                           /* End of FUndef expansion */

/*=========================================================================*/

 Th_special := false;
 Th_LSA     := 0;

 (case (ThumbI and 16#0000F000)
   when 16#0000 => F1orF2Exp
   when 16#1000 => F1orF2Exp
   when 16#2000 => F3Exp
   when 16#3000 => F3Exp
   when 16#4000 =>
     (
     if ((ThumbI and 16#00000800) = 0) then
       (
       if ((ThumbI and 16#00000400) = 0) then F4Exp    
       else                                   F5Exp
       )
     else F6Exp
     )
   when 16#5000 =>
     if ((ThumbI and 16#00000200) = 0) then F7Exp
     else                                   F8Exp
   when 16#6000 => F9Exp
   when 16#7000 => F9Exp
   when 16#8000 => F10Exp
   when 16#9000 => F11Exp
   when 16#A000 => F12Exp
   when 16#B000 =>
     (case (ThumbI and 16#00000F00)
       when 16#0000 => F13Exp
       when 16#0100 => FUndef
       when 16#0200 => FUndef
       when 16#0300 => FUndef
       when 16#0400 => F14Exp
       when 16#0500 => F14Exp
       when 16#0600 => FUndef
       when 16#0700 => FUndef
       when 16#0800 => FUndef
       when 16#0900 => FUndef
       when 16#0A00 => FUndef
       when 16#0B00 => FUndef
       when 16#0C00 => F14Exp
       when 16#0D00 => F14Exp
       when 16#0E00 => FUndef
       when 16#0F00 => FUndef
     )
   when 16#C000 => F15Exp
   when 16#D000 => F1617Exp
   when 16#E000 =>
     (
     if ((ThumbI and 16#00000800) = 0) then F18Exp
     else                                   FUndef
     )
   when 16#F000 => F19Exp
 )

).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
