/*------------------------------------------------------------------------*/
/*                                                                        */
/*             AMULET3 Initialisation & Dispatcher (CPU module)           */
/*                                                                        */
/*                                             Created: JVW 13/12/96      */
/*                                       Last Modified: JDG  28/7/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/*                                                                        */
/*------------------------------------------------------------------------*/

/*
    MemBundle : var(chan(Addr_bus_type)).    /* Instruction fetch address */
      Op_code : var(chan(Data_bus_type)).   /* Snooped value from Imemory */
Dmem_copro_out: var(chan(Mem_copro_bus_type)).      /* Data copro. output */
  Dmem_to_CPU : var(chan(Data_bus_type)).           /* Data memory to CPU */
*/


AMULET3(MemBundle : var(chan(Addr_bus_type)),
          Op_code : var(chan(Data_bus_type)),
   Dmem_copro_out : var(chan(Mem_copro_bus_type)),
      Dmem_to_CPU : var(chan(Data_bus_type))):
             expr(void) =
  (
       Branch : var(chan(Branch_type)).                  /* New PC to IAI */
    Imem_addr : var(chan(MemCtrl_bus_type)). /* Memory control pipe input */
      IndirPC : var(chan(Data_bus_type)).    /* Loaded PC returned to IAI */
    Imem_ctrl : var(chan(MemCtrl_bus_type))./* Memory control pipe output */
     Instr_in : var(chan(Instr_pkt_type)).
     Thumb_in : var(chan(Instr_pkt_type)).
    Thumb_out : var(chan(Instr_pkt_type2)).
Decoded_instr : var(chan(Decoded_instr_type)).
       C_flow : var(chan(int)).
       LSM_go : var(chan(bool)).
     ALU_data : var(chan(Queue_input_type)).
     DAI_addr : var(chan(DAI_addr_pkt_type)).
      LD_data : var(chan(Queue_input_type)).        /* Data memory return */
    DAI_instr : var(chan(DAI_instr_type)).
    Dmem_skip : var(chan(Queue_input_type)).
    Writeback : var(chan(Queue_output_type)).
         QTin : var(chan(void)).	       /* Queue slot freed signal */
        QTout : var(chan(void)).	   /* Queue slot available signal */
     Xpipe_in : var(chan(Xpipe_type)).
    BRpipe_in : var(chan(BRpipe_type)).
   BRpipe_out : var(chan(BRpipe_type)).
   Data_abort : var(chan(PSR_type)).
      Dabt_go : var(chan(void)).      /* Allow Data abort code to execute */

    Areg_addr : var(chan(Reg_addr_type)).
       QA_req : var(chan(Forward_req_type)).
      QA_data : var(chan(Forward_data_type)).
         Areg : var(chan(int)).
    Breg_addr : var(chan(Reg_addr_type)).
       QB_req : var(chan(Forward_req_type)).
      QB_data : var(chan(Forward_data_type)).
         Breg : var(chan(int)).
   C1reg_addr : var(chan(Reg_addr_type)).     /* Note: there are two Creg */
      QC1_req : var(chan(Forward_req_type)).  /*    channels to avoid the */
     QC1_data : var(chan(Forward_data_type)). /*    need for demuxing     */
        C1reg : var(chan(int)).               /*    later. This is a LARD */
   C2reg_addr : var(chan(Reg_addr_type)).     /*    convenience.          */
      QC2_req : var(chan(Forward_req_type)).
     QC2_data : var(chan(Forward_data_type)).
        C2reg : var(chan(int)).

/**/

/*
  init_chan(MemBundle,      "Fetch address");
  init_chan(Op_code,        "Op_code");
  init_chan(Dmem_copro_out, "Data/copro. out");
  init_chan(Dmem_to_CPU,    "Data data to CPU");
*/
  init_chan(Branch,         "New PC");
  init_chan(Imem_addr,      "Fetch ctrl");
  init_chan(Imem_ctrl,      "Fetch ctrl'");
  init_chan(IndirPC,        "Loaded PC");
  init_chan(Instr_in,       "Prefetch in");
  init_chan(Thumb_in,       "Prefetch out");
  init_chan(Thumb_out,      "Thumb out");
  init_chan(Decoded_instr,  "Decoded instr");
  init_chan(C_flow,         "Counterflow colour");
  init_chan(LSM_go,         "LSM go/no go");
  init_chan(ALU_data,       "ALU to queue");
  init_chan(DAI_addr,       "ALU to DAI");
  init_chan(LD_data,        "Returned data");
  init_chan(DAI_instr,      "DAI instruction");
  init_chan(Dmem_skip,      "Dmem skip");
  init_chan(Writeback,      "Writeback");
  init_chan(QTin,           "Queue token in");
  init_chan(QTout,          "Queue token out");
  init_chan(Xpipe_in,       "Xpipe input");
  init_chan(BRpipe_in,      "Base restore in");
  init_chan(BRpipe_out,     "Base restore out");
  init_chan(Data_abort,     "Data abort");
  init_chan(Dabt_go,        "Data abort go");

  init_chan(Areg_addr,      "Areg address");
  init_chan(QA_req,         "A forward req");
  init_chan(QA_data,        "Areg forward");
  init_chan(Areg,           "Areg");
  init_chan(Breg_addr,      "Breg address");
  init_chan(QB_req,         "B forward req");
  init_chan(QB_data,        "Breg forward");
  init_chan(Breg,           "Breg");
  init_chan(C1reg_addr,     "Creg address");
  init_chan(QC1_req,        "C forward req");
  init_chan(QC1_data,       "Creg forward");
  init_chan(C1reg,          "Creg");
  init_chan(C2reg_addr,     "Store reg address");
  init_chan(QC2_req,        "Store forward req");
  init_chan(QC2_data,       "Store reg forward");
  init_chan(C2reg,          "Store data");

/**/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Set up "constants"  (located here for ease of change/recompilation)    */

  Mem_ctrl_len  := 0; /* Must be _shorter_ than Imem pipe (is a throttle) */
  Queue_size    := 4;                                  /* Min. 2, Max. 10 */
  Prefetch_size := 0;                                  /* Min. 0, Max. 10 */
  DAI_Ibuf_size := 0;                                  /* Min. 0, Max. 10 */
  BTC_size      := 30;                    /* Even number, Min. 2, Max. 30 */


        IAI_delay := 100;
   Mem_ctrl_delay := 0;
      In_IF_delay := 0;
In_IF_Indir_delay := 0;
      Ipipe_delay := 0;
     REG_rd_delay := 100;
     REG_wr_delay := 100;
   Thumb_on_delay := 100;
  Thumb_off_delay := 10;
  Copro_dec_delay := 0;
        DEC_delay := 0;
  DEC_trash_delay := 30;
  DEC_multi_delay := 0;
  DEC_alloc_delay := 0;
       EXEC_delay := 100;
  EXEC_nogo_delay := 100;
 EXEC_abort_delay := 100;
EXEC_branch_delay := 30;
      Shift_delay := 0;
  Shift_byp_delay := 0;
ALU_logical_delay := 0;
  ALU_arith_delay := 0;
     Br_Lat_delay := 0;
     BRpipe_delay := 0;
        DAI_delay := 0;
        Qin_delay := 0;
       Qout_delay := 0;
       Qfwd_delay := 100;
  Qfwd_cand_delay := 0;
   Qfwd_byp_delay := 0;
     Qtoken_delay := 0;


/**/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Processor initialisation                                               */

  watchdog := 0;         /* Zero "time" before parallel processes started */
  dis_flag := false;

  STR_semaphore := 0;


  (
  IAI(Branch, MemBundle, Imem_addr, IndirPC)| /* Instruction Address Gen. */
  FIFO(Imem_addr, Imem_ctrl, Mem_ctrl_len, 10, false, "MemCtrl: ")|
  IinfBlock(Op_code, Imem_ctrl, Instr_in, IndirPC)|

  FIFO(Instr_in, Thumb_in, Prefetch_size, 0, false, "Ipipe: ")|
  Thumb(Thumb_in, Thumb_out, Dabt_go) |
  Decoder(Thumb_out, Decoded_instr, C_flow, LSM_go,
          Areg_addr, Breg_addr, C1reg_addr, C2reg_addr,
          QA_req, QB_req, QC1_req, QC2_req, DAI_instr, QTout)|

  Execute(Decoded_instr, C_flow, LSM_go, ALU_data, DAI_addr,
          Areg, Breg, C1reg, Xpipe_in, BRpipe_in, Branch, Data_abort)|

  FIFO(BRpipe_in, BRpipe_out, 2, 10, false, "BRpipe: ")|

  Dint(DAI_instr, DAI_addr, Dmem_copro_out, C2reg, Dmem_to_CPU,
                           LD_data, Dmem_skip, Xpipe_in, Data_abort)|

  Queue(ALU_data, LD_data, Dmem_skip, Writeback, BRpipe_out, QTin, Dabt_go,
        QA_req, QB_req, QC1_req, QC2_req, QA_data, QB_data, QC1_data, QC2_data)|
  Registers(Areg_addr,  QA_data,  Areg,  Breg_addr,  QB_data,  Breg,
            C1reg_addr, QC1_data, C1reg, C2reg_addr, QC2_data, C2reg, Writeback)|
  FIFO(QTin, QTout, Queue_size, 10, false, "Qtoken: ")
                                                    /* Queue token return */
/* |
  Memory_support(MemBundle, Op_code, Dmem_copro_out, Dmem_to_CPU)|
  Environment                                  /* Interrupt (etc.) inputs */
*/
  )

  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
