/*------------------------------------------------------------------------*/
/*                                                                        */
/*                     AMULET3 Dummy Coprocessor Support                  */
/*                                                                        */
/*                                             Created: JDG  13/5/97      */
/*                                       Last Modified: JDG  20/5/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/* Herein is contained a set of routines which pretend to perform some    */
/* coprocessor functions.  They should be regarded as placekeepers only.  */
/*                                                                        */
/*------------------------------------------------------------------------*/

copro_instr_type : type = record(instr::int,
                                 value::int,
                                  dest::int). /* Eventually queue address */

/*------------------------------------------------------------------------*/
/* Dummy - copro 0 exists, others don't                                   */

copro_test(Op_code : val(int)) : expr(bool) =
  (
/*
  printstr("Coprocessor instruction "); printhex(Op_code);
  printstr(" checked and ");
  if (((Op_code >> 8) and 16#F) = 4) then printstr("accepted\n")
  else printstr("rejected\n");
*/
  (((Op_code >> 8) and 16#F) = 4) ||
  (((Op_code >> 8) and 16#F) = 5)
  ).

/*------------------------------------------------------------------------*/

copro_reg_lock : var(int).
copro_register : var(int^16).

/*------------------------------------------------------------------------*/
/* Currently picks up a single register value - not necessarily the right */
/* one.                                                                   */

copro_decode(op_code : val(int)) : expr(copro_instr_type) =
  (
  output : var(copro_instr_type).

  output->instr := op_code;
/*
printstr("Copro. decode - lock = "); printint(copro_reg_lock); printstr("\n");
*/
  if ((op_code and 16#0E000000) = 16#0C000000) then
    (                                                              /* CDT */
    if ((op_code and 16#00100000) = 0) then                        /* STC */
      wait_until(copro_reg_lock = 0)
    else                                                           /* LDC */
      copro_reg_lock := copro_reg_lock + 1;

    output->dest  := (op_code>>12) and 16#F;      /* Value doesn't matter */
    output->value := copro_register[output->dest]              /* for LDC */
    )
  else
    (
    if ((op_code and 16#0F000010) = 16#0E000000) then
      (                                                            /* CDP */
      wait_until(copro_reg_lock = 0);

      output->dest  := (op_code>>12) and 16#F;
      output->value := copro_register[op_code and 16#F];            /* Rm */
      copro_reg_lock := copro_reg_lock + 1
      )
    else
      (                                                            /* CRT */
      if ((op_code and 16#00100000) = 0) then                      /* MCR */
        copro_reg_lock := copro_reg_lock + 1
      else
        wait_until(copro_reg_lock = 0);            /* MRC needs coherency */
      output->dest  := (op_code>>16) and 16#F;                      /* Rn */
      output->value := copro_register[(op_code>>16) and 16#F]       /* Rn */
      )
    );

  output
  ).


/*------------------------------------------------------------------------*/
/* See to any "side effects" such as asserting interrupt lines            */

copro_execute(Command : val(copro_instr_type)) : expr(void) =
  (
  if ((Command->instr and 16#0F000F10) = 16#0E000500) then
    (
    case (Command->instr and 16#00F00000)
      when 16#00100000 =>
        sink(spawn(wait_for(100*(Command->value)); FIQ := true, forparnum))
/* Delayed_write(FIQ, true, 100*(Command->value))  */
      when 16#00200000 =>
        sink(spawn(wait_for(100*(Command->value)); IRQ := true, forparnum))
/* Delayed_write(IRQ, true, 100*(Command->value))  */
      when 16#00300000 => FIQ := false
      when 16#00400000 => IRQ := false
      default null
    )
  ).

/*------------------------------------------------------------------------*/
/* Simplified write procedure                                             */

copro_write(Data : var(chan(Data_bus_type))) : expr(void) =
  (
  latch : var(Data_bus_type).

  forever
    (
    Data?(latch := (?Data));

    if (latch->abort) then
      (
/*
      printstr("Coprocessor result for C"); printint(latch->extra);
      printstr(" discarded\n")
*/ null
      )
    else
      (
      copro_register[latch->extra] := latch->data;
/*
printstr("C"); printint(latch->extra); printstr(" := ");
printhex(latch->data); printstr("\n")
*/ null
      );

    copro_reg_lock := copro_reg_lock - 1 /* Very simple locking mechanism */

    )
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
