/*------------------------------------------------------------------------*/
/*                                                                        */
/*                           AMULET3 Data Interface                       */
/*                                                                        */
/*                                             Created: JDG 19/12/96      */
/*                                            Modified: SM   18/2/97      */
/*                                       Last Modified: JDG  25/9/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/* Timing model should take into account incrementer & (last) cycle spec. */
/*                                                                        */
/* "Xpipe" is contained in this module                                    */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*  N.B. Latches SHOULD be at END of pipeline stage.                      */
/*------------------------------------------------------------------------*/

Dint(instruction : var(chan(DAI_instr_type)),
     address_in  : var(chan(DAI_addr_pkt_type)),
     mem_address : var(chan(Mem_copro_bus_type)),
     store_data  : var(chan(int)),
     mem_data    : var(chan(Data_bus_type)),
     queue_data  : var(chan(Queue_input_type)),
     mem_nogo    : var(chan(Queue_input_type)),
     Xpipe_in    : var(chan(Xpipe_type)),
     Data_abort  : var(chan(PSR_type))):
     expr(void) =
  (

/* Record type declarations for DAI FIFO                                  */

  D_fifo_type : type = record(queue_addr::int,
                                   copro::Copro_type, /* Coprocessor info */
                                               /* May well be unnecessary */
                              mem_colour::int,
                              last_cycle::bool,
                          store_complete::bool,          /* For semaphore */
                                    Read::bool,
                                    Size::int,
                                     sex::bool,            /* Sign extend */
                                     LSM::bool, /* Prevents data rotation */
                                     b2b::int,            /* Byte address */
                               bv_colour::int). /* for block view */

  DAI_instr_out : var(chan(DAI_instr_type)).

  DAI_fifo_in   : var(chan(D_fifo_type)).
  DAI_fifo_out  : var(chan(D_fifo_type)).

  Xpipe_out     : var(chan(Xpipe_type)).
  Dabt_decouple : var(chan(PSR_type)).

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

  DAI(instruction : var(chan(DAI_instr_type)),
      address_in  : var(chan(DAI_addr_pkt_type)),
      mem_address : var(chan(Mem_copro_bus_type)),
      store_data  : var(chan(int)),
      mem_nogo    : var(chan(Queue_input_type)),
      DAI_fifo_in : var(chan(D_fifo_type))) :
      expr(void)=
    (
    addr_int_in   : var(DAI_instr_type).
    addr_int_out  : var(Mem_copro_bus_type).
    nogo_pkt      : var(Queue_input_type).
    address       : var(DAI_addr_pkt_type).
    get_addr      : var(bool).

    /* Set up invariant (or irrelevant) fields                            */
    nogo_pkt->data    := 0;
    nogo_pkt->invalid := true;
    nogo_pkt->abort   := false;
    nogo_pkt->has_BR  := false;             /* Nothing entered in BR pipe */

    get_addr          := true;                 /* Get address after reset */

    forever
      (
      instruction?(addr_int_in:=(?instruction));                 /* Latch */

      active_block("dai",addr_int_in->bv_colour,(

      addr_int_out->Read       := addr_int_in->read;
      addr_int_out->Write      := addr_int_in->write;
      addr_int_out->Size       := addr_int_in->size;
      addr_int_out->copro      := addr_int_in->copro;

      if (addr_int_in->write) then                /* Store data required? */
        (
        store_data?(addr_int_out->Dout := (?store_data));
        if (addr_int_in->size < 4) then
          (
          addr_int_out->Dout := ((addr_int_out->Dout  or 16#FFFF0000)
                      and ((addr_int_out->Dout << 16) or 16#0000FFFF));
          /* stores to the memory a halfword duplicated to a word, "STRH" */
                                       /* ex: ab cd ef gh --> ef gh ef gh */

	  if(addr_int_in->size < 2) then
            addr_int_out->Dout := ((addr_int_out->Dout  or 16#FF00FF00)
                         and ((addr_int_out->Dout <<  8) or 16#00FF00FF))
      /* Stores to the memory a byte duplicated to a word, "STRB", STRBT" */
                                       /* ex: ef gh ef gh --> gh gh gh gh */
          )
        )
      else
        addr_int_out->Dout := 0;                          /* for tidiness */

      wait_for(DAI_delay);

/*    if (addr_int_in->get_addr) then address_in?(address:=(?address_in));*/
      if ((addr_int_in->get_addr) != get_addr) then printstr("Yikes!\n");

      if (get_addr) then address_in?(address:=(?address_in));

                                        /* TIMING ??? */

      addr_int_out->invalidate := (address->invalidate);

      if (address->invalidate) then
        (             /* Abort memory operation and signal queue directly */
        nogo_pkt->queue_addr := addr_int_in->queue_addr;
        nogo_pkt->bv_colour  := addr_int_in->bv_colour;
        mem_nogo!nogo_pkt;
        if (addr_int_in->copro->use_copro) then mem_address!addr_int_out
                             /* Address field not used - just invalidates */
        )             /* NB - FIFO not invoked & no return value expected */
      else
        (                                                /* Valid request */
/*      if ((!addr_int_in->get_addr) || addr_int_in->pre_inc) then        */
/*      if ((!get_addr) || addr_int_in->pre_inc) then                     */
        if ((!get_addr) || address->pre_inc) then
          address->addr := address->addr + 4;

        addr_int_out->addr := address->addr;
        addr_int_out->mode := (if addr_int_in->force_user then 0
                               else address->mode);
        addr_int_out->Qcol := address->colour;
/*      addr_int_out->seq  := (!addr_int_in->get_addr);                   */
        addr_int_out->seq  := (!get_addr);        /* Sequential if didn't */

        (                    /* Output in parallel to memory and DAI FIFO */
          active_block("dmem",addr_int_in->bv_colour,(
            mem_address!addr_int_out
          ))
        |
          (
          /* queue_addr, last_cycle, Read, Size, sex, LSM, b2b,           */
          /*  store_complete go through DAI FIFO                          */
          f_in          : var(D_fifo_type).

          f_in->queue_addr     := addr_int_in->queue_addr;
          f_in->copro          := addr_int_in->copro;
                           /* In case it's needed - probably okay to omit */
          f_in->mem_colour     := address->colour and 2;/* Memory bit only*/
          f_in->store_complete := addr_int_in->write && addr_int_in->last_cycle;
          f_in->last_cycle     := addr_int_in->last_cycle;
          f_in->Read           := addr_int_in->read;
          f_in->Size           := addr_int_in->size;
          f_in->sex            := addr_int_in->sex;
          f_in->LSM            := addr_int_in->LSM;
          if (addr_int_in->LSM || addr_int_in->copro->use_copro) then
            f_in->b2b          := 0
          else
            f_in->b2b          := address->addr and 16#00000003;
          f_in->bv_colour      := addr_int_in->bv_colour;

          DAI_fifo_in!f_in
          )

        )                                       /* End of parallel output */

        );
      get_addr := (addr_int_in->last_cycle) || (address->invalidate)
        /* Get new address only if this is the end of current instruction */
        )) /* active_block */
      )                            /* End of forever ----  Memory look up */
    ).                                                      /* End of DAI */

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

  D2Q(DAI_fifo_out  : var(chan(D_fifo_type)),
      mem_data      : var(chan(Data_bus_type)),
      queue_data    : var(chan(Queue_input_type)),
      Dabt_decouple : var(chan(PSR_type)),
      Xpipe_out     : var(chan(Xpipe_type))):
      expr(void) =
    (
    data_output : var(Queue_input_type).
    Data        : var(int).
    QB          : var(int).
    Aborted     : var(bool).
    Abort_col   : var(int).

                    /* ______________________________________*/

    Aborted := false;

    forever				
      (
      mem_data? DAI_fifo_out?
        (
/* *** TIMING MISSING ***/

        /*  Join FIFO and memory fields together                          */
        data_output->queue_addr := (?DAI_fifo_out)->queue_addr;
        data_output->invalid    := (!(?mem_data)->abort)
                                && (!(?DAI_fifo_out)->Read);
                        /* write back if load or if aborted (for PC=>LR)  */
        data_output->abort      := (?mem_data)->abort;
        data_output->has_BR     := (?DAI_fifo_out)->last_cycle
                                && (?DAI_fifo_out)->LSM;
        data_output->bv_colour  := (?DAI_fifo_out)->bv_colour;

        QB := 8 * (?DAI_fifo_out)->b2b;

        if (!(?DAI_fifo_out)->LSM) then      /* LSM now zeros LSBs in DAI */
          (
          if (QB = 0) then
            Data := (?mem_data)->data         /* LARD can't manage << 32  */
          else
            Data := (((?mem_data)->data >> QB) and (not(16#FFFFFFFF << (32 - QB)))
                  or ((?mem_data)->data << (32 - QB)));
                       /* rotates by 0, 8, 16 or 24 bits according to b2b */

          if ((?DAI_fifo_out)->Size < 4) then
            (
            Size : var(int).

            Size := (?DAI_fifo_out)->Size;
            if ((?DAI_fifo_out)->sex && ((Data and (1<<(8*Size - 1))) != 0 )) then
              Data := Data or (not((1 << (8*Size))-1))
                                /* Sign-extension for "LDRSH", "LDRSB"
                                   i.e data or 16#FFFF0000 for halfwords
                                        data or 16#FFFFFF00 for bytes     */
            else
              Data := Data and ((1 << (8*Size))-1)
                               /* Zero-extension for "LDRH", "LDRB", "LDRBT"
                                i.e data and 16#0000FFFF for halfwords
                                    data and 16#000000FF for bytes        */
            )
          )
        else
          Data := (?mem_data)->data;

        data_output->data := Data;
                                /* Normal data return (poss. overwritten) */

        if ((!Aborted) || (Abort_col != (?DAI_fifo_out)->mem_colour)) then
          (
          Aborted   := (?mem_data)->abort;           /* Abort if required */
          Abort_col := (?DAI_fifo_out)->mem_colour;
          if ((?mem_data)->abort) then/* Value according to memory result */
            (                /* Read Xpipe value without acknowledging it */
            wait_until(chan_ready(Xpipe_out));
            data_output->data := (?Xpipe_out)->PC;        /* Value for LR */
            Dabt_decouple!(?Xpipe_out)->PSR
            )            /* send CPSR back to execute as abort initiation */
          )
        else
          data_output->invalid := true;                /* In abort shadow */


        /* Pull PC from Xpipe */
        if ((?DAI_fifo_out)->last_cycle) then Xpipe_out?null;
                                       /* Values already read if required */

        if (?DAI_fifo_out)->store_complete then
          STR_semaphore := STR_semaphore - 1;

        queue_data!data_output
        )                              /* memory and DAI FIFO acknowledge */
      )                                                 /* End of forever */
    ).                                                      /* End of D2Q */

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

    init_chan(DAI_instr_out, "DAI instr out");

    init_chan(DAI_fifo_in, "Dai_fifoin");
    init_chan(DAI_fifo_out, "Dai_fifout");
    init_chan(Dabt_decouple, "Dabt_latch");
    init_chan(Xpipe_out, "Xpipe_out");

/* Parallel Blocks                                                        */
/* There are six parallel processes created below.  These are:            */
/* DAI_Ipipe which buffers DAI instructions (may well be zero length)     */
/* DAI which receives commands, reads addresses and outputs to memory     */
/* DAI_PIPE which is a memory control pipeline                            */
/* D2Q receives returned data and packages it for the queue.              */
/*    It also responds to data aborts by outputting a command to:         */
/*       DABT_LATCH which feeds the execution unit                        */
/*    and reads extra information from:                                   */
/*       X_PIPE, which holds old PC/PSRs                                  */

     (
     FIFO(instruction, DAI_instr_out, DAI_Ibuf_size, 10,
          false, "DAI Ipipe") |                 /* DAI instruction buffer */

/* This latch avoids a deadlock problem by decoupling the DAI so that it  */
/* can start complete operation without waiting for the execution unit to */
/* finish a cycle.                                                        */
     Latch(Dabt_decouple, Data_abort, 0, false, "Dabt_latch") |
                                                   /* Spawn latch process */

     DAI(DAI_instr_out, address_in, mem_address, store_data, mem_nogo,
            DAI_fifo_in)  |

     FIFO(DAI_fifo_in, DAI_fifo_out, 4, 10, false, "DAI_PIPE") |
                                     /* Instructions are pipelined in DAI */

     FIFO(Xpipe_in, Xpipe_out, 2, 10, false, "X_PIPE") |
                                  /* PC through Xpipe  when load or store */

     D2Q(DAI_fifo_out, mem_data, queue_data, Dabt_decouple, Xpipe_out)
    )
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
