/*------------------------------------------------------------------------*/
/*                                                                        */
/*                   AMULET3 Instruction Decoder - part 1                 */
/*                                                                        */
/*                                             Created: JDG  31/1/97      */
/*                                       Last Modified: JDG  11/7/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/*  Check that ALL output bits defined for every instruction category     */
/*                                                                        */
/*------------------------------------------------------------------------*/

Queue_CAM_type : type = record(addr::int,
                               mode::int).

Queue_CAM  : var(Queue_CAM_type^10).
DEC_Q_addr : var(int).
Q_Fcol     : var(int).

/*------------------------------------------------------------------------*/
/* LOOKUP DELAY MODEL MISSING                                             */
/* NB this procedure does not return until the register read is accepted  */

Queue_CAM_lookup(Register    : val(int),
                 Mode        : val(int),
                 Colour      : val(int),
                 Forward_req : var(chan(Forward_req_type))):
                 expr(void)=
  (
  to_queue : var(Forward_req_type).
  i : var(int).

  if (Register = 15) then                       /* Never forward R15 (PC) */
    forseq i in 0 to (Queue_size-1) do to_queue->vector[i] := false
  else
    forseq i in 0 to (Queue_size-1) do
      to_queue->vector[i] := (Reg_no(Register, Mode)
                            = Reg_no((Queue_CAM[i])->addr, (Queue_CAM[i])->mode));

  /* DELAYS ??  */

  forseq i in Queue_size to 9 do to_queue->vector[i] := false;/* Prettify */

  to_queue->start  := DEC_Q_addr;
  to_queue->Fcol   := Q_Fcol;
  to_queue->colour := Colour;               /* Used in across abort entry */
  Forward_req!to_queue       /* Ask queue forwarding process to send data */
  ).

/*------------------------------------------------------------------------*/
/* If the operation has an ALU destination of R15 then no action results  */
/* in which case ignore the return value                                  */

Dec_allocate(QTout    : var(chan(void)),
             Reg_addr : val(int),
             Reg_mode : val(int),
             Mem_op   : val(bool)):
             expr(int)=
  (
  if ((Reg_addr != 15) || Mem_op) then
    (
    QTout?null;            /* Wait for queue location to become available */
    DEC_Q_addr := (DEC_Q_addr+1) mod Queue_size;    /* Next queue address */
    bview_set_colour("queue"@i2s(DEC_Q_addr),"white");
    if (DEC_Q_addr=0) then Q_Fcol := Q_Fcol xor 1;/*Change expected colour*/

    (Queue_CAM[DEC_Q_addr])->addr := Reg_addr; /* Also saving register ID */
    (Queue_CAM[DEC_Q_addr])->mode := Reg_mode
    )
  else
    printstr("Queue allocation attempt failed\n");

  wait_for(DEC_alloc_delay);

  DEC_Q_addr
  ).

/*------------------------------------------------------------------------*/
/* Shift */

shift(op_code : val(int)): expr(bool) =
  (((op_code and 16#0E000000) = 16#06000000)              /* LDR/STR reg. */
|| ((op_code and 16#0E000000) = 16#00000000))                 /* Data op. */
  && ((op_code and 16#00000FF0) != 16#00000000)    /* Not a shift by zero */
  .

/*------------------------------------------------------------------------*/
/*
store(op_code : val(int)): expr(bool) =
  (
     ((op_code and 16#0C100000) = 16#04000000)                     /* STR */
  || ((op_code and 16#0E100000) = 16#08000000)                     /* STM */
  || ((op_code and 16#0FB00FF0) = 16#01000090)                     /* SWP */
  ||(((op_code and 16#0E100090) = 16#00000090) /* Half-word (etc.) stores */
           && ((op_code and 16#00000060) != 0))
  || ((op_code and 16#0E100000) = 16#0C000000)                     /* STC */
  || ((op_code and 16#0F100010) = 16#0E000010)                     /* MCR */
  ).                     /* ^              ^         Possibly losable ??? */
*/
/*------------------------------------------------------------------------*/
/* Flags which instructions take "undefined" trap                         */

  /* NB *maybe* some transfers missing (reg. offset, non zero bits 11..8) */
Undef_fn(Op_code : val(int),
         Exec_op : val(Decoded_instr_type)):
         expr(bool) =
  (
  Arith_ext    : var(bool).
  Ctrl_ext     : var(bool).
  Transfer_ext : var(bool).

  Arith_ext    := ((Op_code and 16#0F0000F0)  = 16#00000090);
  Ctrl_ext     := ((Op_code and 16#0D900000)  = 16#01000000)
               && ((Op_code and 16#02000090) != 16#00000090);
  Transfer_ext := ((Op_code and 16#0E000090)  = 16#00000090)
               && ((Op_code and 16#01000060) != 16#00000000);


  ((Op_code and 16#0E000010)  = 16#06000010)
|| (Arith_ext && (!Exec_op->operation->use_MUL))
|| (Transfer_ext && (!(((Exec_op->Op_code and 16#00100000) != 0)  /* LDRH */
                   ||  ((Exec_op->Op_code and 16#00000060) = 16#000000020)
                                                     /* unsigned halfword */
                   ||  ((Exec_op->Op_code and 16#00B00F60) = 16#000000000))))
                                                      /* Enough to ID SWP */
|| (Ctrl_ext && (!(Exec_op->branch->BX
                || Exec_op->flags->MRS
                || Exec_op->flags->MSR)))
  ).

/*------------------------------------------------------------------------*/
/* Find which input buses are used by the execution unit (NB not stores)  */
/* THIS IS NOT GUARANTEED COMPLETE YET                                    */
/* EXPERIMENT - try decoding and suppressing A := R0 read on MOV, MVN     */

Register_usage(op_code : val(int),
               A       : var(bool),
               B       : var(bool),
               C       : var(bool),
               Imm     : var(bool)):
               expr(void) =
  (
  A   := ((op_code and 16#0A000000) != 16#0A000000);/* Lose branches, etc.*/
  B   := ((op_code and 16#0E000000) = 16#00000000)      /* Reg. data ops. */
      || ((op_code and 16#0E000000) = 16#06000000);       /* Reg. LDR/STR */
  C   := ((op_code and 16#0E000090) = 16#00000010)   /* Reg. shift (& BX) */
      || ((op_code and 16#0F0000F0) = 16#00000090);                /* MUL */

  Imm := (((op_code xor (op_code >> 1)) and 16#02000000) != 0)   /* 26!=25 */
       || ((op_code and 16#0C000000) = 16#08000000)
       ||  (((op_code and 16#0C400090)  = 16#00400090)    /* half wd xfer */
         && ((op_code and 16#00000060) != 16#00000000))
                    /* Immediate indicates usage rather than transmission */
                    /* Immediate field is contained in instruction bundle */
  ).

/*------------------------------------------------------------------------*/
/* Find least significant bit set in "Rd_gen_temp", then clear it         */

Rd_gen_temp : var(int).

Rd_gen(last : var(bool)) : expr(int) =
  (
  Rd :var(int).

  Rd := 0;
  if (Rd_gen_temp != 0) then
    (
    while (((16#0001 << Rd) and Rd_gen_temp) = 0) do Rd := Rd + 1;
    Rd_gen_temp := Rd_gen_temp and not(16#0001 << Rd) /* Clear latest bit */
    );
  last := (Rd_gen_temp = 0);                       /* Any more bits left? */
  Rd                   /* Error case (input of Rd_gen_temp = 0) returns 0 */
  ).

/*------------------------------------------------------------------------*/

Ngen(op_code : val(int)) : expr(int) =
  (
  count : var(int).
  mask  : var(int).

  count := 0;
  mask  := 16#8000;

  while (mask != 0) do
    (
    if ((op_code and mask) != 0) then count := count + 1;
    mask := mask >> 1                       /* Shift to zero to terminate */
    );

  count                        /* Outputs zero if no bits set; acceptable */
  ).

/*------------------------------------------------------------------------*/
/* Immediate extraction */

Immediate(op_code : val(int), LSB : val(int), special : val(bool)) :
          expr(Immediate_type) =
  (
  rotate : var(int).
  value  : var(int).
  result : var(Immediate_type).

  result->default_C := true;         /* For appearance - really undefined */
  result->Imm_C     := 0;

  if (!special) then
    (case ((op_code >> 25) and 7)

      when 0 =>
        result->Immediate := ((op_code and 16#00000F00) >> 4)   /* OK-ish */
                           or (op_code and 16#0000000F)      /* Half word */
      when 1 =>
        (                                               /* Data operation */
        rotate := (op_code>>7) and 16#0000001E;
        value  := op_code and 16#000000FF;
        result->default_C := (rotate = 0);           /* Use the CPSR flag */
        if (!result->default_C) then
          result->Imm_C := (value >> (rotate-1)) and 1;/* Carry otherwise */
        result->Immediate := (value >> rotate) or (value << (32-rotate))
        )

      when 2 => result->Immediate := (op_code and 16#00000FFF) /* LDR/STR */

      when 4 => result->Immediate := (Ngen(op_code) << 2)      /* LDM/STM */

      when 5 =>
        (                                                       /* Branch */
        value := op_code and 16#00FFFFFF;
        if ((value and 16#00800000) != 0) then             /* Sign extend */
          value := -((value xor 16#00FFFFFF) + 1);
        result->Immediate := (value << 2) or (LSB << 1)
        )                                  /* Needs extra bit (for Thumb) */

      when 6 => result->Immediate := (op_code and 16#000000FF) << 2
                                             /* Coprocessor data transfer */
      default   result->Immediate := 0             /* For appearance only */
    )
  else
    (                                 /* Special Thumb BL immediate modes */
    if (((op_code and 16#00001000) = 0)                /* If Rd is LR ... */
    && ((op_code and 16#00000200) != 0)) then  /* ... and offset negative */
      result->Immediate := ((op_code << 2) or (LSB << 1))  or 16#FFFFF000
    else
      result->Immediate := ((op_code << 2) or (LSB << 1)) and 16#00000FFE;
    if ((op_code and 16#00001000) = 0) then          /* If BL0 then << 11 */
      result->Immediate := result->Immediate << 11
    );

  result
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
