/*------------------------------------------------------------------------*/
/*                                                                        */
/*                   AMULET3 Instruction Decoder - part 2                 */
/*                                                                        */
/*                                             Created: JDG  31/1/97      */
/*                                       Last Modified: JDG  25/9/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/*  Check that ALL output bits defined for every instruction category     */
/*                                                                        */
/*------------------------------------------------------------------------*/

Decoder(Instruction : var(chan(Instr_pkt_type2)),
          Instr_out : var(chan(Decoded_instr_type)),
             C_flow : var(chan(int)),
             LSM_go : var(chan(bool)),
          Areg_addr : var(chan(Reg_addr_type)),
          Breg_addr : var(chan(Reg_addr_type)),
          Creg_addr : var(chan(Reg_addr_type)),
     Store_reg_addr : var(chan(Reg_addr_type)),
             QA_req : var(chan(Forward_req_type)),
             QB_req : var(chan(Forward_req_type)),
             QC_req : var(chan(Forward_req_type)),
         QStore_req : var(chan(Forward_req_type)),
             to_DAI : var(chan(DAI_instr_type)),
              QTout : var(chan(void))):
              expr(void)=
  (
  IR     : var(Instr_pkt_type2).
  to_ALU : var(Decoded_instr_type).
  DAI_op : var(DAI_instr_type).
  C_flow_col : var(int).

  default_C  : var(bool).                               /* Ignore "Imm_C" */
  Imm_C      : var(int).                /* Carry from immediate extractor */

  Link_class : var(bool).                  /* Anything which saves an R14 */
  BL1        : var(bool).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Decode bit crunching                                                   */

(DEC(Op_code   : val(int),
     BL1       : val(bool),
     Extension : val(int),
     Exec_op   : var(Decoded_instr_type),
     DAI_op    : var(DAI_instr_type),
     Link_class: var(bool)):               /* Anything which saves an R14 */
     expr(void) =
  (
  Operation : var(int).
  Rn15 : var(bool).
  Rd15 : var(bool).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Decode instruction with extension                                      */

DEC_ext : expr(void) =
  (
  Instr_ext    : var(bool).                      /* FIQ, IRQ, Pabt, Undef */
  Copro        : var(bool).            /* Untaken coprocessor instruction */

  Instr_ext    := (Extension < 16#20);           /* FIQ, IRQ, Pabt, Undef */
  Copro        := (Extension = 16#15); /* Untaken coprocessor instruction */

/*
  Exec_op->condition  := (if Instr_ext then
                           (if Copro then
                              (Op_code >> 28) and 16#F /* Copro.->undefCC */
                            else 16#E)      /* Other exceptions -> always */
                          else Extension and 16#F); /* Predicted branches */
*/
  Exec_op->condition  := (Op_code >> 28) and 16#F;/*Exceptions already in */

  Exec_op->Exception  := (if (Extension = 16#1F) then 12          /* Pabt */
                          else (if Copro then 4 else 0));     /* Und/none */

  Exec_op->flags->LSM := false;
  Exec_op->flags->MRS := false;
  Exec_op->flags->MSR := false;

  Exec_op->operation->shift     := false;
  Exec_op->operation->shift_op  := 0;                       /* Don't care */
  Exec_op->operation->shift_imm := false;                   /* Don't care */
  Exec_op->operation->shift_amount := 0;                    /* Don't care */
  Exec_op->operation->use_MUL   := false;                           /* No */
  Exec_op->operation->use_ALU   := false;
  Exec_op->operation->set_flags := false;

  Exec_op->branch->predicted := (Extension >= 16#20);
  Exec_op->branch->branch    := Exec_op->branch->predicted;
  Exec_op->branch->link      := ((Extension and 16#10) != 0);
                               /* FIQ, IRQ, Pabt, Undef  and predicted BL */


  Link_class := Exec_op->branch->link;        /* TEMP - while passed back */
       /* Looks to get (correct) mode from wrong place below for pred. BL */

  Exec_op->branch->BX        := false;
  Exec_op->branch->Thumb_BL  := false;

  Exec_op->to_queue  := Exec_op->branch->link;

/* NB An "Extension" of 16#15 indicates a coprocessor operation and maybe */
/*   an invalidation signal should be sent to the DAI (currently omitted) */
/* If not the coprocessor instruction must have been destroyed at decode  */

  Exec_op->to_DAI    := false;

  Exec_op->to_SPSR   := Instr_ext;
  Exec_op->LD_PC     := false;
  Exec_op->dest_PC   := false;
  DAI_op->read       := false;                              /* Don't care */
  DAI_op->write      := false;                              /* Don't care */
  Exec_op->store     := Exec_op->to_DAI && DAI_op->write;
  Exec_op->addr_pre_inc := false;                           /* Don't care */

  DAI_op->get_addr   := true;                               /* Don't care */
  DAI_op->force_user := false;                              /* Don't care */
  DAI_op->pre_inc    := false;                              /* Don't care */
  DAI_op->LSM        := Exec_op->flags->LSM;
  DAI_op->size       := 4;                                  /* Don't care */
  DAI_op->sex        := false;                              /* Don't care */

  DAI_op->copro->use_copro := Copro;
  DAI_op->copro->copro_no  := (Op_code>>8) and 16#F;
  DAI_op->copro->instr     := Op_code and 16#03100010;

  Exec_op->operation->ALU_op := 0;                          /* Don't care */

  Exec_op->PSR_ctrl->mode := (case Extension
                                when 16#19 =>  1              /* FIQ mode */
                                when 16#1A =>  2              /* IRQ mode */
                                when 16#1F =>  7        /* Prefetch abort */
                                when 16#15 => 11 /* Undefined (no copro.) */
                                default Exec_op->PSR_ctrl->mode
                             );

                      /* Disable interrupts in the case of prefetch abort */
  if (Extension = 16#1F) then Exec_op->PSR_ctrl->IRQ_En := false;

  if ((Extension < 32) || (Extension > 127)) then
    (case Extension
      when 16#00 => null
      when 16#19 => null
      when 16#1A => null
      when 16#1F => null
      when 16#15 => null
      default (printint(Extension); printstr(" is an illegal extension");
               Run := false)
    )

  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Decode instruction without extension                                   */

DEC_normal : expr(void) =
  (
  Arith_type   : var(bool).         /* Probably an arithmetic instruction */

  Arith_ext    : var(bool).     /* Arithmetic instruction extension space */
  Ctrl_ext     : var(bool).        /* Control instruction extension space */
  Transfer_ext : var(bool).     /* Load/store instruction extension space */

  SWP_instr    : var(bool).                            /* SWP instruction */
  SWI_instr    : var(bool).                            /* SWI instruction */
  LSR_instr    : var(bool).      /* LDR or STR (or undefined) instruction */
  LSM_instr    : var(bool).                     /* LDM or STM instruction */
  Undefined    : var(bool).                      /* Undefined instruction */

  Coprocessor  : var(bool).                    /* Coprocessor instruction */

  P_bit        : var(bool).
  U_bit        : var(bool).
  B_bit        : var(bool).
  W_bit        : var(bool).
  L_bit        : var(bool).
  bits_6_5     : var(bool).

  Halfword_bits: var(bool).


  P_bit        := ((Op_code and 16#01000000) != 0);
  U_bit        := ((Op_code and 16#00800000) != 0);
  B_bit        := ((Op_code and 16#00400000) != 0);
  W_bit        := ((Op_code and 16#00200000) != 0);
  L_bit        := ((Op_code and 16#00100000) != 0);
  bits_6_5     := ((Op_code and 16#00000060) != 0);

  Arith_type   := ((Op_code and 16#0C000000)  = 16#00000000);

  Arith_ext    := ((Op_code and 16#0F0000F0)  = 16#00000090);
  Ctrl_ext     := ((Op_code and 16#0D900000)  = 16#01000000)
               && ((Op_code and 16#02000090) != 16#00000090);
  Transfer_ext := ((Op_code and 16#0E000090)  = 16#00000090)
               && ((Op_code and 16#01000060) != 16#00000000);

  SWP_instr    := Transfer_ext && (!bits_6_5);
  SWI_instr    := ((Op_code and 16#0F000000)  = 16#0F000000);      /* SWI */
  LSR_instr    := ((Op_code and 16#0C000000)  = 16#04000000);
  LSM_instr    := ((Op_code and 16#0E000000)  = 16#08000000);

  Coprocessor  := ((Op_code and 16#0E000000)  = 16#0C000000)   /* LDC/STC */
               || ((Op_code and 16#0F000000)  = 16#0E000000);  /* CDP/CRT */

  Exec_op->condition  := (Op_code >> 28) and 16#F;

  Exec_op->flags->LSM := LSM_instr;
  Exec_op->flags->MRS := Ctrl_ext && ((Op_code and 16#022F0FFF) = 16#000F0000);
  Exec_op->flags->MSR := Ctrl_ext && (((Op_code and 16#0220FFF0) = 16#0020F000)
                                   || ((Op_code and 16#0220F000) = 16#0220F000));

  Exec_op->branch->BX := Ctrl_ext
                      && ((Op_code and 16#026FFFF0) = 16#002FFF10);
  Exec_op->branch->Thumb_BL     := BL1;

  Exec_op->operation->shift     := (shift(Op_code)   /* Separate function */
                    && (!(Arith_ext || Transfer_ext || Exec_op->branch->BX)));

  Exec_op->operation->shift_op  :=
              (if ((Op_code and 16#00000FF0) = 16#00000060) then 4 /* RRX */
               else (Op_code >> 5) and 3);
  Exec_op->operation->shift_imm := ((Op_code and 16#010) = 0);
  Exec_op->operation->shift_amount := (if ((Op_code and 16#F80) = 0) then 32
                                       else (Op_code >> 7) and 16#1F);
                               /* LSL #0 bypasses, RRX assumes amount = 1 */

  Exec_op->operation->use_MUL   := Arith_ext
                                && ((Op_code and 16#00C00000) != 16#00400000);

  Exec_op->operation->use_ALU   := ((Op_code and 16#0A000000) != 16#0A000000);
  Exec_op->operation->set_flags := Arith_type && L_bit   /* L_bit = S bit */
                                && (!(((Op_code and 16#02000090)  = 16#00000090)
                                   && bits_6_5));   /* Filter undefineds? */

  Undefined  := Undef_fn(Op_code, Exec_op);

  Link_class := ((Op_code and 16#0B000000)  = 16#0B000000)      /* BL/SWI */
             || Undefined || BL1;

  Exec_op->branch->branch    := ((Op_code and 16#0E000000) = 16#0A000000);
  Exec_op->branch->predicted := false;
  Exec_op->branch->link      := Link_class;

  Exec_op->to_queue  := (Arith_type && (!Rd15)        /* and exclude CMPs */
                        && ((Op_code and 16#01900000) != 16#01100000)
                        && (!Transfer_ext))
                     || Link_class            /* Includes undefineds etc. */
                     || ((!Rn15) &&         /* Data transfers; writeback? */
                         ((LSM_instr && W_bit)                   /* LSM ! */
                       || (LSR_instr && (W_bit || (!P_bit)))     /* LSR ! */
                       || (Transfer_ext && (W_bit || (!P_bit)))
                                                             /* NB no SWP */
                       || ((Op_code and 16#0E200000) = 16#0C200000)));
                                                /* LDC/STC with writeback */

  Exec_op->to_DAI  := ((LSR_instr || LSM_instr || Transfer_ext) && (!Undefined))
                     || Coprocessor;

  Exec_op->to_SPSR := Undefined || SWI_instr || (Exec_op->flags->MSR && B_bit);

  Exec_op->LD_PC   := Exec_op->to_DAI && L_bit            /* Excludes SWP */
                   && (!Coprocessor)
                   && (if LSM_instr then ((Op_code and 16#8000) != 0)
                       else Rd15);                     /* Destination R15 */

  Exec_op->dest_PC := Arith_type && Rd15 && (!Arith_ext) && (!Undefined);
/* Assumes no CMP destinations of Rd = 15.  Includes MSR instructions.    */

  DAI_op->read     := L_bit || SWP_instr;                     /* L or SWP */
  DAI_op->write    := !L_bit;

  Exec_op->store   := Exec_op->to_DAI && DAI_op->write;
                                        /* Includes STC, MCR (& some CDP) */

  Exec_op->addr_pre_inc := LSM_instr && (P_bit = U_bit);

  DAI_op->get_addr   := true;  /* Get address unless otherwise instructed */
  DAI_op->force_user := LSR_instr && (!P_bit) && W_bit;
  DAI_op->pre_inc    := LSM_instr && (P_bit = U_bit);
  DAI_op->LSM        := LSM_instr;

  DAI_op->copro->use_copro := Coprocessor;
  DAI_op->copro->copro_no  := (Op_code>>8) and 16#F;
  DAI_op->copro->instr     := Op_code and 16#03100010;
/*                                              0x 0  x => STC            */
/*                                              0x 1  x => LDC            */
/*                                              10 x  0 => CDP            */
/*                                              10 0  1 => MCR            */
/*                                              10 1  1 => MRC            */

  DAI_op->size := (if (LSM_instr || Coprocessor
                   || ((!B_bit) && (SWP_instr || LSR_instr))) then 4
                   else
                     (if (Transfer_ext && (!SWP_instr)
                      && ((Op_code and 16#00000020) != 0)) then 2
                      else 1));

  DAI_op->sex  := Transfer_ext && ((Op_code and 16#00000040) != 0);
                                                /* SWPs never sign extend */

  Exec_op->operation->ALU_op := (if Exec_op->to_DAI then/*SWP doesn't care*/
                                   ((Op_code >> 23) and 1) or 2#1010
                                               /* CMP/CMN (sub/add) codes */
                                 else
                                   (
                                   if (Exec_op->operation->use_MUL) then
                                     2#0100                        /* ADD */
                                   else
                                     (
                                     if Exec_op->branch->BX then  /* 1101 */
                                       ((Op_code >> 21) and 16#F) or 2#0100
                                     else (Op_code >> 21) and 16#F
                                     )
                                   )
                                 );

  Exec_op->Exception  := (if Undefined then 4
                          else (if SWI_instr then 8 else 0));

  if SWI_instr then                   /* SWI   NB else _default_ to input */
    (
    Exec_op->PSR_ctrl->mode   := 3;
    Exec_op->PSR_ctrl->IRQ_En := false
    );
                   
  if Undefined then                 /* Undef   NB else _default_ to input */
    (
    Exec_op->PSR_ctrl->mode   := 11;
    Exec_op->PSR_ctrl->IRQ_En := false
    )
                   
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  Rn15 := (Op_code and 16#000F0000) = 16#000F0000;
  Rd15 := (Op_code and 16#0000F000) = 16#0000F000;       /* Frequent test */

  if ((Extension and 16#30) = 0) then DEC_normal else DEC_ext
                                                        /* Derive outputs */

  )
).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

DEC_Q_assign(Op_code   : val(int),
             Mode      : val(int),
             Xfer_mode : val(int),
             Link_class: val(bool),
             BL1       : val(bool),
             Exec_op   : var(Decoded_instr_type),
             DAI_op    : var(DAI_instr_type),
             QTout     : var(chan(void))):
             expr(void) =
  (

  if (Exec_op->to_DAI) then
    (                         /* M_reg calculation may be _unconditional_ */
    M_reg : var(int).                       /* Register number for memory */

/*  if ((store(Op_code)   /* Dummy placeholder on all stores (except SWP) */*/
    if ((((IR->Op_code and 16#00100000) = 0)
     && ((Op_code and 16#0FB00FF0) != 16#01000090))
    || ((Op_code and 16#0E000000) = 16#0C000000)                   /* LSC */
    || ((Op_code and 16#0E000010) = 16#0E000000)) then M_reg := 15
    else              /* CDP (plus some SWI bits which are never present) */
      (
      if DAI_op->LSM then M_reg := Rd_gen(DAI_op->last_cycle)
           /* NB Returns last cycle indicator - only invoked for LDM here */
      else                M_reg := (Op_code >> 12) and 15
      );

    DAI_op->queue_addr := Dec_allocate(QTout, M_reg, Xfer_mode, true)
    )
  else
    DAI_op->queue_addr := -1;             /* Out of range - error if used */

  if (Exec_op->to_queue) then
    (
    Q_reg : var(int).                        /* Register number for queue */
    Q_mode : var(int).

    if ((!Link_class) || ((Op_code and 16#0F000000) = 16#0B000000) || BL1) then
      Q_mode := Mode                                         /* ... or BL */
                                         /* Also include predicted BL (?) */
    else Q_mode := Exec_op->PSR_ctrl->mode;       /* Save R14 in new mode */

    if Link_class then Q_reg := 14
    else
      (
      if (Exec_op->to_DAI            /* Writeback or short multiplication */
     || (Exec_op->operation->use_MUL && ((Op_code and 16#00800000) = 0))) then
        Q_reg := (Op_code >> 16) and 15
      else
        Q_reg := (Op_code >> 12) and 15
      );

    Exec_op->queue_addr := Dec_allocate(QTout, Q_reg, Q_mode, false)
    )
  else
    Exec_op->queue_addr := -1             /* Out of range - error if used */

  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  DEC_Q_addr := Queue_size - 1;         /* Initialisation (pre-inc. to 0) */
  Q_Fcol     := 0;                   /* Expect "previous" colour at start */
  C_flow_col := 0;
  (i : var(int).
  forseq i in 0 to 9 do (Queue_CAM[i])->addr := 15          /* Invalidate */
  );


  forever
    (
    Instruction?(IR:=(?Instruction));

    active_block("decode",IR->bv_colour,(

    wait_for(DEC_delay);

    to_ALU->Op_code   := IR->Op_code;/* For disassembler only (eventually)*/
    to_ALU->extension := IR->extension;          /* For disassembler only */
    to_ALU->PSR_ctrl  := IR->PSR_ctrl;
    to_ALU->colour    := IR->colour;
    to_ALU->PC        := IR->PC;
    to_ALU->bv_colour := IR->bv_colour;
    DAI_op->bv_colour := IR->bv_colour;

             /* Make next test "true" to disable counterflow colour check */
    if ((IR->colour = C_flow_col)     /* Always lets copro. stuff through */
    || ((IR->Op_code and 16#0C000000) = 16#0C000000)) then
      (                                         /* Begin decode and issue */
      Rn : var(int).
      Rd : var(int).
      Rs : var(int).
      Rm : var(int).

      Ra : var(int).
      Rb : var(int).
      Rc : var(int).

      Xfer_mode : var(int).


      Rn := (IR->Op_code >> 16) and 15;  /* Diverges from ARM name in MUL */
      Rd := (IR->Op_code >> 12) and 15;  /* Diverges from ARM name in MUL */
      Rs := (IR->Op_code >>  8) and 15;
      Rm := (IR->Op_code)       and 15;

      Rd_gen_temp := IR->Op_code and 16#0000FFFF;/* Register mask for LSM */

                                              /* Set up "global" register */
      if (IR->PSR_ctrl->Thumb) then
        (
        Register_bank_PC := IR->PC + 4;
                                    /* Bodge to align   LDR Rd, [PC, #nn] */
        if ((IR->Op_code and 16#0C000000) = 16#04000000) then      /* LSR */
          Register_bank_PC := Register_bank_PC and 16#FFFFFFFC
        )
      else
        Register_bank_PC := IR->PC + 8;

      DAI_op->last_cycle := true;   /* Default state - altered by "Rdgen" */


      if (((IR->Op_code and 16#0E508000) = 16#08500000)            /* LDM */
       || ((IR->Op_code and 16#0E500000) = 16#08400000)) then      /* STM */
        Xfer_mode := 0                                 /* Force user mode */
      else
        Xfer_mode := IR->PSR_ctrl->mode;

                           /* Flag which registers (& immediate) required */
      Register_usage(IR->Op_code, to_ALU->reg_usage->uses_A,
                                  to_ALU->reg_usage->uses_B,
                                  to_ALU->reg_usage->uses_C,
                                  to_ALU->reg_usage->uses_Imm);

      if ((IR->Op_code and 16#0F8000F0) = 16#00000090) then Ra := Rd /*MUL*/
      else Ra := Rn;                   /* (including long multiplication) */
      Rb := Rm;                                      /* Always this field */
      if ((IR->Op_code and 16#0FB00FF0) = 16#01000090) then Rc := Rm /*SWP*/
      else
/*
        (
        if (store(IR->Op_code)) then
          (
          if ((IR->Op_code and 16#0E100000) = 16#08000000) then    /* STM */
            Rc := Rd_gen(DAI_op->last_cycle)        /*returns "last_cycle"*/
          else
            Rc := Rd                             /* SWPs already excluded */
          )
        else
          Rc := Rs
        );
*/
        (
        if ((IR->Op_code and 16#0E100000) = 16#08000000) then      /* STM */
          Rc := Rd_gen(DAI_op->last_cycle)        /* returns "last_cycle" */
        else
          (
          if (to_ALU->reg_usage->uses_C) then Rc := Rs
          else Rc := Rd                          /* SWPs already excluded */
          )
        );

/*
printstr("Instruction at "); printhex(IR->PC);
if (IR->PSR_ctrl->Thumb) then printstr(" Thumb ") else printstr(" ARM ");
printstr(" value "); printhex(IR->Op_code);
printstr("   PC = "); printhex(Register_bank_PC);
printstr("\n"); 
*/
      to_ALU->Immediate := Immediate(IR->Op_code, IR->LSB, IR->special);

      BL1 := IR->special && ((IR->Op_code and 16#00001000) != 0);
                             /* Uses Destination LSB to determine BL0/BL1 */

/*   Is actually passed back from DEC - fix ????
      Link_class  := ((Op_code and 16#0B000000)  = 16#0B000000)/*BL/SWI*/
                  || Undef_fn(Op_code, Exec_op)
                  || BL1;
*/

/* Perform decode (mostly output routeing)                                */
      DEC(IR->Op_code, BL1,    IR->extension,
          to_ALU,      DAI_op,
          Link_class);

        (                                                  /* First cycle */

/* Read register bank */
          (
          A : var(Reg_addr_type).
          if (to_ALU->reg_usage->uses_A) then
            (A->addr := Ra; A->mode := IR->PSR_ctrl->mode; Areg_addr!A)
          )
        |
          (
          B : var(Reg_addr_type).
          if (to_ALU->reg_usage->uses_B) then
            (B->addr := Rb; B->mode := IR->PSR_ctrl->mode; Breg_addr!B)
          )
        |
          (
          C : var(Reg_addr_type).
          C->addr := Rc;
          if to_ALU->store then               /* Definitive store command */
            (C->mode := Xfer_mode; Store_reg_addr!C)
          else
            (if (to_ALU->reg_usage->uses_C) then
              (C->mode := IR->PSR_ctrl->mode; Creg_addr!C))
          )

        |

          (

/* Parallel CAM look up and forward requests (*3) */
          (
          if (to_ALU->reg_usage->uses_A) then
            Queue_CAM_lookup(Ra, IR->PSR_ctrl->mode, IR->colour, QA_req) |
          if (to_ALU->reg_usage->uses_B) then
            Queue_CAM_lookup(Rb, IR->PSR_ctrl->mode, IR->colour, QB_req) |
          if to_ALU->store then               /* Definitive store command */
            Queue_CAM_lookup(Rc, Xfer_mode, IR->colour, QStore_req)
          else
            (
            if (to_ALU->reg_usage->uses_C) then
              Queue_CAM_lookup(Rc, IR->PSR_ctrl->mode, IR->colour, QC_req)
            )
          );

                             /* TIMING MODEL NEEDS ATTENTION IN THIS AREA */

                                      /* Assign any required queue spaces */
          DEC_Q_assign(IR->Op_code, IR->PSR_ctrl->mode, Xfer_mode,
                       Link_class, BL1,
                       to_ALU, DAI_op, QTout);

          (                       /* Parallel output (conditional to DAI) */
          Instr_out!to_ALU
          |
          C_flow?(C_flow_col:=(?C_flow))            /* Counterflow colour */
          |
          if (to_ALU->to_DAI) then to_DAI!DAI_op
          )

          )
        );                                                 /* First cycle */



/* Long multiplication? */
/*    if ((operation->MUL and 7) != 0) then
        read/forward Rn onto A bus  *** ONLY IF DIFFERENT FROM Rd ***
        allocate Rn
        and keep instruction
No second read for LMUL - if Rd=Rn then convert to pseudo LMUL?
*/


      if (to_ALU->flags->LSM) then
        (                                         /* Start of LSM repeats */
        if (LSM_go?(?LSM_go)) then                    /* Check for LSM go */
          (
          DAI_op->get_addr := false;        /* Addresses generated by DAI */
                                           /* Operation remains unchanged */
          while (!(DAI_op->last_cycle)) do
            (
            wait_for(DEC_multi_delay);

/*          if store(IR->Op_code) then                                    */
            if ((IR->Op_code and 16#00100000) = 0) then
              (                                                    /* STM */
              C : var(Reg_addr_type).
              C->addr := Rd_gen(DAI_op->last_cycle);
              C->mode := Xfer_mode;

              (
                Store_reg_addr!C
              |
                (
                Queue_CAM_lookup(C->addr, Xfer_mode, IR->colour, QStore_req);
                DAI_op->queue_addr := Dec_allocate(QTout, 15,
                                                          Xfer_mode, true);
                to_DAI!DAI_op
                )
              )
              )
            else
              (                                                    /* LDM */
              DAI_op->queue_addr := Dec_allocate(QTout,
                         Rd_gen(DAI_op->last_cycle), Xfer_mode, true);

              to_DAI!DAI_op
              )
            )                                           /* End of "while" */
          )
        )                                           /* End of LSM repeats */
      )                                        /* End of decode and issue */
    else
      (
/*
      printstr("DEC: colour mismatch: operation "); printhex(IR->Op_code);
      printstr("\n");
*/
      wait_for(DEC_trash_delay)              /* Instruction discard delay */
      )

    )
  
    )) /* end of active_block */
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
