/*------------------------------------------------------------------------*/
/*                                                                        */
/*                   AMULET3 Instruction Decoder - part 2                 */
/*                             Substitute version                         */
/*                                                                        */
/*                                             Created: JDG  26/3/97      */
/*                                       Last Modified: JDG  27/3/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/*  Check that ALL output bits defined for every instruction category     */
/*                                                                        */
/*------------------------------------------------------------------------*/

(Decoder(Instruction : var(chan(Instr_pkt_type2)),
           Instr_out : var(chan(Decoded_instr_type)),
              C_flow : var(chan(int)),
              LSM_go : var(chan(bool)),
           Areg_addr : var(chan(Reg_addr_type)),
           Breg_addr : var(chan(Reg_addr_type)),
           Creg_addr : var(chan(Reg_addr_type)),
      Store_reg_addr : var(chan(Reg_addr_type)),
              QA_req : var(chan(Forward_req_type)),
              QB_req : var(chan(Forward_req_type)),
              QC_req : var(chan(Forward_req_type)),
          QStore_req : var(chan(Forward_req_type)),
              to_DAI : var(chan(DAI_instr_type)),
               QTout : var(chan(void))):
             expr(void)=
  (
  IR     : var(Instr_pkt_type2).
  to_ALU : var(Decoded_instr_type).
  DAI_op : var(DAI_instr_type).
  C_flow_col : var(int).

  default_C : var(bool).                                /* Ignore "Imm_C" */
  Imm_C     : var(int).                 /* Carry from immediate extractor */

  Link_class   : var(bool).                /* Anything which saves an R14 */
  BL1          : var(bool).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Decode bit crunching                                                   */

(DEC(Op_code   : val(int),
     BL1       : val(bool),
     Extension : val(int),
     Exec_op   : var(Decoded_instr_type),
     DAI_op    : var(DAI_instr_type),
     Link_class: var(bool)):               /* Anything which saves an R14 */
     expr(void) =
  (
  Operation : var(int).
  Rn15 : var(bool).
  Rd15 : var(bool).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Decode instruction with extension                                      */

DEC_ext : expr(void) =
  (
  Instr_ext    : var(bool).                             /* FIQ, IRQ, Pabt */

  Link_class := false;        /* TEMP - while passed back */


  Instr_ext    := (Extension < 32);                     /* FIQ, IRQ, Pabt */

  Exec_op->condition  := (if Instr_ext then 16#E   /* Exception -> always */
                          else Extension and 16#F); /* Predicted branches */

  Exec_op->Exception  := (if (Extension = 4) then 12 else 0);/* Pabt/none */

  Exec_op->flags->LSM := false;
  Exec_op->flags->MRS := false;
  Exec_op->flags->MSR := false;

  Exec_op->operation->shift     := false;
  Exec_op->operation->shift_op  := 0;                       /* Don't care */
  Exec_op->operation->shift_imm := false;                   /* Don't care */
  Exec_op->operation->shift_amount := 0;                    /* Don't care */
  Exec_op->operation->use_MUL   := false;                           /* No */
  Exec_op->operation->use_ALU   := false;
  Exec_op->operation->set_flags := false;

  Exec_op->branch->branch    := Exec_op->branch->predicted;
  Exec_op->branch->predicted := (Extension >= 32);
  Exec_op->branch->link      := Instr_ext               /* FIQ, IRQ, Pabt */
                             || (Exec_op->branch->predicted
                               && ((Extension and 16#10) != 0));
  Exec_op->branch->BX        := false;
  Exec_op->branch->Thumb_BL  := false;

  Exec_op->to_queue  := Exec_op->branch->link;
  Exec_op->to_DAI    := false;
  Exec_op->to_SPSR   := Instr_ext;
  Exec_op->LD_PC     := false;
  Exec_op->dest_PC   := false;
  DAI_op->read       := false;                              /* Don't care */
  DAI_op->write      := false;                              /* Don't care */
  Exec_op->store     := Exec_op->to_DAI && DAI_op->write;

  DAI_op->get_addr   := true;                               /* Don't care */
  DAI_op->force_user := false;                              /* Don't care */
  DAI_op->pre_inc    := false;                              /* Don't care */
  DAI_op->LSM        := Exec_op->flags->LSM;
  DAI_op->size       := 4;                                  /* Don't care */
  DAI_op->sex        := false;                              /* Don't care */

  Exec_op->operation->ALU_op := 0;                          /* Don't care */

  Exec_op->PSR_ctrl->mode := (case Extension
                                when 2 =>  1                  /* FIQ mode */
                                when 3 =>  2                  /* IRQ mode */
                                when 4 =>  7            /* Prefetch abort */
                                default Exec_op->PSR_ctrl->mode
                             );

  if (((Extension > 4) && (Extension < 32)) || (Extension > 63)) then
    (printint(Extension); printstr(" is an illegal extension");
     Run := false);

  if Exec_op->branch->predicted then
    (
    printstr("Predicted branch\n");
    null
    )

  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Decode instruction without extension                                   */
/* "Don't care" cases are NOT set in an instruction                       */

DEC_normal : expr(void) =
  (
  P_bit    : var(bool).
  U_bit    : var(bool).
  B_bit    : var(bool).
  W_bit    : var(bool).
  L_bit    : var(bool).
  bits_6_5 : var(bool).

DEC_undef : expr(void) =
  (
  Exec_op->operation->use_ALU := false;
  Exec_op->branch->link   := true;
  Exec_op->to_queue       := true;
  Exec_op->to_SPSR        := true;
  Exec_op->PSR_ctrl->mode := 11;
  Exec_op->Exception      :=  4
  ).

DEC_arith : expr(void) =
  (
  if ((Op_code and 16#0F0000F0)  = 16#00000090) then
    (                                             /* Arithmetic extension */
    if ((Op_code and 16#00C00000) != 16#00400000) then
      (                                                 /* Multiplication */
      Exec_op->operation->use_MUL   := true;
      Exec_op->operation->set_flags := L_bit;            /* L_bit = S bit */
      Exec_op->to_queue             := !Rd15;
      Exec_op->operation->ALU_op    := 2#0100                      /* ADD */
      )
    else DEC_undef
    )
  else
    (
    if (((Op_code and 16#0D900000)  = 16#01000000)
     && ((Op_code and 16#02000090) != 16#00000090)) then
      (                                              /* Control extension */
      if ((Op_code and 16#026FFFF0) = 16#002FFF10) then
        (
        Exec_op->branch->BX := true;
        Exec_op->operation->ALU_op := ((Op_code >> 21) and 16#F) or 2#0100
        )
      else
        (
        Exec_op->flags->MRS := ((Op_code and 16#022F0FFF) = 16#000F0000);
        Exec_op->flags->MSR := ((Op_code and 16#0220FFF0) = 16#0020F000)
                            || ((Op_code and 16#0220F000) = 16#0220F000);
        Exec_op->to_SPSR    := Exec_op->flags->MSR && B_bit;
        Exec_op->to_queue   := !Rd15
        )
      )
    else
      (
      if (((Op_code and 16#0E000090)  = 16#00000090)
       && ((Op_code and 16#01000060) != 16#00000000)) then
        (                                         /* Load/store extension */

/* filter undefineds(?) */

        SWP_instr : var(bool).

        SWP_instr := ((Op_code and 16#01B00FF0)  = 16#01000090);

        Exec_op->to_queue  := (!Rd15) && (W_bit || (!P_bit));/* NB no SWP */
        Exec_op->to_DAI    := true;
        Exec_op->operation->ALU_op := ((Op_code >> 23) and 1) or 2#1010;
                                               /* CMP/CMN (sub/add) codes */

        DAI_op->read     := L_bit || SWP_instr;               /* L or SWP */
        DAI_op->write    := !L_bit;
        Exec_op->store   := Exec_op->to_DAI && DAI_op->write;
        DAI_op->get_addr := true;/*Get address unless otherwise instructed*/

        DAI_op->size := (if ((!B_bit) && SWP_instr) then 4
                         else
                         (if ((!SWP_instr) && ((Op_code and 16#00000020) != 0)) then 2
                          else 1));
        DAI_op->sex  := ((Op_code and 16#00000040) != 0)
                                                /* SWPs never sign extend */
        )
      else
        (
        Exec_op->to_queue                             /* and exclude CMPs */
                     := (!Rd15) && ((Op_code and 16#01900000) != 16#01100000);

        Exec_op->operation->shift := shift(Op_code); /* Separate function */
        Exec_op->operation->shift_op :=
          (if ((Op_code and 16#00000FF0) = 16#00000060) then 4     /* RRX */
           else (Op_code >> 5) and 3);
        Exec_op->operation->shift_imm    := ((Op_code and 16#010) = 0);
        Exec_op->operation->shift_amount :=
          (if ((Op_code and 16#F80) = 0) then 32 else (Op_code >> 7) and 16#1F);
                               /* LSL #0 bypasses, RRX assumes amount = 1 */

        Exec_op->operation->set_flags := L_bit           /* L_bit = S bit */
                                && (!(((Op_code and 16#0E000090)  = 16#00000090)
                                   && bits_6_5));   /* Filter undefineds? */

        Exec_op->operation->ALU_op := (Op_code >> 21) and 16#F;

        Exec_op->dest_PC := Rd15
        )
      )
    )
  ).

DEC_LSR : expr(void) =
  (
  if ((Op_code and 16#0E000010) = 16#06000010) then DEC_undef/* Undefined */
  else
    (
    Exec_op->to_queue   := (!Rn15) && (W_bit || (!P_bit));       /* LSR ! */
    Exec_op->LD_PC      := L_bit && Rd15;
    Exec_op->store      := !L_bit;
    Exec_op->operation->ALU_op := ((Op_code >> 23) and 1) or 2#1010;
                                               /* CMP/CMN (sub/add) codes */
    Exec_op->to_DAI     := true;

    DAI_op->read        := L_bit;
    DAI_op->write       := !L_bit;
    DAI_op->size        := (if B_bit then 1 else 4);
    DAI_op->sex         := false;
    DAI_op->get_addr    := true;/* Get address unless otherwise instructed*/
    DAI_op->force_user  := (!P_bit) && W_bit;
    DAI_op->pre_inc     := false;
    DAI_op->LSM         := false
    )
  ).

DEC_LSM : expr(void) =
  (
  Exec_op->flags->LSM := true;
  Exec_op->to_queue   := (!Rn15) && W_bit;
  Exec_op->LD_PC      := L_bit && ((Op_code and 16#00008000) != 0);
  Exec_op->store      := !L_bit;
  Exec_op->operation->ALU_op := ((Op_code >> 23) and 1) or 2#1010;
                                               /* CMP/CMN (sub/add) codes */
  Exec_op->to_DAI     := true;

  DAI_op->read        := L_bit;
  DAI_op->write       := !L_bit;
  DAI_op->size        := 4;
  DAI_op->sex         := false;
  DAI_op->get_addr    := true; /* Get address unless otherwise instructed */
  DAI_op->force_user  := false;
  DAI_op->LSM         := true;
  DAI_op->pre_inc     := (P_bit = U_bit)
  ).

DEC_BRA : expr(void) =
  (
  Exec_op->operation->use_ALU := false;
  Exec_op->branch->branch := true;
  if ((Op_code and 16#01000000) != 0) then
    (
    Exec_op->branch->link   := ((Op_code and 16#01000000) != 0);
    Exec_op->to_queue       := true
    )
  ).

DEC_LSC : expr(void) = (DEC_undef).                          /* Undefined */

DEC_SWI : expr(void) =
  (
  if ((Op_code and 16#01000000) = 0) then DEC_undef
  else
    (                                                              /* SWI */
    Exec_op->operation->use_ALU := false;
    Exec_op->branch->link   := true;
    Exec_op->to_queue       := true;
    Exec_op->to_SPSR        := true;
    Exec_op->PSR_ctrl->mode :=  3;
    Exec_op->Exception      :=  8
    )
  ).

/*DAI_op->queue_addr := 10;    /* Magic statement to make LDB work - no idea why */*/

  P_bit    := ((Op_code and 16#01000000) != 0);
  U_bit    := ((Op_code and 16#00800000) != 0);
  B_bit    := ((Op_code and 16#00400000) != 0);
  W_bit    := ((Op_code and 16#00200000) != 0);
  L_bit    := ((Op_code and 16#00100000) != 0);
  bits_6_5 := ((Op_code and 16#00000060) != 0);

  Exec_op->condition  := (Op_code >> 28) and 16#F;
  Exec_op->Exception  := 0;

  Exec_op->operation->shift     := false;
  Exec_op->operation->use_MUL   := false;
  Exec_op->operation->use_ALU   := true;          /* Default to operation */
  Exec_op->operation->set_flags := false;

  Exec_op->dest_PC   := false;

  Exec_op->LD_PC     := false;
  Exec_op->store     := false;
  Exec_op->to_queue  := false;

  Exec_op->branch->branch    := false;
  Exec_op->branch->predicted := false;
  Exec_op->branch->link      := false;
  Exec_op->branch->BX        := false;
  Exec_op->branch->Thumb_BL  := false;

  Exec_op->to_DAI     := false;
  Exec_op->to_SPSR    := false;

  Exec_op->flags->LSM := false;
  Exec_op->flags->MRS := false;
  Exec_op->flags->MSR := false;

  (case (Op_code and 16#0E000000)
    when 16#00000000 => DEC_arith
    when 16#02000000 => DEC_arith
    when 16#04000000 => DEC_LSR
    when 16#06000000 => DEC_LSR
    when 16#08000000 => DEC_LSM
    when 16#0A000000 => DEC_BRA
    when 16#0C000000 => DEC_LSC
    when 16#0E000000 => DEC_SWI
  );

  Link_class := Exec_op->branch->link

  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  Rn15 := (Op_code and 16#000F0000) = 16#000F0000;
  Rd15 := (Op_code and 16#0000F000) = 16#0000F000;       /* Frequent test */

  Exec_op->branch->Thumb_BL := BL1;

  if (Extension < 2) then DEC_normal else DEC_ext;      /* Derive outputs */

  if BL1 then Exec_op->branch->link := true

  )
).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

DEC_Q_assign(Op_code   : val(int),
             Mode      : val(int),
             Xfer_mode : val(int),
             Link_class: val(bool),
             BL1       : val(bool),
             Exec_op   : var(Decoded_instr_type),
             DAI_op    : var(DAI_instr_type),
             QTout     : var(chan(void))):
             expr(void) =
  (

  if (Exec_op->to_DAI) then
    (                         /* M_reg calculation may be _unconditional_ */
    M_reg : var(int).                       /* Register number for memory */

    if store(Op_code) then M_reg := 15 /* Dummy placeholder on all stores */
    else
      (
      if DAI_op->LSM then M_reg := Rd_gen(DAI_op->last_cycle)
           /* NB Returns last cycle indicator - only invoked for LDM here */
      else                M_reg := (Op_code >> 12) and 15
      );

    DAI_op->queue_addr := Dec_allocate(QTout, M_reg, Xfer_mode, true)
    )
  else
    DAI_op->queue_addr := -1;             /* Out of range - error if used */

  if (Exec_op->to_queue) then
    (
    Q_reg : var(int).                        /* Register number for queue */
    Q_mode : var(int).

    if ((!Link_class) || ((Op_code and 16#0F000000) = 16#0B000000) || BL1) then
      Q_mode := Mode                                         /* ... or BL */
    else Q_mode := Exec_op->PSR_ctrl->mode;       /* Save R14 in new mode */

    if Link_class then Q_reg := 14
    else
      (
      if (Exec_op->to_DAI            /* Writeback or short multiplication */
     || (Exec_op->operation->use_MUL && ((Op_code and 16#00800000) = 0))) then
        Q_reg := (Op_code >> 16) and 15
      else
        Q_reg := (Op_code >> 12) and 15
      );

    Exec_op->queue_addr := Dec_allocate(QTout, Q_reg, Q_mode, false)
    )
  else
    Exec_op->queue_addr := -1             /* Out of range - error if used */

  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  DEC_Q_addr := Queue_size - 1;         /* Initialisation (pre-inc. to 0) */
  Q_Fcol     := 0;                   /* Expect "previous" colour at start */
  C_flow_col := 0;
  (i : var(int).
  forseq i in 0 to 9 do (Queue_CAM[i])->addr := 15          /* Invalidate */
  );


  forever
    (
    Instruction?(IR:=(?Instruction));

    wait_for(DEC_delay);

    to_ALU->Op_code   := IR->Op_code;/* For disassembler only (eventually)*/
    to_ALU->extension := IR->extension;          /* For disassembler only */
    to_ALU->PSR_ctrl  := IR->PSR_ctrl;
    to_ALU->colour    := IR->colour;
    to_ALU->PC        := IR->PC;

                  /* Remove next test to disable counterflow colour check */
    if (IR->colour = C_flow_col) then
      (                                         /* Begin decode and issue */
      Rn : var(int).
      Rd : var(int).
      Rs : var(int).
      Rm : var(int).

      Ra : var(int).
      Rb : var(int).
      Rc : var(int).

      Xfer_mode : var(int).


      Rn := (IR->Op_code >> 16) and 15;  /* Diverges from ARM name in MUL */
      Rd := (IR->Op_code >> 12) and 15;  /* Diverges from ARM name in MUL */
      Rs := (IR->Op_code >>  8) and 15;
      Rm := (IR->Op_code)       and 15;

      Rd_gen_temp := IR->Op_code and 16#0000FFFF;/* Register mask for LSM */

                                              /* Set up "global" register */
      if (IR->PSR_ctrl->Thumb) then Register_bank_PC := IR->PC + 4
      else                          Register_bank_PC := IR->PC + 8;

      DAI_op->last_cycle := true;   /* Default state - altered by "Rdgen" */


      if (((IR->Op_code and 16#0E508000) = 16#08500000)            /* LDM */
       || ((IR->Op_code and 16#0E500000) = 16#08400000)) then      /* STM */
        Xfer_mode := 0                                 /* Force user mode */
      else
        Xfer_mode := IR->PSR_ctrl->mode;

      if ((IR->Op_code and 16#0F8000F0) = 16#00000090) then Ra := Rd /*MUL*/
      else Ra := Rn;                   /* (including long multiplication) */
      Rb := Rm;                                      /* Always this field */
      if ((IR->Op_code and 16#0FB00FF0) = 16#01000090) then Rc := Rm /*SWP*/
      else
        (
        if (store(IR->Op_code)) then
          (
          if ((IR->Op_code and 16#0E100000) = 16#08000000) then    /* STM */
            Rc := Rd_gen(DAI_op->last_cycle)        /*returns "last_cycle"*/
          else
            Rc := Rd                             /* SWPs already excluded */
          )
        else
          Rc := Rs
        );

                           /* Flag which registers (& immediate) required */
      Register_usage(IR->Op_code, to_ALU->reg_usage->uses_A,
                                  to_ALU->reg_usage->uses_B,
                                  to_ALU->reg_usage->uses_C,
                                  to_ALU->reg_usage->uses_Imm);
/*
printstr("Instruction at "); printhex(IR->PC);
if (IR->PSR_ctrl->Thumb) then printstr(" Thumb ") else printstr(" ARM ");
printstr(" value "); printhex(IR->Op_code);
printstr("   PC = "); printhex(Register_bank_PC);
printstr("\n"); 
*/
      to_ALU->Immediate := Immediate(IR->Op_code, IR->LSB, IR->special);

      BL1 := IR->special && ((IR->Op_code and 16#00001000) != 0);
                             /* Uses Destination LSB to determine BL0/BL1 */

/*   Is actually passed back from DEC - fix !!!!
      Link_class       := ((Op_code and 16#0B000000)  = 16#0B000000)/*BL/SWI*/
                       || Undefined || BL1;
*/

/* Perform decode (mostly output routeing)                                */
      DEC(IR->Op_code,        BL1,    IR->extension,
          to_ALU,             DAI_op,
          Link_class);

        (                                                  /* First cycle */

/* Read register bank */
          (
          A : var(Reg_addr_type).
          if (to_ALU->reg_usage->uses_A) then
            (A->addr := Ra; A->mode := IR->PSR_ctrl->mode; Areg_addr!A)
          )
        |
          (
          B : var(Reg_addr_type).
          if (to_ALU->reg_usage->uses_B) then
            (B->addr := Rb; B->mode := IR->PSR_ctrl->mode; Breg_addr!B)
          )
        |
          (
          C : var(Reg_addr_type).
          C->addr := Rc;
          if to_ALU->store then               /* Definitive store command */
            (C->mode := Xfer_mode; Store_reg_addr!C)
          else
            (if (to_ALU->reg_usage->uses_C) then
              (C->mode := IR->PSR_ctrl->mode; Creg_addr!C))
          )

        |

          (

/* Parallel CAM look up and forward requests (*3) */
          (
          if (to_ALU->reg_usage->uses_A) then
            Queue_CAM_lookup(Ra, IR->PSR_ctrl->mode, QA_req) |
          if (to_ALU->reg_usage->uses_B) then
            Queue_CAM_lookup(Rb, IR->PSR_ctrl->mode, QB_req) |
          if to_ALU->store then               /* Definitive store command */
            Queue_CAM_lookup(Rc, Xfer_mode, QStore_req)
          else
            (
            if (to_ALU->reg_usage->uses_C) then
              Queue_CAM_lookup(Rc, IR->PSR_ctrl->mode, QC_req)
            )
          );

                             /* TIMING MODEL NEEDS ATTENTION IN THIS AREA */

                                      /* Assign any required queue spaces */
          DEC_Q_assign(IR->Op_code, IR->PSR_ctrl->mode, Xfer_mode,
                       Link_class, BL1,
                       to_ALU, DAI_op, QTout);

          (                       /* Parallel output (conditional to DAI) */
          Instr_out!to_ALU
          |
          C_flow?(C_flow_col:=(?C_flow))            /* Counterflow colour */
          |
          if (to_ALU->to_DAI) then to_DAI!DAI_op
          )

          )
        );                                                 /* First cycle */



/* Long multiplication? */
/*    if ((operation->MUL and 7) != 0) then
        read/forward Rn onto A bus  *** ONLY IF DIFFERENT FROM Rd ***
        allocate Rn
        and keep instruction
No second read for LMUL - if Rd=Rn then convert to pseudo LMUL?
*/


      if (to_ALU->flags->LSM) then
        (                                         /* Start of LSM repeats */
        if (LSM_go?(?LSM_go)) then                    /* Check for LSM go */
          (
          DAI_op->get_addr := false;        /* Addresses generated by DAI */
                                           /* Operation remains unchanged */
          while (!(DAI_op->last_cycle)) do
            (
            wait_for(DEC_multi_delay);

            if store(IR->Op_code) then
              (                                                    /* STM */
              C : var(Reg_addr_type).
              C->addr := Rd_gen(DAI_op->last_cycle);
              C->mode := Xfer_mode;

              (
                Store_reg_addr!C
              |
                (
                Queue_CAM_lookup(C->addr, Xfer_mode, QStore_req);
                DAI_op->queue_addr := Dec_allocate(QTout, 15,
                                                          Xfer_mode, true);
                to_DAI!DAI_op
                )
              )
              )
            else
              (                                                    /* LDM */
              DAI_op->queue_addr := Dec_allocate(QTout,
                         Rd_gen(DAI_op->last_cycle), Xfer_mode, true);

              to_DAI!DAI_op
              )
            )                                           /* End of "while" */
          )
        )                                           /* End of LSM repeats */
      )                                        /* End of decode and issue */
    else
      (
/*
      printstr("DEC: colour mismatch: operation "); printhex(IR->Op_code);
      printstr("\n");
*/
      wait_for(DEC_trash_delay)              /* Instruction discard delay */
      )

    )
  )
).

/*------------------------------------------------------------------------*/

(null)

/*------------------------------------------------------------------------*/
