/*------------------------------------------------------------------------*/
/*                                                                        */
/*                             ARM Disassembler                           */
/*                                                                        */
/*                                             Created: JDG  31/1/97      */
/*                                       Last Modified: JDG   4/7/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/*                                                                        */
/*------------------------------------------------------------------------*/

addr_tab(position : var(int)) : expr(void) =
  while (position < 8) do (printstr(" "); position := position + 1).

/*------------------------------------------------------------------------*/

print_hex(x : val(int), digits : val(int)) : expr(void) =
  (
  i : var(int).

  print_digit(x : val(int)) : expr(void) =
    (
    y : var(int).
    y := x and 15;
    if (y < 10) then printint(y) else print(cast(char, y+55))
    ).

  i := (if (digits > 8) then 8 else (if (digits < 1) then 1 else digits));

  while (i > 0) do (i := i - 1; print_digit(x >> (4*i)))
  ).

/*------------------------------------------------------------------------*/

printasc(z : val(int), Thumb : val(bool)) : expr(void) =
  (
  p_char(x : val(int)) : expr(void) =
    (if (((x and 16#FF) > 16#1F) && ((x and 16#FF) < 16#7F)) then ASCII_out(x)
    else printstr(".")).

  p_char(z); p_char(z>>8); if !Thumb then (p_char(z>>16); p_char(z>>24))
  ).

/*------------------------------------------------------------------------*/
/* Print all the disassemble preamble                                     */

print_instr(PC : val(int), op_code : val(int)) : expr(void) =
  (
  printint(now);     printstr(":  "); printhex(PC);              printstr("  ");
  printhex(op_code); printstr("  ");  printasc(op_code, false);  printstr("  ")
  ).

/*------------------------------------------------------------------------*/
/* Print an instruction extension field in the same format as disassembly */

print_instr_ext(PC : val(int)) : expr(void) =
  (printint(now); printstr(":  "); printhex(PC); printstr("  ........  ....  ")).

/*------------------------------------------------------------------------*/

Preg(reg_in : val(int), high_reg : val(bool)) : expr(void) =
  (
  reg : var(int).

  reg := reg_in; if high_reg then reg := reg + 8;

  if (reg < 13) then (printstr("R"); printint(reg))
  else
    (case reg
      when 13 => printstr("SP")
      when 14 => printstr("LR")
      when 15 => printstr("PC")
    )
  ).

/*------------------------------------------------------------------------*/

Preg_list(bits : val(int), LR : val(bool), PC : val(bool)) : expr(void) =
  (
  x : var(int).
  y : var(int).
  range  : var(bool).
  reg_no : var(int).

  x := bits;
  y := 1;
  range  := false;
  reg_no := 0;

  printstr("{");

  while (x != 0) do
    (
    if ((x and y) != 0) then
      (
      x := x and not(y);
      if range then
        (
        if ((x and (y<<1)) = 0) then
          (
          printstr("R"); printint(reg_no);
          range := false;
          if (x != 0) then printstr(",")
          )
        )
      else
        (
        printstr("R"); printint(reg_no);
        if (x != 0) then
          (
          if (((x and (y<<1)) != 0) && ((x and (y<<2)) != 0)) then
            (printstr(".."); range := true)
          else
            printstr(",")
          )
        )
      );
    y := y << 1;
    reg_no := reg_no + 1
    );

  if LR then (if ((bits and 16#FF) != 0) then printstr(","); printstr("LR"));
  if PC then (if ((bits and 16#FF) != 0) then printstr(","); printstr("PC"));

  printstr("}")
  ).

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

Disassemble(PC      : val(int),  op_code : val(int),  extension : val(int),
            Thumb   : val(bool), special : val(bool), LSB       : val(int)):
            expr(void)=
  (

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

condition(CC : val(int), position : var(int)) : expr(void) =
  (
  if (CC != 14) then position := position + 2;
  (case CC
    when  0 => printstr("EQ")
    when  1 => printstr("NE")
    when  2 => printstr("CS")
    when  3 => printstr("CC")
    when  4 => printstr("MI")
    when  5 => printstr("PL")
    when  6 => printstr("VS")
    when  7 => printstr("VC")
    when  8 => printstr("HI")
    when  9 => printstr("LS")
    when 10 => printstr("GE")
    when 11 => printstr("LT")
    when 12 => printstr("GT")
    when 13 => printstr("LE")
    when 14 => null
    when 15 => printstr("NV")
  )
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

Sbit(S : val(bool), position : var(int)) : expr(void) =
  if S then (printstr("S"); position := position + 1).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

data_op(op : val(int), S : val(bool), position : var(int)) : expr(void) =
  (case op
    when  0 => printstr("AND")
    when  1 => printstr("EOR")
    when  2 => printstr("SUB")
    when  3 => printstr("RSB")
    when  4 => printstr("ADD")
    when  5 => printstr("ADC")
    when  6 => printstr("SBC")
    when  7 => printstr("RSC")
    when  8 => printstr("TST")
    when  9 => printstr("TEQ")
    when 10 => printstr("CMP")
    when 11 => printstr("CMN")
    when 12 => printstr("ORR")
    when 13 => printstr("MOV")
    when 14 => printstr("BIC")
    when 15 => printstr("MVN")
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

addr_mode(imm : val(bool), addr : val(int)) : expr(void) =
  (
  if imm then
    (
    x : var(int).
    y : var(int).
    printstr("#");
    x := addr and 16#FF;
    y := (addr >> 7) and 16#1E;
    printhex(((x>>y) and not(16#FFFFFFFF<<(32-y))) or (x<<(32-y)))
    )
  else
    (
    Preg(addr and 15, false);
    if ((addr and 16#FF0) != 0) then
      (

      (print_shift(x : val(int)) : expr(void) =

      if ((addr and 16#010) = 16#000) then      /* LSL #0 case dealt with */
        (printstr("#"); print_hex((((addr >> 7) - 1) and 16#1F) + 1, 2))
      else Preg((addr >> 8) and 15, false)
      ).

      (case (addr and 16#060)
        when 16#000 => (printstr(" LSL "); print_shift(addr))

        when 16#020 => (printstr(" LSR "); print_shift(addr))

        when 16#040 => (printstr(" ASR "); print_shift(addr))

        when 16#060 => if ((addr and 16#F90) = 16#000) then
                         printstr(" RRX #1")
                       else
                         (printstr(" ROR "); print_shift(addr))
      )

      )
    )
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Variables here                                                         */

  position : var(int).
  read_PC  : var(int).                        /* Value read by programmer */
  BL0      : var(bool).
  BL1      : var(bool).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

Arith_etc : expr(void) =
  (
  if (((op_code and 16#0F0000F0) = 16#00000090)
   ||  (((op_code and 16#0D900000)  = 16#01000000)
     && ((op_code and 16#02000090) != 16#00000090))
   ||  (((op_code and 16#0E000090)  = 16#00000090)
     && ((op_code and 16#01000060) != 16#00000000))) then
    (                                                       /* Extensions */
    if ((op_code and 16#0F0000F0) = 16#00000090) then
      (                                                            /* MUL */
      if ((op_code and 16#00C00000) = 16#00000000) then
        (
        if ((op_code and 16#00200000) = 0) then printstr("MUL")
        else                                    printstr("MLA");
        condition((op_code>>28) and 15, position);
        Sbit((op_code and 16#00100000) != 0, position);
        addr_tab(position);
        Preg((op_code>>16) and 15, false); printstr(", ");
        Preg(op_code and 15, false);       printstr(", ");
        Preg((op_code>>8) and 15, false);
        if ((op_code and 16#00200000) != 0) then            /* Accumulate */
          (printstr(", "); Preg((op_code>>12) and 15, false))
        );

      if ((op_code and 16#00C00000) = 16#00400000) then
        (
        printstr("Undef");
        condition((op_code>>28) and 15, position)
        );

      if ((op_code and 16#00800000) = 16#00800000) then
        (
        if ((op_code and 16#00400000) = 0) then printstr("U")
        else                                    printstr("S");
        if ((op_code and 16#00200000) = 0) then printstr("MLAL")
        else                                    printstr("MULL");
        position := position + 2;
        condition((op_code>>28) and 15, position);
        Sbit((op_code and 16#00100000) != 0, position);
        addr_tab(position);
        Preg((op_code>>12) and 15, false); printstr(", ");
        Preg((op_code>>16) and 15, false); printstr(", ");
        Preg(op_code and 15, false);       printstr(", ");
        Preg((op_code>>8) and 15, false)
        )

      );

    if (((op_code and 16#0D900000)  = 16#01000000)
     && ((op_code and 16#02000090) != 16#00000090)) then
      (
      if ((op_code and 16#020F0FFF)  = 16#000F0000) then
        (
        printstr("MRS");
        condition((op_code>>28) and 15, position);
        addr_tab(position);
        Preg((op_code>>12) and 15, false); printstr(", ");
        if ((op_code and 16#00400000)  = 0) then printstr("CPSR")
        else                                     printstr("SPSR")
        )
      else
        (
        if ((op_code and 16#0FFFFFF0)  = 16#012FFF10) then
          (
          printstr("BX"); position := position - 1;
          condition((op_code>>28) and 15, position);
          addr_tab(position);
          Preg(op_code and 15, false)
          )
        else
          (
          if (((op_code and 16#0FB0F000) = 16#0320F000)
           || ((op_code and 16#0FB0FFF0) = 16#0120F000)) then
            (
            printstr("MSR");
            condition((op_code>>28) and 15, position);
            addr_tab(position);
            if ((op_code and 16#00400000) = 0) then printstr("CPSR")
            else                                    printstr("SPSR");
            if ((op_code and 16#000F0000) = 16#000F0000) then
              printstr("_all")
            else
              (
              if ((op_code and 16#00080000) != 0) then printstr("_f");
              if ((op_code and 16#00040000) != 0) then printstr("_s");
              if ((op_code and 16#00020000) != 0) then printstr("_x");
              if ((op_code and 16#00010000) != 0) then printstr("_c")
              );
            printstr(", ");
            addr_mode(((op_code and 16#02000000) != 0), (op_code))
            )
          else
            (
            printstr("Undef"); condition((op_code>>28) and 15, position)
            )
          )
        )
      );

    if (((op_code and 16#0E000090)  = 16#00000090)
     && ((op_code and 16#01000060) != 16#00000000)) then
      (
      if ((op_code and 16#00000060) = 16#00000000) then
        (
        if ((op_code and 16#00B00F00) = 16#01000000) then
          (
          printstr("SWP");
          condition((op_code>>28) and 15, position);
          if ((op_code and 16#00400000) != 0) then
            (printstr("B"); position := position + 1);
          addr_tab(position);
          Preg((op_code>>12) and 15, false); printstr(", ");
          Preg(op_code and 15, false);       printstr(", [");
          Preg((op_code>>16) and 15, false); printstr("]")
          )
        else
          (printstr("Undef"); condition((op_code>>28) and 15, position))
        )
      else
        (
        if (((op_code and 16#00100040) = 16#00000040)
        || (((op_code and 16#00400000)  = 16#00000000)
         && ((op_code and 16#00000F00) != 16#00000000))) then
          (printstr("Undef"); condition((op_code>>28) and 15, position))
                                                /* Signed halfword stores */
        else
          (
          if ((op_code and 16#00100000) != 0) then printstr("LDR")
          else                                     printstr("STR");
          condition((op_code>>28) and 15, position);
          if ((op_code and 16#00000040) != 0) then
            (printstr("S"); position := position + 1);
          if ((op_code and 16#00000020) != 0) then printstr("H")
          else                                     printstr("B");
          position := position + 1;
          addr_tab(position);
          Preg((op_code>>12) and 15, false); printstr(", [");


          if ((op_code and 16#0F6F0000) = 16#014F0000) then
            (                                              /* PC relative */
            if ((op_code and 16#00800000) = 0) then
              printhex(read_PC - ((op_code>>4) and 16#F0) - (op_code and 16#0F))
            else
              printhex(read_PC + ((op_code>>4) and 16#F0) + (op_code and 16#0F));
            printstr("]")
            )
          else
            (
            Preg((op_code >> 16) and 15, false);
            if ((op_code and 16#01000000) = 0) then printstr("]");
            if ((op_code and 16#00400000) != 0) then
              (
              if ((op_code and 16#F0F) != 0) then
                (
                if ((op_code and 16#00800000) = 0) then printstr(" - #")
                else                                    printstr(" + #");
                printhex(((op_code>>4) and 16#F0) or (op_code and 16#0F))
                )
              )
            else
              (
              if ((op_code and 16#00800000) = 0) then printstr(" - ")
              else                                    printstr(" + ");
              Preg(op_code and 15, false) 
              );

            if ((op_code and 16#01000000) != 0) then printstr("]")
            );
          if ((op_code and 16#00200000) != 0) then printstr("!")
          )
        )
      )
    )
  else
    (
    data_op((op_code >> 21) and 15, (op_code and 16#00100000) != 0, position);
    if BL1 then (printstr("L"); position := position + 1);      /* & link */
    condition((op_code>>28) and 15, position);
    if ((op_code and 16#01800000) != 16#01000000) then
      Sbit((op_code and 16#00100000) != 0, position);
    addr_tab(position);
    if ((op_code and 16#01800000) != 16#01000000) then              /* Rd */
      (Preg((op_code >> 12) and 15, false); printstr(", "));
    if ((op_code and 16#01A00000) != 16#01A00000) then              /* Rn */
      (Preg((op_code >> 16) and 15, false); printstr(", "));

    if (!special) then
      addr_mode((op_code and 16#02000000) != 0, op_code)
    else
      (
      printstr("#");
      if BL0 then
        (
        if ((op_code and 16#00000200) = 0) then
          printhex(((op_code<<13) or (LSB<<12)) and 16#007FF000)
        else
          printhex(((op_code<<13) or (LSB<<12))  or 16#FF800000)
        )
      else
        print_hex(((op_code<<2) or (LSB<<1)) and 16#00000FFE, 3)
      )
    )
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

LSR_etc : expr(void) =
  (
  if ((op_code and 16#0E000010) = 16#06000010) then
    (
    printstr("Undef");
    condition((op_code>>28) and 15, position)
    )
  else
    (
    if ((op_code and 16#00100000) != 0) then printstr("LDR")
    else                                     printstr("STR");
    condition((op_code>>28) and 15, position);
    if ((op_code and 16#00400000) != 0) then
      (printstr("B"); position := position + 1);
    if ((op_code and 16#01200000) = 16#00200000) then
      (printstr("T"); position := position + 1 );
    addr_tab(position);
    Preg((op_code >> 12) and 15, false); printstr(", [");
    if ((op_code and 16#032F0000) = 16#010F0000) then
      (                                                    /* PC relative */
      x : var(int).

      if ((op_code and 16#00800000) = 0) then
        x := read_PC - (op_code and 16#FFF)
      else
        x := read_PC + (op_code and 16#FFF);
      if Thumb then x := x and 16#FFFFFFFC;      /* PROBABLY - WHO KNOWS? */
      printhex(x); printstr("]")

/*    if ((op_code and 16#00800000) = 0) then
        printhex(read_PC - (op_code and 16#FFF))
      else
        printhex(read_PC + (op_code and 16#FFF));
      printstr("]")           */
      )
    else
      (
      Preg((op_code >> 16) and 15, false);
      if ((op_code and 16#01000000) = 0) then printstr("]");
      if ((op_code and 16#02000000) = 0) then
        (
        if ((op_code and 16#FFF) != 0) then
          (
          if ((op_code and 16#00800000) = 0) then printstr(" - #")
          else                                    printstr(" + #");
          print_hex(op_code and 16#FFF, 3)
          )
        )
      else
        (
        if ((op_code and 16#00800000) = 0) then printstr(" - ")
        else                                    printstr(" + ");
        addr_mode(false, op_code) 
        );

      if ((op_code and 16#01000000) != 0) then
        (
        printstr("]");
        if ((op_code and 16#00200000) != 0) then printstr("!")
        )
      )
    )
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

LSM : expr(void) =
  (
  if ((op_code and 16#00100000) != 0) then printstr("LDM")
  else                                     printstr("STM");
  condition((op_code>>28) and 15, position);
  if ((op_code and 16#00800000) != 0) then printstr("I")
  else                                     printstr("D");
  if ((op_code and 16#01000000) != 0) then printstr("B")
  else                                     printstr("A");
  position := position + 2;
  addr_tab(position);
  Preg((op_code >> 16) and 15, false);
  if ((op_code and 16#00200000) != 0) then printstr("!");
  printstr(", ");
  Preg_list(op_code and 16#FFFF, false, false);
  if ((op_code and 16#00400000) != 0) then printstr("^")
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

Branches : expr(void) =
  (
  printstr("B");
  if ((op_code and 16#01000000) != 0) then
    (printstr("L"); position := position - 1)
  else
    (
    if ((op_code and 16#F0000000) = 16#E0000000) then printstr("RA")
    else position := position - 2
    );
  condition((op_code>>28) and 15, position);
  addr_tab(position);
  if ((op_code and 16#00800000) = 0) then
    printhex(read_PC + (((op_code<<2) + (LSB<<1)) and 16#03FFFFFE))
  else
    printhex(read_PC + (((op_code<<2) + (LSB<<1)) or  16#FC000000))
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

Co_pro : expr(void) =
  (
  if ((op_code and 16#00100000) = 0) then printstr("STC")
  else                                    printstr("LDC");
  condition((op_code>>28) and 15, position);
  addr_tab(position);
  printstr("p");    printint((op_code >> 8) and 16#F);
  printstr(", CR"); printint((op_code >> 12) and 16#F);
  printstr(", [R"); printint((op_code >> 16) and 16#F);
  if ((op_code and 16#01000000) = 0) then printstr("]");
  if ((op_code and 16#00800000) = 0) then printstr(" - #")
  else                                    printstr(" + #");
  print_hex((op_code and 16#FF)<<2, 3);
  if ((op_code and 16#01000000) != 0) then printstr("]");
  if ((op_code and 16#01200000) = 16#01200000) then printstr("!")
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

SWI_etc : expr(void) =
  (
  if ((op_code and 16#01000000) = 0) then
    (
    if ((op_code and 16#00000010) = 0) then
      (
      printstr("CDP");
      condition((op_code>>28) and 15, position);
      addr_tab(position);
      printstr("p");    printint((op_code >> 8) and 16#F);
      printstr(", ");   printint((op_code >> 20) and 16#F);
      printstr(", CR")
      )
    else
      (
      if ((op_code and 16#00100000) = 0) then printstr("MCR")
      else                                    printstr("MRC");
      condition((op_code>>28) and 15, position);
      addr_tab(position);
      printstr("p");    printint((op_code >> 8) and 16#F);
      printstr(", ");   printint((op_code >> 21) and 16#7);
      printstr(", R")
      );
    printint((op_code >> 12) and 16#F);
    printstr(", CR"); printint((op_code >> 16) and 16#F);
    printstr(", CR"); printint((op_code) and 16#F);
    printstr(", ");   printint((op_code >> 5) and 16#7)
    )
  else
    (
    printstr("SWI");
    condition((op_code>>28) and 15, position);
    addr_tab(position);
    print_hex(op_code and 16#00FFFFFF, 6)
    )
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  if Thumb then read_PC := PC + 4 else read_PC := PC + 8;

  BL0 := special && ((op_code and 16#00001000) = 0);
  BL1 := special && (!BL0);

  if (extension != 0) then  print_instr_ext(PC)         /* Op. superceded */
  else                      print_instr(PC, op_code); /* Normal operation */

  position := 3;                              /* Assume 3 letter mnemonic */

  if ((extension and 16#30) = 0) then
    (case (op_code and 16#0E000000)
      when 16#00000000 => Arith_etc
      when 16#02000000 => Arith_etc
      when 16#04000000 => LSR_etc
      when 16#06000000 => LSR_etc
      when 16#08000000 => LSM
      when 16#0A000000 => Branches
      when 16#0C000000 => Co_pro
      when 16#0E000000 => SWI_etc
    )
  else
    (
    if ((extension and 16#20) = 0) then
      (case extension
        when 16#19 => printstr("FIQ")
        when 16#1A => printstr("IRQ")
        when 16#1F => printstr("Prefetch abort")
        default   printstr("Illegal extension")
      )
    else          printstr("Predicted branch  ")
    );
  printstr("\n")
  ).

/*------------------------------------------------------------------------*/
/* Print an unbranch marker in the same format as disassembly             */

Print_unbranch(PC : val(int)) : expr(void) =
  (print_instr_ext(PC); printstr("Unbranch\n")).

/*------------------------------------------------------------------------*/
/* Print an abort marker in the same format as disassembly                */

Print_abort(PC : val(int)) : expr(void) =
  (print_instr_ext(PC); printstr("Data abort\n")).

/*------------------------------------------------------------------------*/
/* Print an abandoned instruction in the same format as disassembly       */

Print_CC_fail(PC : val(int), op_code : val(int)) : expr(void) =
  (print_instr(PC, op_code); printstr("CC failure\n")).

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

Tdisassemble(PC : val(int), op_in : val(int), extension : val(int)):
            expr(void)=
  (

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

Pbranch(CC : val(int)) : expr(void) =
  (case CC
    when  0 => printstr("BEQ")
    when  1 => printstr("BNE")
    when  2 => printstr("BCS")
    when  3 => printstr("BCC")
    when  4 => printstr("BMI")
    when  5 => printstr("BPL")
    when  6 => printstr("BVS")
    when  7 => printstr("BVC")
    when  8 => printstr("BHI")
    when  9 => printstr("BLS")
    when 10 => printstr("BGE")
    when 11 => printstr("BLT")
    when 12 => printstr("BGT")
    when 13 => printstr("BLE")
    when 14 => printstr("BRA")
    when 15 => printstr("SWI")
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  op_code  : var(int).
  position : var(int).

  op_code := op_in and 16#FFFF;

  printint(now); printstr(":  ");
  printhex(PC);  printstr("      "); print_hex(op_code, 4); printstr("    ");
  printasc(op_code, true); printstr("  ");

  position := 3;                              /* Assume 3 letter mnemonic */

  if (extension = 0) then
    (
    case ((op_code>>13) and 7)

    when 0 =>
      (
      if ((op_code and 16#1800) != 16#1800) then
        (
        (case (op_code and 16#1800)
          when 16#0000 => printstr("LSL")
          when 16#0800 => printstr("LSR")
          when 16#1000 => printstr("ASR")
        );
        addr_tab(position);
        Preg(op_code and 7, false);
        printstr(", "); Preg((op_code >> 3) and 7, false);
        printstr(", #"); print_hex((op_code >> 6) and 16#1F, 2)
        )
      else
        (
        if ((op_code and 16#0200) = 0) then printstr("ADD")
        else                                printstr("SUB");
        addr_tab(position);
        Preg(op_code and 7, false);
        printstr(", "); Preg((op_code >> 3) and 7, false);
        if ((op_code and 16#0400) = 0) then
          (printstr(", "); Preg((op_code >> 6) and 7, false))
        else
          (printstr(", #"); print_hex((op_code >> 6) and 7, 1))
        )
      )

    when 1 =>
      (
      (case (op_code and 16#1800)
        when 16#0000 => printstr("MOV")
        when 16#0800 => printstr("CMP")
        when 16#1000 => printstr("ADD")
        when 16#1800 => printstr("SUB")
      );
      addr_tab(position);
      Preg((op_code >> 8) and 7, false); printstr(", #");
      print_hex(op_code and 16#FF, 2)
      )

    when 2 =>
      (
      if ((op_code and 16#1000) = 0) then
        (
        if ((op_code and 16#0800) = 0) then
          (
          if ((op_code and 16#0400) = 0) then
            (
            (case (op_code and 16#03C0)
              when 16#0000 => printstr("AND")
              when 16#0040 => printstr("EOR")
              when 16#0080 => printstr("LSL")
              when 16#00C0 => printstr("LSR")
              when 16#0100 => printstr("ASR")
              when 16#0140 => printstr("ADC")
              when 16#0180 => printstr("SBC")
              when 16#01C0 => printstr("R0R")
              when 16#0200 => printstr("TST")
              when 16#0240 => printstr("NEG")
              when 16#0280 => printstr("CMP")
              when 16#02C0 => printstr("CMN")
              when 16#0300 => printstr("ORR")
              when 16#0340 => printstr("MUL")
              when 16#0380 => printstr("BIC")
              when 16#03C0 => printstr("MVN")
            );
            addr_tab(position);
            Preg(op_code and 7, false); printstr(", ");
            Preg((op_code >> 3) and 7, false)
            )
          else
            (
            (case (op_code and 16#0300)
              when 16#0000 => printstr("ADD")
              when 16#0100 => printstr("CMP")
              when 16#0200 => printstr("MOV")
              when 16#0300 => printstr("BX ")  /* NB also "undef. case */
            );
            addr_tab(position);
            if ((op_code and 16#0300) != 16#0300) then
              (Preg(op_code and 7, (op_code and 16#0080) != 0); printstr(", "));
            Preg((op_code >> 3) and 7, (op_code and 16#0040) != 0)
            )
          )
        else
          (
          printstr("LDR");
          addr_tab(position);
          Preg((op_code >> 8) and 7, false);           printstr(", [");
          printhex((PC + 4 + ((op_code and 16#FF)<<2)) and 16#FFFFFFFC);
          printstr("]")                         /* BODGE ? ^^^^^^^^^^^ */
          )
        )
      else
        (
        (case (op_code and 16#0E00)
          when 16#0000 => printstr("STR  ")
          when 16#0200 => printstr("STRH ")
          when 16#0400 => printstr("STRB ")
          when 16#0600 => printstr("LDRSB")
          when 16#0800 => printstr("LDR  ")
          when 16#0A00 => printstr("LDRH ")
          when 16#0C00 => printstr("LDRB ")
          when 16#0E00 => printstr("LDRSH")
        );
        position := position + 2;
        addr_tab(position);
        Preg( op_code and 7,       false); printstr(", [");
        Preg((op_code >> 3) and 7, false); printstr(", ");
        Preg((op_code >> 6) and 7, false); printstr("]")
        )
      )

    when 3 =>
      (
      if ((op_code and 16#0800) = 0) then printstr("STR")
      else                                printstr("LDR");
      if ((op_code and 16#1000) != 0) then
        (printstr("B"); position := position + 1);
      addr_tab(position);
      Preg( op_code and 7,       false); printstr(", [");
      Preg((op_code >> 3) and 7, false); printstr(", #");
      if ((op_code and 16#1000) = 0) then
        print_hex((op_code >> 4) and 16#7C, 2)    /* Word offset */
      else
        print_hex((op_code >> 6) and 16#1F, 2);   /* Byte offset */
      printstr("]")
      )

    when 4 =>
      (
      if ((op_code and 16#1000) = 0) then
        (
        if ((op_code and 16#0800) = 0) then printstr("STRH")
        else                                printstr("LDRH");
        position := position + 1;
        addr_tab(position);
        Preg( op_code and 7,       false); printstr(", [");
        Preg((op_code >> 3) and 7, false); printstr(", #");
        print_hex((op_code >> 5) and 16#1F, 2); printstr("]")
        )
      else
        (
        if ((op_code and 16#0800) = 0) then printstr("STR")
        else                                printstr("LDR");
        addr_tab(position);
        Preg((op_code >> 8) and 7, false); printstr(", [SP, #");
        print_hex((op_code and 16#FF)<<2, 3); printstr("]")
        )
      )

    when 5 =>
      (
      if ((op_code and 16#1000) = 0) then
        (
        if ((op_code and 16#0800) = 0) then printstr("LDA") /* Load address */
        else                                printstr("ADD");
        addr_tab(position);
        Preg((op_code >> 8) and 7, false);
        if ((op_code and 16#0800) = 0) then
          (
          printstr(", #");
          printhex((PC + 4 + ((op_code and 16#FF)<<2)) and 16#FFFFFFFC)
          )
        else
          (printstr(", SP, #"); print_hex((op_code and 16#FF)<<2, 3))
        )
      else
        (
        case (op_code and 16#0E00)

          when 16#0000 =>
            (
            if ((op_code and 16#0100) = 0) then
              (
              if ((op_code and 16#0080) = 0) then printstr("ADD")
              else                                printstr("SUB");
              addr_tab(position);
              printstr("SP, SP, #"); print_hex((op_code and 16#7F)<<2, 2)
              )
            else
              printstr("Not specified")
            )

          when 16#0400 =>
            (
            printstr("PUSH");   position := position + 1;
            addr_tab(position);
            Preg_list(op_code and 16#FF, (op_code and 16#0100) != 0, false)
            )

          when 16#0C00 =>
            (
            printstr("POP"); addr_tab(position);
            Preg_list(op_code and 16#FF, false, (op_code and 16#0100) != 0)
            )

          default  printstr("Not specified")
        )
      )

    when 6 =>
      (
      if ((op_code and 16#1000) = 0) then
        (
        if ((op_code and 16#0800) = 0) then printstr("STM")
        else                                printstr("LDM");
        addr_tab(position);
        Preg((op_code >> 8) and 7, false); printstr("!, ");
        Preg_list(op_code and 16#FF, false, false)
        )
      else
        (
        Pbranch((op_code>>8) and 16#F);
        addr_tab(position);
        if ((op_code and 16#0F00) != 16#0F00) then
          (
          if ((op_code and 16#80) = 0) then
            printhex(PC + 4 + ((op_code and 16#000000FF) << 1))
         else
            printhex(PC + 4 + ((op_code  or 16#FFFFFF00) << 1))
          )
        else
          print_hex(op_code and 16#FF, 2)                /* SWI */
        )
      )

    when 7 =>
      (
      case (op_code and 16#1800)

        when 16#0000 =>
          (
          printstr("BRA"); addr_tab(position);
          if ((op_code and 16#0400) = 0) then
            printhex(PC + 4 + ((op_code and 16#000007FF) << 1))
          else
            printhex(PC + 4 + ((op_code  or 16#FFFFF800) << 1))
          )

        when 16#0800 => printstr("Undef")

        when 16#1000 =>
          (
          printstr("BL0"); addr_tab(position);
          if ((op_code and 16#0400) = 0) then
            printhex(PC + 4 + ((op_code and 16#000007FF) << 12))
          else
            printhex(PC + 4 + ((op_code  or 16#FFFFF800) << 12))
          )

        when 16#1800 =>
          (
          printstr("BL1"); addr_tab(position);
          printhex((op_code and 16#07FF) << 1)
          )
      )
    )
  else
    (
    printstr("extension")
    );

  printstr("\n")

  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
