/*------------------------------------------------------------------------*/
/*                                                                        */
/*                       AMULET3 Execution Unit - part 1                  */
/*                                                                        */
/*                                             Created: JDG  31/1/97      */
/*                                       Last Modified: JDG  28/7/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/* Condition code check                                                   */

Flag_match(op_condition : val(int), flags : val(int)) : expr(bool) =
  (
  status    : var(bool).

  status    := ((op_condition and 1) = 0);

  (case (op_condition and 16#E)
    when  0 => (((flags and 2#0100) != 0) = status)              /* EQ/NE */
    when  2 => (((flags and 2#0010) != 0) = status)              /* CS/CC */
    when  4 => (((flags and 2#1000) != 0) = status)              /* MI/PL */
    when  6 => (((flags and 2#0001) != 0) = status)              /* VS/VC */
    when  8 => (((flags and 2#0110) = 2#0010) = status)          /* HI/LS */
    when 10 => ((((flags and 2#1000) != 0) = ((flags and 2#0001) != 0)) = status)
                                                                 /* GE/LT */
    when 12 => ((((flags and 2#0100) = 0) &&                     /* GT/LE */
          (((flags and 2#1000) != 0) = ((flags and 2#0001) != 0))) = status)
    when 14 => (status)                                          /* AL/NV */
  )

  ).

/*------------------------------------------------------------------------*/

Shift(input     : val(int),
      Cin       : val(int),
      Cout      : var(int),
      quantity  : val(int),
      operation : val(int),
      delay     : var(int)):
      expr(int) =
  (

  delay := delay + Shift_delay;

  (case operation

    when 0 =>                                                      /* LSL */
      (      
      if (quantity = 0) then Cout := Cin
      else Cout := (if (quantity > 32) then 0
                    else (input >> (32-quantity)) and 1);
      if (quantity >= 32) then 0  /* Should be unnecessary - isn't in v.4 */
      else input << quantity
      )

    when 1 =>                                                      /* LSR */
      (      
      if (quantity = 0) then Cout := Cin
      else Cout := (if (quantity > 32) then 0
                    else (input >> (quantity-1)) and 1);
      if (quantity >= 32) then 0
      else (input >> quantity) and ((1<<(32-quantity))-1)/* Zero extend */
      )

    when 2 =>                                                      /* ASR */
      (      
      if (quantity = 0) then Cout := Cin
      else Cout := (if (quantity < 32) then (input >> (quantity-1)) and 1
                    else (input >> 31) and 1);       /* Stick at sign bit */
      if ((input and 16#80000000) = 0) then
        (                                                     /* Positive */
        if (quantity >= 32) then 0
        else (input >> quantity) and ((1<<(32-quantity))-1)     /* as LSR */
        )
      else
        (                                                     /* Negative */
        if (quantity >= 32) then -1
        else (input >> quantity) or (not((1<<(32-quantity))-1)) /* 1 ext. */
        )
      )

    when 3 =>                                                      /* ROR */
      (      
      if (quantity = 0) then Cout := Cin
      else
        (
        if ((quantity and 16#1F) != 0) then
          Cout := (input >> ((quantity and 16#1F)-1)) and 1        /* bit */
        else
          Cout := (input >> 31) and 1                    /* Copy sign bit */
        );
      ((input >> quantity) and not(16#FFFFFFFF << (32-quantity)))
                                     or (input << (32-quantity))
      )

    when 4 =>                                                      /* RRX */
      (      
      Cout := input and 1;
      ((input >> 1) and 16#7FFFFFFF) or (Cin<<31)
      )

    default  (printstr("Unrecognised shift code\n"); 0)

  )

  ).

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

NZC_flags(value : val(int), carry : val(int), flags : var(int)):
          expr(void) =
  (
  if ((value and 16#80000000) != 0) then flags := flags or  2#1000
  else                                   flags := flags and 2#0111;
  if (value = 0) then flags := flags or  2#0100
  else                flags := flags and 2#1011;
  if (carry != 0) then flags := flags or  2#0010
  else                 flags := flags and 2#1101
  ).

/*------------------------------------------------------------------------*/

assemble_PSR(input : val(int)) : expr(PSR_type)=
  (
  temp : var(PSR_type).

  temp->ctrl->IRQ_En := ((input and 16#00000080) = 0);
  temp->ctrl->FIQ_En := ((input and 16#00000040) = 0);
  temp->ctrl->Thumb  := ((input and 16#00000020) != 0);
  temp->ctrl->mode   := input and 16#F;
  temp->flags        := (input >> 28) and 16#F;
  temp
  ).

/*------------------------------------------------------------------------*/

disassemble_PSR(input : val(PSR_type)) : expr(int)=
  (
    (input->flags << 28)
 or (if (input->ctrl->IRQ_En) then 16#00 else 16#80)
 or (if (input->ctrl->FIQ_En) then 16#00 else 16#40)
 or (if (input->ctrl->Thumb)  then 16#20 else 16#00)
 or 16#10
 or input->ctrl->mode
  ).

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

ALU_eval(operation : val(int),
         set_flags : val(bool),
           A       : val(int),
           B       : val(int),
           C       : val(int),
           Shift_C : val(int),
           flags   : var(int),
           delay   : var(int)):
           expr(int) =
  (

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

op_mov(B : val(int), S : val(bool), flags : var(int), SC : val(int)):
        expr(int) =
  (
  Result : var(int).

  delay := delay + ALU_logical_delay;

  Result := B;
  if S then NZC_flags(Result, SC, flags);
  Result
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

op_add(A : val(int), B : val(int), C : val(int),
       S : val(bool), flags : var(int)):
       expr(int) =
  (
  Result : var(int).

  delay := delay + ALU_arith_delay;

  Result := A + B + C;

  if S then
    (
    NZC_flags(Result,
             (if ((((A and B) or ((A xor B) and not(Result)))
                                            and 16#80000000) != 0) then 1
              else 0),                            /* Deal with carry flag */
              flags);

    if (((not(A xor B) and (A xor Result)) and 16#80000000) != 0) then
      flags := flags or  2#0001                /* Deal with overflow flag */
    else
      flags := flags and 2#1110
    );
  Result
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

op_sub(A : val(int), B : val(int), C : val(int),
       S : val(bool), flags : var(int)):
       expr(int) =
  (
  Result : var(int).

  delay := delay + ALU_arith_delay;

  Result := A - B - (C xor 1);

  if S then
    (
    NZC_flags(Result,
             (if ((((A and not(B)) or ((A xor not(B)) and not(Result)))
                                             and 16#80000000) != 0) then 1
              else 0),                            /* Deal with carry flag */
              flags);

    if ((((A xor B) and (A xor Result)) and 16#80000000) != 0) then
      flags := flags or  2#0001                /* Deal with overflow flag */
    else
      flags := flags and 2#1110
    );
  Result
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

op_and(A : val(int), B : val(int), S : val(bool),
       flags : var(int), SC : val(int)): expr(int) =
  (
  Result : var(int).

  delay := delay + ALU_logical_delay;

  Result := A and B;
  if S then NZC_flags(Result, SC, flags);
  Result
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

op_orr(A : val(int), B : val(int), S : val(bool),
       flags : var(int), SC : val(int)): expr(int) =
  (
  Result : var(int).

  delay := delay + ALU_logical_delay;

  Result := A or B;
  if S then NZC_flags(Result, SC, flags);
  Result
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

op_eor(A : val(int), B : val(int), S : val(bool),
       flags : var(int), SC : val(int)): expr(int) =
  (
  Result : var(int).

  delay := delay + ALU_logical_delay;

  Result := A xor B;
  if S then NZC_flags(Result, SC, flags);
  Result
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  (case operation
    when  0 => op_and(A, B,      set_flags, flags, Shift_C)
    when  1 => op_eor(A, B,      set_flags, flags, Shift_C)
    when  2 => op_sub(A, B, 1,   set_flags, flags)
    when  3 => op_sub(B, A, 1,   set_flags, flags)
    when  4 => op_add(A, B, 0,   set_flags, flags)
    when  5 => op_add(A, B, C,   set_flags, flags)
    when  6 => op_sub(A, B, C,   set_flags, flags)
    when  7 => op_sub(B, A, C,   set_flags, flags)
    when  8 => op_and(A, B,      set_flags, flags, Shift_C)
    when  9 => op_eor(A, B,      set_flags, flags, Shift_C)
    when 10 => op_sub(A, B, 1,   set_flags, flags)
    when 11 => op_add(B, A, 0,   set_flags, flags)
    when 12 => op_orr(A, B,      set_flags, flags, Shift_C)
    when 13 => op_mov(B,         set_flags, flags, Shift_C)
    when 14 => op_and(A, not(B), set_flags, flags, Shift_C)
    when 15 => op_mov(not(B),    set_flags, flags, Shift_C)
    default  (printstr("ALU op. out of range\n"); 0)
  )

  ).

/*------------------------------------------------------------------------*/

(null)

/*------------------------------------------------------------------------*/
