/*------------------------------------------------------------------------*/
/*                                                                        */
/*                       AMULET3 Execution Unit - part 2                  */
/*                                                                        */
/*                                             Created: JDG  31/1/97      */
/*                                       Last Modified: JDG  25/9/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/

/* To do: Various multiplies                                              */
/*                                                                        */

/* Analyse and debug PSR stuff - e.g. a prefetch abort has caused a CPSR  */
/* restore from user mode :-(                                             */

/*------------------------------------------------------------------------*/

Execute(Instr       : var(chan(Decoded_instr_type)),
        C_flow      : var(chan(int)),
        LSM_go      : var(chan(bool)),
        Q_out       : var(chan(Queue_input_type)),
        address     : var(chan(DAI_addr_pkt_type)),
        Areg        : var(chan(int)),
        Breg        : var(chan(int)),
        C1reg       : var(chan(int)),
        Xpipe       : var(chan(Xpipe_type)),
        BRpipe      : var(chan(BRpipe_type)),
        Branch_chan : var(chan(Branch_type)),
        Data_abort  : var(chan(PSR_type))):
        expr(void) =
  (

  Branch_decouple : var(chan(Branch_type)).

  Colour      : var(int).
  New_Colour  : var(int).
  CPSR        : var(PSR_type).
  New_CPSR    : var(PSR_type).
  SPSR        : var(PSR_type^16).

  A_pipe      : var(int).
  B_pipe      : var(int).
  C_pipe      : var(int).
  Shift_carry : var(int).
  ALU_result  : var(int).

  Command     : var(Decoded_instr_type).
  Branch      : var(Branch_type).
  to_DAI      : var(DAI_addr_pkt_type).
  to_queue    : var(Queue_input_type).
  to_Xpipe    : var(Xpipe_type).
  to_BRpipe   : var(BRpipe_type).

  colour_match : var(bool).
  all_match    : var(bool).
  predicted    : var(bool).
  Thumb_En     : var(bool).

  Execution_delay : var(int).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Here for appearance sake only - (aids readability of main prog.)       */

Exec_data_abort : expr(void) =
  (
  wait_for(EXEC_abort_delay);

  if dis_flag then Print_abort(Command->PC);

  Data_abort?(CPSR := (?Data_abort));            /* Handshake too early ??*/

  New_Colour := Colour xor 2;                          /* Toggle Qcol bit */
  SPSR[7]    := CPSR;
  CPSR->ctrl->mode   := 7;                      /* May be unnecessary (?) */
  CPSR->ctrl->Thumb  := false;                  /* May be unnecessary (?) */
  CPSR->ctrl->IRQ_En := false;                  /* May be unnecessary (?) */
  Branch->Exception  := 16;                            /* Data abort code */
  Branch->New_col    := New_Colour;

  Branch->New_PC     := 16#00000010;              /* Dabt handler address */
  Branch->New_PSR_ctrl->mode   := 7;
  Branch->New_PSR_ctrl->Thumb  := false;
  Branch->New_PSR_ctrl->IRQ_En := false;
  Branch->New_PSR_ctrl->FIQ_En := CPSR->ctrl->FIQ_En;
  Branch->Indir_req  := false;      /* Unnecessary - included for display */
  Branch->Old_PC     := Command->PC;       /*        Ditto                */

  Branch_decouple!Branch;
  Colour := New_Colour
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

DAI_output : expr(void) =
  (
  wait_for(Execution_delay);

  /* Pre- or post- increment mode                                         */
  if (((Command->Op_code and 16#05000000) = 16#05000000)     /* LSR & LSC */
   ||  (((Command->Op_code and 16#0F000000) = 16#01000000)        /* LSRH */
     && ((Command->Op_code and 16#00000060) != 0))         /* but not SWP */
   || ((Command->Op_code and 16#0E800000) = 16#08000000)) then     /* LSM */
    to_DAI->addr := ALU_result
  else
    to_DAI->addr := A_pipe;

  to_DAI->mode       := CPSR->ctrl->mode;      /* Mode before instruction */
  to_DAI->colour     := Colour and 2;                /* Just the Qcol bit */
  to_DAI->invalidate := false;
  to_DAI->pre_inc    := Command->addr_pre_inc;    /* Goes at address time */

  to_Xpipe->CPSR_flag := true;           /* Save PC if memory cycle to go */
  to_Xpipe->PC        := Command->PC + 8;
                               /* Or increment at far end (_if needed_ ?) */
  to_Xpipe->PSR       := CPSR;

  if (Command->store) then
    STR_semaphore := STR_semaphore + 1;

  (address!to_DAI | Xpipe!to_Xpipe)
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* SPSRs may share Xpipe with memory instructions, but are mutually exclusive */

SPSR_output : expr(void) =
  (
  wait_for(Execution_delay);            /* Timing should be function dependent */
/*
History buffer or locking mechanism
  to_Xpipe->CPSR_flag := false;                           /* SPSR instead */
  to_Xpipe->PC        := CPSR->ctrl->mode;            /* Artificial reuse */
  to_Xpipe->PSR       := SPSR[CPSR->ctrl->mode];
  Xpipe!to_Xpipe
*/
/*
printstr("Changing SPSR["); printint(New_CPSR->ctrl->mode); printstr("]\n");
*/
  SPSR[New_CPSR->ctrl->mode] :=                           /* in new mode. */
                (if (Command->flags->MSR) then assemble_PSR(B_pipe)
                else CPSR)                    /* Copy for exception entry */
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

BR_output : expr(void) =
  (                                                 /* Save base register */
  wait_for(Execution_delay);
  to_BRpipe->base := A_pipe;                  /* Save base register value */
  to_BRpipe->reg  := (Command->Op_code >> 16) and 16#F;             /* Rn */
  to_BRpipe->mode := CPSR->ctrl->mode;         /* Mode before instruction */
  BRpipe!to_BRpipe
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

Branch_output : expr(void) =
  (
  New_Colour        := Colour xor 1;              /* Change branch colour */

  Branch->New_col   := New_Colour;
  Branch->Indir_req := Command->LD_PC;
  Branch->Exception := Command->Exception;
             /* Possibilities are none (0), Undef (4), SWI (8), Pabt (12) */

  Branch->Old_PC    := Command->PC;
  Branch->Cacheable := Command->branch->branch;      /* also Thumb BL ??? */
  Branch->link      := Command->branch->link;
  Branch->CC        := Command->condition;
  Branch->flush     := false;                        /* Clear BTC command */
  Branch->halt      := (Command->Op_code = 16#EAFFFFFE);
                                            /* Thumb may affect halt etc. */

  if (Command->branch->branch) then
    (
    wait_for(EXEC_branch_delay);       /* Allows faster branch simulation */
    if Command->PSR_ctrl->Thumb then
      Branch->New_PC := Command->PC + 4 + Command->Immediate->Immediate
    else
      Branch->New_PC := Command->PC + 8 + Command->Immediate->Immediate
    )
  else
    (
    wait_for(Execution_delay);
    if (Command->flags->MSR) then                         /* Must be CSPR */
      (                                            /* Beware of mode etc. */
      temp : var(PSR_type).                           /* Operand assembly */

      Branch->New_PC := Command->PC + 4;                /* Never in Thumb */

      temp := assemble_PSR(B_pipe);

      if ((Command->Op_code and 16#00080000) != 0) then
        New_CPSR->flags := temp->flags;                   /* Not CPSR yet */
      if (((Command->Op_code and 16#00010000) != 0)
        && (CPSR->ctrl->mode != 0)) then              /* Not in user mode */
        New_CPSR->ctrl  := temp->ctrl                     /* Not CPSR yet */
      )
    else
      (
      if (  ((Command->flags->LSM) && (Command->LD_PC)             /* PC^ */
         && ((Command->Op_code and 16#00400000) != 0))
       ||   ((!Command->LD_PC)       /* ALU operations to PC - test S bit */
         && ((Command->Op_code and 16#00100000) != 0))) then
        (
        if (((CPSR->ctrl->mode and 16#C) = 16#0)
         || ((CPSR->ctrl->mode and 16#3) = 16#3)) then      /* Legal mode */
          (
          if ((CPSR->ctrl->mode != 16#0) && (CPSR->ctrl->mode != 16#F)) then
            New_CPSR := SPSR[CPSR->ctrl->mode]/* Restore CPSR (BX immune) */
          else
            (
            New_CPSR := CPSR;   /* Not in definition but in verification! */
            printstr("Attempting CPSR restore from user/system mode\n");
            printstr("Time = "); printint(now);
            printstr("  PC = "); printhex(Command->PC);
            printstr("\n")
            )
          )
        else
          printstr("Attempting CPSR restore from illegal mode\n")
        );

      if (Command->branch->BX) then
        New_CPSR->ctrl->Thumb := ((B_pipe and 1) != 0);  /* or ALU_res. ? */

      if (Command->LD_PC
      && (    (Command->flags->LSM
          && ((Command->Op_code and 16#00800000) = 0))
        ||   ((!Command->flags->LSM)
          && ((Command->Op_code and 16#01000000) = 0)))) then/*Test "P" bit*/
        Branch->New_PC := A_pipe
      else
        Branch->New_PC := ALU_result;

      if ((Command->flags->LSM) && (Command->LD_PC)) then
        Branch->New_PC := Branch->New_PC and 16#FFFFFFFC;
                                                /* No data rotate for LDM */

      if ((Command->flags->LSM) && (Command->LD_PC)   /* & maybe adjust ? */
      && (((Command->Op_code xor (Command->Op_code>>1))
          and 16#00800000) != 0)) then                         /* P xor U */
        Branch->New_PC := Branch->New_PC - 4               /* Predecrement*/
      )
    );

  Branch->New_PSR_ctrl := New_CPSR->ctrl;                /* New mode etc. */
                                             /* Here to be after MSR code */
  Branch_decouple!Branch
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

ALU_output : expr(void) =
  (                                                     /* Return results */
  wait_for(Execution_delay);
  if (Command->branch->link) then
    (
    if (Command->PSR_ctrl->Thumb && Thumb_En) then
      (                                       /* Thumb BL sets bottom bit */
      if (Command->branch->Thumb_BL) then to_queue->data := Command->PC + 3
      else                                to_queue->data := Command->PC + 2
      )
    else to_queue->data := Command->PC + 4
    )
  else
    to_queue->data     := ALU_result;

  to_queue->queue_addr := Command->queue_addr;
  to_queue->invalid    := false;
  to_queue->bv_colour  := Command->bv_colour;

  Q_out!to_queue
  ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  Colour     := 0;
  New_Colour := Colour;  /* To maintain consistency over parallel threads */

  CPSR->flags        := 0;
  CPSR->ctrl->mode   := 3;                    /* Start in supervisor mode */
  CPSR->ctrl->IRQ_En := false;
  CPSR->ctrl->FIQ_En := false;
  CPSR->ctrl->Thumb  := false;

  to_DAI->addr       := 0;              /* Just for tidiness - not needed */
  to_DAI->mode       := CPSR->ctrl->mode;
  to_DAI->colour     := Colour and 2;                /* Just the Qcol bit */

  to_queue->abort    := false;                            /* Cannot abort */
  to_queue->has_BR   := false;                      /* Never uses BR pipe */

/* This latch avoids a deadlock problem by decoupling the execution       */
/* process so that it can start to empty the prefetch without waiting for */
/* the IAI arbiter.                                                       */

  init_chan(Branch_decouple, "Branch_latch");
    (
    Latch(Branch_decouple, Branch_chan, Br_Lat_delay, false, "Branch_latch") |
                                                   /* Spawn latch process */

    forever
      (
                  /* First arbitrate between instructions and data aborts */
      wait_until(chan_ready(Instr)||chan_ready(Data_abort));

      Execution_delay := EXEC_delay;       /* Initialise cumulative delay */

      if chan_ready(Data_abort) then
        Exec_data_abort                             /* Data abort handler */
      else
        (                                  /* Begin instruction execution */
        (Instr?(Command := (?Instr)) | C_flow!Colour);

        /* Collect the input data                                         */
        if (Command->reg_usage->uses_A) then  Areg?(A_pipe :=  (?Areg));
        if (Command->reg_usage->uses_B) then  Breg?(B_pipe :=  (?Breg));
        if (Command->reg_usage->uses_C) then C1reg?(C_pipe := (?C1reg));

        active_block("execute",Command->bv_colour,(

        New_CPSR->flags := CPSR->flags;
        New_CPSR->ctrl  := Command->PSR_ctrl;

        colour_match := (Command->colour = Colour);
        all_match    := colour_match
                     && Flag_match(Command->condition, CPSR->flags);
        predicted := Command->branch->predicted;  /* Copy may be molested */

         /* Enable special operation PC+2 for Thumb                       */
        Thumb_En  := ((Command->Exception = 4)                   /* Undef */
                   || (Command->Exception = 8)                     /* SWI */
                   ||  Command->branch->Thumb_BL);/* Thumb BL (2nd instr.)*/

        if (Command->flags->LSM) then LSM_go!all_match;
                /* Signal instruction continuation status (CHECK TIMINGS) */
                                       /* MAY BE WORTH EXPERIMENTING WITH */

        if (!colour_match) then predicted := false; /* Scrap "unbranches" */

        if (dis_flag && colour_match
        && (!Flag_match(Command->condition, CPSR->flags))) then
          Print_CC_fail(Command->PC, Command->Op_code);

        if all_match then                 /* Execute instruction normally */
          (
          if dis_flag then Disassemble(Command->PC,
                                       Command->Op_code,
                                       Command->extension,
                                       Command->PSR_ctrl->Thumb,
                                       Command->branch->Thumb_BL,
                                    /* Only set for second part of BL !!! */
                                       (Command->Immediate->Immediate>>1) and 1);

/* Multiply or Immediate or Shift (which may be bypassed)                 */

          if (Command->operation->use_MUL) then
            (                                                 /* Multiply */
            if ((Command->Op_code and 16#00800000) = 0) then
              (
              (case (Command->Op_code and 16#00600000)
                when 16#00000000 => B_pipe := B_pipe * C_pipe
                when 16#00200000 => B_pipe := (B_pipe * C_pipe) + A_pipe
                default (printstr("Illegal MUL"); Run := false)
              );
              A_pipe := 0                 /* Ensure nothing else added in */
              )
            else
              (                                    /* Long multiplication */
              t1 : var(int). t2 : var(int). t3 : var(int).
              t4 : var(int). t5 : var(int). t6 : var(int).
/*
Input XX
              t1 := (B_pipe and 16#0000FFFF) * (C_pipe and 16#0000FFFF);
              t2 := ((B_pipe>>16) and 16#0000FFFF) * (C_pipe and 16#0000FFFF);
              t3 := (B_pipe and 16#0000FFFF) * ((C_pipe>>16) and 16#0000FFFF);
              t4 := (t2 and 16#0000FFFF) + (t3 and 16#0000FFFF)
                 + ((t1>>16) and 16#0000FFFF);
              outL := (t4<<16) or (t1 and 16#0000FFFF);
              outH := ((t2>>16) and 16#0000FFFF) + ((t3>>16) and 16#0000FFFF)
                    + ((t4>>16) and 16#0000FFFF)
                      + (((B_pipe>>16) and 16#0000FFFF)
                       * ((C_pipe>>16) and 16#0000FFFF));
*/ null
              )
            )
          else
            (
            if (Command->reg_usage->uses_Imm) then          /* Immediate? */
              (
              B_pipe := Command->Immediate->Immediate;
              Shift_carry := (if (Command->Immediate->default_C) then
                                (CPSR->flags >> 1) and 1
                              else
                                Command->Immediate->Imm_C)
              )
            else
              (                                                  /* Shift */
              if (Command->operation->shift) then
                B_pipe := Shift(B_pipe,
                      (CPSR->flags >> 1) and 1,
                      Shift_carry,
                      if (Command->operation->shift_imm) then
                        Command->operation->shift_amount    /* Imm. shift */
                      else                        /* Register based shift */
                        C_pipe and 16#FF,
                      Command->operation->shift_op,
                      Execution_delay)
              else
                (
                Shift_carry := (CPSR->flags >> 1) and 1; /* Default value */
                Execution_delay := Execution_delay + Shift_byp_delay
                )
              )                                                  /* Shift */
            );

/* End of Multiply/Immediate/Shift                                        */


          if (Command->operation->use_ALU) then
            ALU_result := ALU_eval(Command->operation->ALU_op,
                                   Command->operation->set_flags,
                                   A_pipe,
                                   B_pipe,
                                   (CPSR->flags >> 1) and 1, /* Old carry */
                                   Shift_carry,
                                   New_CPSR->flags,
                                   Execution_delay);

          if (Command->flags->MRS) then
            (                                         /* Operand assembly */
            if (((Command->Op_code and 16#00400000) = 0)
              || (CPSR->ctrl->mode = 0)) then
            /* (Second condition for verification prog. "TSPU-32.s" only) */
              ALU_result := disassemble_PSR(CPSR)
            else
              ALU_result := disassemble_PSR(SPSR[CPSR->ctrl->mode])
            );

          (                              /* Begin parallel output threads */
            if (Command->to_DAI) then DAI_output   /* Memory Instructions */
          |
            if (Command->to_SPSR) then SPSR_output
          |
            if (Command->flags->LSM) then BR_output   /* LSM Instructions */
          |
            if ((Command->dest_PC && (!Command->to_SPSR))
                                       /* Data ops. to PC and MSR to CPSR */
             || (Command->branch->branch && (!predicted))
                                                  /* Unpredicted branches */
             || (Command->Exception != 0)               /* SWI/Undef/etc. */
             || (Command->LD_PC)) then                        /* PC loads */
              Branch_output
          |
            if (Command->to_queue) then ALU_output  /* Return any results */
          );                                      /* End parallel threads */

          CPSR := New_CPSR;             /* Now safe to update mode (etc.) */
          Colour := New_Colour
          )

        else

          (                         /* Invalidated instruction - trash it */
          wait_for(EXEC_nogo_delay);
          (
          if (Command->to_queue) then        /* Is queue entry allocated? */
            (
            to_queue->data    := Command->Op_code; /* Makes O/P easier to read */
            to_queue->queue_addr := Command->queue_addr;
            to_queue->invalid := true;          /* Trash queue allocation */
            Q_out!to_queue
            )
          |
          if (Command->to_DAI) then        /* Is DAI expecting something? */
            (
            to_DAI->invalidate := true;
            address!to_DAI
            )
          |
          if (Command->branch->branch && predicted) then  /* CC fail only */
            (                                                 /* Unbranch */
            if dis_flag then Print_unbranch(Command->PC);

            New_Colour         := Colour xor 1;   /* Change branch colour */

            Branch->New_col    := New_Colour;
            Branch->New_PSR_ctrl := CPSR->ctrl;  /* Could use new mode if easier ??? */
            Branch->Indir_req  := false;
            Branch->Exception  := 0;
            Branch->Old_PC     := Command->PC;               /* Undefined */
            Branch->Cacheable  := false;
            Branch->link       := false;                     /* Undefined */
            Branch->CC         := 0;                         /* Undefined */
            Branch->halt       := false;
            if Command->PSR_ctrl->Thumb then Branch->New_PC := Command->PC + 2
            else                             Branch->New_PC := Command->PC + 4;

            Branch_decouple!Branch
            )
/*
          |                                       /* "nogo" code for LSMs */
          if (Command->flags->LSM) then LSM_go!false
                  /* This is done earlier - check for timing model though */
*/
          );

          Colour := New_Colour;

/*
          printstr("Exec: Instruction discarded - wrong colour\n")
*/        null
          )                               /* End of invalidation routines */
        )) /* active_block */
        )                                 /* End of instruction execution */
      )
    )
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
