/*------------------------------------------------------------------------*/
/*                                                                        */
/*                 AMULET3 Instruction Address Interface                  */
/*                                                                        */
/*                                             Created: JVW 13/12/96      */
/*                                       Last Modified: JDG   4/7/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/* Notes: Some state is kept in output bundle to minimise copying.        */
/*        "IRQ" & "FIQ" are global, Boolean variables set by "Environment"*/
/*        "Run" is a global variable set to false to halt execution       */
/*                                                                        */
/* Missing: Halt                                                          */
/*                                                                        */
/* "seq" - sequential access indicator a bit bodged                       */
/*                                                                        */
/* Interrupt handling somewhat bodged                                     */
/*                                                                        */
/* Designed to fetch always as 32 bit words (reflected in BTC design)     */
/*                                                                        */
/*------------------------------------------------------------------------*/

IAI(Branch     : var(chan(Branch_type)),
    to_Mem     : var(chan(Addr_bus_type)),
    to_MemCtrl : var(chan(MemCtrl_bus_type)),
    indirPC    : var(chan(Data_bus_type))):
    expr(void)=
  (
  MAR_type : type = record(PC::int,
                          col::int,
                     PSR_ctrl::Ctrl_bits_type,
                      IndirPC::bool,
                   Mem_bypass::bool,
                    extension::int,
                     Dabt_svc::bool,
                          seq::bool).

  BTC_CAM   : var(int^30).
  BTC_valid : var(bool^30).
  BTC_PC    : var(int^30).
  BTC_link  : var(bool^30).
  BTC_CC    : var(int^30).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  BTC_insert(Old_PC : val(int),
             New_PC : val(int),
             link   : val(bool),
             CC     : val(int),
             Thumb  : val(bool)):
             expr(void)=
    (
    BTC_ptr : var(int).

    BTC_ptr := rand(BTC_size);

    if Thumb then                       /* Select appropriate half of BTC */
      (if ((Old_PC and 2) = 0) then BTC_ptr := BTC_ptr and 16#FFFFFFFE
       else                         BTC_ptr := BTC_ptr or  16#00000001);

    BTC_CAM[BTC_ptr]   := Old_PC and 16#FFFFFFFC;
    BTC_valid[BTC_ptr] := true;
    BTC_PC[BTC_ptr]    := New_PC;
    BTC_link[BTC_ptr]  := link;
    BTC_CC[BTC_ptr]    := CC
    ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

    BTC_search(PC : val(int), Thumb : val(bool), BTC_index : var(int)):
               expr(int) =
      (
      local_PC   : var(int).
      i          : var(int).
      prediction : var(int).

      prediction := 16#00;          /* Bits 6,5: 00=none, x1=even, 10=odd */
      local_PC := PC and 16#FFFFFFFC;

      if (!Thumb) then
        forseq i in 0 to (BTC_size-1) do               /* ARM mode search */
          (
          if ((BTC_CAM[i] = local_PC) && BTC_valid[i]) then
            (BTC_index := i; prediction := 16#20)      /* ARM mode: found */
          )
      else
        (                                            /* Thumb mode search */
        index_e : var(int).
        index_o : var(int).

        forseq i in 0 to (BTC_size-1) do
          (
          if ((BTC_CAM[i] = local_PC) && BTC_valid[i]) then
            (                                   /* Test for even or odd ? */
            if ((i and 1) = 0) then             /* Thumb mode: found even */
              (index_e := i; prediction := prediction or 16#20)
            else                                /* Thumb mode: found odd  */
              (index_o := i; prediction := prediction or 16#40)
            )
          );

        if ((PC and 2) = 0) then                /* Both halves considered */
          BTC_index  := (if ((prediction and 16#20) != 0) then index_e
                         else index_o)
        else
          (                                   /* Only odd half considered */
          prediction := prediction and 16#5F; /* Ensure "even" bit clear */
          BTC_index  := index_o
          )
        );

      prediction
      ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

    BTC_what_branch(BTC_index : val(int)) : expr(int) =
      ((if BTC_link[BTC_index] then 16#10 else 16#00)
       or (BTC_CC[BTC_index] and 16#0F)).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

    BTC_read(BTC_index : val(int)) : expr(int) = BTC_PC[BTC_index].

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  Mem_Addr_Reg(to_MAR     : var(chan(MAR_type)),
               to_Mem     : var(chan(Addr_bus_type)),
               to_MemCtrl : var(chan(MemCtrl_bus_type))):
               expr(void) =
    (
    MAR       : var(MAR_type).
    MemBundle : var(Addr_bus_type).
    MCBundle  : var(MemCtrl_bus_type).

/* Initialise static or unused record elements                            */
    MemBundle->Read  := true;
    MemBundle->Write := false;
    MemBundle->Size  := 4;
    MemBundle->Dout  := 0;                                      /* Unused */

    forever
      (
      to_MAR?(MAR := (?to_MAR));

      wait_for(0);

      MCBundle->PC         := MAR->PC;
      MCBundle->col        := MAR->col;
      MCBundle->PSR_ctrl   := MAR->PSR_ctrl;
      MCBundle->IndirPC    := MAR->IndirPC;
      MCBundle->Mem_bypass := MAR->Mem_bypass;
      MCBundle->extension  := MAR->extension;
      MCBundle->Dabt_svc   := MAR->Dabt_svc;
      MCBundle->bv_colour  := next_bview_colour();

      MemBundle->addr  := MAR->PC and 16#FFFFFFFC;        /* Always words */
      MemBundle->mode  := MAR->PSR_ctrl->mode;
      MemBundle->Qcol  := MAR->col;       /* This one probably irrelevant */
      MemBundle->extra := (if MAR->IndirPC then 1 else 0);
      MemBundle->seq   := MAR->seq;

      (                                                /* Parallel output */
        if !(MAR->Mem_bypass) then 
          active_block("iai",MCBundle->bv_colour,(
            to_Mem!MemBundle
          ))
      |
        to_MemCtrl!MCBundle
      )
      )

    ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  to_MAR  : var(chan(MAR_type)).

  MAR : var(MAR_type).

  FIQ_svc : var(bool).
  IRQ_svc : var(bool).

  Loaded_PC : var(Data_bus_type).

  BTC_prediction : var(int).        /* Bits 6,5: 00=none, x1=even, 10=odd */
  BTC_index      : var(int).

  init_chan(to_MAR, "MAR Input");

/* Initialise BTC                                                         */
  (i : var(int). forseq i in 0 to (BTC_size-1) do BTC_valid[i] := false);

  MAR->PC               := 0;                       /* Initialise PC etc. */
  MAR->col              := 0;
  MAR->PSR_ctrl->IRQ_En := false;                  /* Interrupts disabled */
  MAR->PSR_ctrl->FIQ_En := false;
  MAR->PSR_ctrl->Thumb  := false;                        /* Full ARM mode */
  MAR->PSR_ctrl->mode   := 3;               /* Reset into supervisor mode */
  MAR->seq              := false;                /* Non-sequential access */

  FIQ_svc := false;
  IRQ_svc := false;

  wait_until(Run);                  /* At least have a chance of starting */

  (                                                 /* Parallel processes */
  Mem_Addr_Reg(to_MAR, to_Mem, to_MemCtrl) |      /* Parallel MAR process */

/* forever */
  while Run do
    (
    watchdog := now;                                     /* Keep time log */
    wait_for(IAI_delay);

    MAR->IndirPC    := false;        /* No indirection unless "branch"ing */
    MAR->Mem_bypass := false;
    MAR->extension  := 0;                            /* No extension code */
    MAR->Dabt_svc   := false;

    if chan_ready(Branch) then               /* See if a branch is wanted */
      (Branch?                                             /* Yes - do it */
        (
        MAR->col  := (?Branch)->New_col;                    /* Regardless */
        MAR->seq := false;                       /* Non-sequential access */

        if ((?Branch)->Exception != 0) then
          (                        /* Exception handler - code is address */
            /* Possibilities are Undef (4), SWI (8), Pabt (12), Dabt (16) */
          MAR->PC       := (?Branch)->Exception;         /* Branch to ... */
          MAR->PSR_ctrl := (?Branch)->New_PSR_ctrl;

/* Checking routines */
          if (MAR->PSR_ctrl->IRQ_En) then
            printstr("Exception entry - IRQ still enabled\n");

          (case ((?Branch)->Exception)
            when  4 => if (MAR->PSR_ctrl->mode != 11) then
                         printstr("Undefined entry - wrong mode\n")
            when  8 => if (MAR->PSR_ctrl->mode !=  3) then
                         printstr("SWI entry - wrong mode\n")
            when 12 => if (MAR->PSR_ctrl->mode !=  7) then
                         printstr("Pabt entry - wrong mode\n")
            when 16 => if (MAR->PSR_ctrl->mode !=  7) then
                         printstr("Dabt entry - wrong mode\n")
            default printstr("Bad exception entry\n")
          );                                  /* End of confidence checks */

          if ((?Branch)->Exception = 16) then
            (
            MAR->Dabt_svc  := true
            )

          )
        else
          (
          MAR->PC       := (?Branch)->New_PC;
          MAR->IndirPC  := (?Branch)->Indir_req;
          MAR->PSR_ctrl := (?Branch)->New_PSR_ctrl;
/* printstr("Branched to ");printhex(MAR->PC);printstr("\n"); */

          if ((?Branch)->Cacheable && BTC_Enable) then   /* New BTC entry */
            BTC_insert((?Branch)->Old_PC,
                       (?Branch)->New_PC,
                       (?Branch)->link,
                       (?Branch)->CC,
                       (?Branch)->New_PSR_ctrl->Thumb)
          )
        );
      FIQ_svc := false;   /* Override any interrupt operation in progress */
      IRQ_svc := false
      );         /* NB Branch "ack"ed before address dispatched - WRONG ? */



/*Insert interrupt service address (currently overwites incremented value)*/
/*  Hangover from previous cycle                                          */
    if FIQ_svc then (MAR->PC := 28; FIQ_svc := false);
    if IRQ_svc then (MAR->PC := 24; IRQ_svc := false);

    if !(MAR->IndirPC) then          /* PC load is non-interruptable */
      /* Test for interrupts                                              */
      (
      if (MAR->PSR_ctrl->FIQ_En && FIQ) then
        (
        FIQ_svc := true;
        MAR->extension        := 16#19;
        MAR->PSR_ctrl->IRQ_En := false;
        MAR->PSR_ctrl->FIQ_En := false;
        MAR->PSR_ctrl->Thumb  := false;
        MAR->PSR_ctrl->mode   := 1
        )
      else
        (
        if (MAR->PSR_ctrl->IRQ_En && IRQ) then
          (
          IRQ_svc := true;
          MAR->extension        := 16#1A;
          MAR->PSR_ctrl->IRQ_En := false;
          MAR->PSR_ctrl->Thumb  := false;
          MAR->PSR_ctrl->mode   := 2
          )
        )
      );


/*  Bypass for any unusual case except "entry" to Dabt                    */
    if (MAR->extension != 0) then MAR->Mem_bypass := true;

/*
    printstr("IAI: MAR = "); printhex(MAR->PC); printstr("  Colour ");
    printint(MAR->col); printstr("\n");
*/
    if (MAR->IndirPC) then wait_until(STR_semaphore = 0);
                             /* Wait for stores to complete if loading PC */

    if (BTC_Enable && (MAR->extension = 0)) then              /* CHECK !!!*/
      BTC_prediction := BTC_search(MAR->PC,                  /* Check BTC */
                                   MAR->PSR_ctrl->Thumb,
                                   BTC_index)
    else
      BTC_prediction := 16#00;

    if (BTC_prediction != 0) then               /* Get branch instruction */
      (
      MAR->extension  := BTC_prediction or BTC_what_branch(BTC_index);
      MAR->Mem_bypass := ((BTC_prediction and 16#20) != 0)
                         /* Only if "even" prediction - includes ARM mode */
      );

    to_MAR!MAR;                                                 /* Output */

    if (MAR->IndirPC) then                   /* Did we request a PC load? */
      (                                        /* Yes - wait for response */
      indirPC?(Loaded_PC:=(?indirPC));  /* Input from indirection channel */
                                        /* Early ack. - should not matter */
      if (Loaded_PC->abort) then
        (                          
        while (wait_until(chan_ready(Branch));
               !((?Branch)->Exception = 16)) do   /* Data abort = code 16 */
             Branch?null            /* Spurious branch discard and repeat */
                                   /* Shouldn't happen - remove when sure */
        )                                   /* End of indirect load stuff */
      else
        MAR->PC := Loaded_PC->data                     /* Adopt loaded PC */
      )
    else
      (
      if (BTC_prediction != 0) then
        (
        MAR->seq := false;                       /* Non-sequential access */
        MAR->PC  := BTC_read(BTC_index)       /* Predicted branch address */
        )
      else
        (
        MAR->seq := true;                            /* Sequential access */
        MAR->PC  := ((MAR->PC) + 4) and 16#FFFFFFFC
        )                                  /* PC increment and word align */
      )
    )                                                      /* End of loop */

  )                                                 /* Parallel processes */
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
