/*------------------------------------------------------------------------*/
/*                                                                        */
/*                           AMULET3 Header File                          */
/*                                                                        */
/*                                             Created: JVW 13/12/96      */
/*                                       Last Modified: JDG  25/9/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/

/* Use files                                                              */

/*------------------------------------------------------------------------*/
/* Record type declarations                                               */

Ctrl_bits_type : type = record(IRQ_En::bool,
                               FIQ_En::bool,
                                Thumb::bool,
                                 mode::int).

PSR_type : type = record(flags::int,
                          ctrl::Ctrl_bits_type).

Branch_type : type = record(New_PC::int,
                           New_col::int,
                      New_PSR_ctrl::Ctrl_bits_type,
                         Indir_req::bool,
                         Exception::int,
                         Cacheable::bool,
                            Old_PC::int,
                              link::bool,
                                CC::int,
                             flush::bool,
                              halt::bool).

MemCtrl_bus_type : type = record( PC::int,
                                 col::int,
                            PSR_ctrl::Ctrl_bits_type,
                             IndirPC::bool,
                          Mem_bypass::bool,
                           extension::int,
                            Dabt_svc::bool,
                           bv_colour::int). /* for block view display */

Instr_pkt_type : type = record(Op_code::int,
                             extension::int,
                              Dabt_svc::bool,
                              PSR_ctrl::Ctrl_bits_type,
                                colour::int,
                                    PC::int,
                             bv_colour::int). /* for block view display */

Instr_pkt_type2 : type = record(Op_code::int,
                              extension::int,
                                    LSB::int, /* for odd address branches */
                                special::bool,            /* for Thumb BL */
                               Dabt_svc::bool,
                               PSR_ctrl::Ctrl_bits_type,
                                 colour::int,
                                     PC::int,
                              bv_colour::int). /* for block view display */

Branch_dec_type : type = record(branch::bool,
                             predicted::bool,
                                  link::bool,
                                    BX::bool,
                              Thumb_BL::bool).

Reg_usage_type : type = record(uses_A::bool,
                               uses_B::bool,
                               uses_C::bool,
                             uses_Imm::bool).

Odd_Functions_type : type = record(LSM::bool,
                                   MSR::bool,
                                   MRS::bool).

Operation_type :  type = record(shift::bool,            /* Shift required */
                             shift_op::int,            /* Shift operation */
                            shift_imm::bool,  /* Shift by immediate value */
                         shift_amount::int,/* Immediate value to shift by */
                              use_MUL::bool,   /* Multiplication required */
                              use_ALU::bool,
                               ALU_op::int,
                            set_flags::bool).

Immediate_type :  type = record(Immediate::int,
                                default_C::bool,
                                    Imm_C::int).

Decoded_instr_type : type = record(condition::int,
                                     Op_code::int, /* Mostly for disassembler */
                                   extension::int, /* For disassembler only */
                                   Exception::int,
                                   Immediate::Immediate_type,
                                   operation::Operation_type,
                                    PSR_ctrl::Ctrl_bits_type,
                                      colour::int,
                                   reg_usage::Reg_usage_type,
                                     dest_PC::bool,
                                       LD_PC::bool,
                                       store::bool,
                                    to_queue::bool,
                                  queue_addr::int,
                                      branch::Branch_dec_type,
                                      to_DAI::bool,
                                     to_SPSR::bool,
                                       flags::Odd_Functions_type,
                                          PC::int,
                                addr_pre_inc::bool,   /* indirectly to DAI */
                                   bv_colour::int).  /* for block view */

DAI_instr_type : type = record(read::bool,
                              write::bool,
                               size::int,
                                sex::bool,
                           get_addr::bool,     /* lose */
                         force_user::bool,
                                LSM::bool,
                              copro::Copro_type,
                            pre_inc::bool,     /* lose */
                         last_cycle::bool,
                         queue_addr::int,
                          bv_colour::int).

DAI_addr_pkt_type : type = record(addr::int,
                                  mode::int,
                                colour::int,
                            invalidate::bool,
                               pre_inc::bool).

Queue_input_type : type = record(data::int,
                           queue_addr::int,
                              invalid::bool,
                                abort::bool,
                               has_BR::bool,
                            bv_colour::int). /* for block view */


Queue_output_type : type = record(data::int,
                              reg_addr::int,
                                  mode::int,
                             bv_colour::int).

Xpipe_type : type = record(CPSR_flag::bool,
                                  PC::int,
                                 PSR::PSR_type).

BRpipe_type : type = record(base::int,
                             reg::int,
                            mode::int).

Reg_addr_type : type = record(addr::int,
                              mode::int).

Forward_req_type : type = record(vector::bool^10,
                                  start::int,
                                   Fcol::int,
                                 colour::int).

Forward_data_type : type = record(data::int,
                                 found::bool).


/*------------------------------------------------------------------------*/
/* Global variables                                                       */

STR_semaphore : var(int).
                     /* Prevents loading PC whilst stores are outstanding */

/*------------------------------------------------------------------------*/
/* Key to some encoded types                                              */
/*                                                                        */
/* Op-code extensions: 00 - none                                          */
/*                     19 - FIQ                                           */
/*                     1A - IRQ                                           */
/*                     1F - Pabt                                          */
/*                     15 - Undefined (e.g. no such coprocessor)          */
/*                  20-7F - Predicted branches                            */
/*                                                                        */
/* Modes:  ARM's code    Internal code                                    */
/*         0 - user            0                                          */
/*         1 - FIQ             1                                          */
/*         2 - IRQ             2                                          */
/*         3 - supervisor      3                                          */
/*         7 - abort           7                                          */
/*         B - undef           5                                          */
/*         F - system          4                                          */
/*                                                                        */
/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/
/* "Constants"  (actually global variables set at top level)              */

Mem_ctrl_len  : var(int).   /* Must be _shorter_ than Imem pipeline depth */
Queue_size    : var(int).                              /* Min. 2, Max. 10 */
Prefetch_size : var(int).                              /* Min. 0, Max. ?? */
DAI_Ibuf_size : var(int).                              /* Min. 0, Max. ?? */
BTC_size      : var(int).                 /* Even number, Min. 2, Max. 20 */

/*------------------------------------------------------------------------*/
/*                                                                        */
/* Timing parameters for processor core                                   */
/*                                                                        */

        IAI_delay : var(int).
   Mem_ctrl_delay : var(int).
      In_IF_delay : var(int).
In_IF_Indir_delay : var(int).
      Ipipe_delay : var(int).
     REG_rd_delay : var(int).
     REG_wr_delay : var(int).
   Thumb_on_delay : var(int).
  Thumb_off_delay : var(int).
  Copro_dec_delay : var(int).
        DEC_delay : var(int).
  DEC_trash_delay : var(int).
  DEC_multi_delay : var(int).
  DEC_alloc_delay : var(int).
       EXEC_delay : var(int).
  EXEC_nogo_delay : var(int).
 EXEC_abort_delay : var(int).
EXEC_branch_delay : var(int).
      Shift_delay : var(int).
  Shift_byp_delay : var(int).
ALU_logical_delay : var(int).
  ALU_arith_delay : var(int).
     Br_Lat_delay : var(int).
     BRpipe_delay : var(int).
        DAI_delay : var(int).
        Qin_delay : var(int).
       Qout_delay : var(int).
       Qfwd_delay : var(int).
  Qfwd_cand_delay : var(int).
   Qfwd_byp_delay : var(int).
     Qtoken_delay : var(int).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
