/*------------------------------------------------------------------------*/
/*                                                                        */
/*                      AMULET3 Instruction Input Stage                   */
/*                                                                        */
/*                                             Created: JVW 13/12/96      */
/*                                       Last Modified: JDG  24/7/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/

/* This "stage" does not latch data */

/* Structure may be improvable - output occurs in one of two places       */
/* NB. be careful not to acknowledge input before output complete         */

/*------------------------------------------------------------------------*/

IinfBlock(Op_code   : var(chan(Data_bus_type)),
          Ctrl_info : var(chan(MemCtrl_bus_type)),
          Inst      : var(chan(Instr_pkt_type)),
          IndirPC   : var(chan(Data_bus_type))):
          expr(void) =
  (

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  Iinf_output(Instr_pkt : var(Instr_pkt_type),
              Inst      : var(chan(Instr_pkt_type))):
              expr(void) =
    (
    wait_for(In_IF_delay);

               /* Pre-decode unit - gates in exception/prediction classes */
    (case (Instr_pkt->extension and 16#30)
      when 16#00 => null
      when 16#10 =>
        Instr_pkt->Op_code := (Instr_pkt->Op_code and 16#00FFFFFF) or 16#EF000000
      when 16#20 =>
        Instr_pkt->Op_code := (Instr_pkt->Op_code and 16#00FFFFFF) or 16#0A000000
                          or ((Instr_pkt->extension and 16#F) << 28)
      when 16#30 =>
        Instr_pkt->Op_code := (Instr_pkt->Op_code and 16#00FFFFFF) or 16#0B000000
                          or ((Instr_pkt->extension and 16#F) << 28)
/* NB in Thumb mode in the latter two cases the _second_ instruction will */
/*   be corrupted, but so what?                                           */
    );

    Inst!Instr_pkt                                              /* Output */
    ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  Instr_pkt : var(Instr_pkt_type).

  forever
    (
    Ctrl_info?            /* First wait for control information to arrive */
      (
      if ((?Ctrl_info)->IndirPC) then                /* Look for PC loads */
        (
        temp : var(Data_bus_type).

        wait_for(In_IF_Indir_delay);
        Op_code?                             /* Output to PC load channel */
          (
            (case ((?Ctrl_info)->PC and 3)   /* Rotate if misaligned load */
              when 0 => temp->data := (?Op_code)->data
              when 1 =>
                temp->data := ((((?Op_code)->data >>  8) and 16#00FFFFFF)
                            or (((?Op_code)->data << 24) and 16#FF000000))
              when 2 =>
                temp->data := ((((?Op_code)->data >> 16) and 16#0000FFFF)
                            or (((?Op_code)->data << 16) and 16#FFFF0000))
              when 3 =>
                temp->data := ((((?Op_code)->data >> 24) and 16#000000FF)
                            or (((?Op_code)->data <<  8) and 16#FFFFFF00))
            );
          temp->abort := (?Op_code)->abort;
          temp->extra := 0;
          IndirPC!temp
          )                                       /* Acknowledge op. code */
        )
      else
        (                                  /* Assemble instruction packet */
        wait_for(In_IF_delay);

        Instr_pkt->colour    := (?Ctrl_info)->col;
        Instr_pkt->PSR_ctrl  := (?Ctrl_info)->PSR_ctrl;
        if (?Ctrl_info)->PSR_ctrl->Thumb then
          Instr_pkt->PC      := (?Ctrl_info)->PC and 16#FFFFFFFE
        else
          Instr_pkt->PC      := (?Ctrl_info)->PC and 16#FFFFFFFC;
        Instr_pkt->extension := (?Ctrl_info)->extension;
        Instr_pkt->Dabt_svc  := (?Ctrl_info)->Dabt_svc;
        Instr_pkt->bv_colour := (?Ctrl_info)->bv_colour;

        if (!(?Ctrl_info)->Mem_bypass) then
          (Op_code?                    /* Read op. code (unless bypassed) */
            active_block("imem",Instr_pkt->bv_colour,(
            (
            Instr_pkt->Op_code := (?Op_code)->data;

              /* NB Cannot abort if FIQ/IRQ/Bpred because memory bypassed */
            if ((?Op_code)->abort) then
              (                                    /* Prefetch abort code */
              Instr_pkt->extension       := 16#1F;/*Partly redundant soon */
              Instr_pkt->PSR_ctrl->Thumb := false;/*Disable Thumb decoder */
              Instr_pkt->PSR_ctrl->mode  := 7     /* Change to abort mode */
              );

            Iinf_output(Instr_pkt, Inst)                     /* Output ...*/
            )
            ))
          )                                 /* ... and acknowledge memory */
        else
          (
          Iinf_output(Instr_pkt, Inst)             /* or else just output */
          )
        )                                     /* Acknowledge control pipe */
      )
    )
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
