/*------------------------------------------------------------------------*/
/*                                                                        */
/*                          AMULET3 Dual port Memory                      */
/*                                                                        */
/*                                             Created: JDG 11/12/96      */
/*                                       Last Modified: JDG  23/5/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/

freadfile(filename     : val(string),
          memory_array : var(int^40000)):
          expr(void) =
  (
  handle : var(file).
  number : var(int).

  i : var(int).

  handle := openin(filename);
  i := 0;
  while (!eof(handle)) do
    (
    number := freadint(handle);
    memory_array[i] := number;
    i := i + 1
    );

  printstr("Memory loaded\n")
  ).

/*------------------------------------------------------------------------*/

Memory(Iaddr : var(chan(Addr_bus_type)),
       op_code : var(chan(Data_bus_type)),
       Daddr   : var(chan(Addr_bus_type)),
       Dout    : var(chan(Data_bus_type))):
       expr(void) =
  (
  memory_array : var(int^40000).
  i            : var(int).
  block_lock   : var(bool^16).


  Mem_port(Addr : var(chan(Addr_bus_type)),
           Data : var(chan(Data_bus_type))):
           expr(void) =
    (
    Qcol       : var(int).
    Aborted    : var(bool).
    Block_addr : var(int).

    data_value : var(Data_bus_type).
    mem_addr   : var(int).
    byte_sel   : var(int).

    Aborted := false;

    forever
      (
      Addr?                                           /* Wait for request */
        (
        mem_addr := (?Addr)->addr>>2;
        byte_sel := ((?Addr)->addr) and 3;

        block_addr := (mem_addr >> 2) and mem_addr_mask;  /* 4 words/line */
        if (!dual_port) then
          (
          wait_until(!block_lock[block_addr]);  /* Wait if block conflict */
          block_lock[block_addr] := true          /* Signal block entered */
          );

        wait_for(MEM_delay);                              /* Memory delay */

        data_value->extra := (?Addr)->extra;     /* Auxilliary data field */

        if (Aborted && (Qcol = (?Addr)->Qcol)) then
          (                                  /* true _following_ an abort */
          data_value->abort := true
          )                  /* Prevent further action until colour match */
        else
          (
          Qcol    := (?Addr)->Qcol;               /* Adopt current colour */

          data_value->data  := 0;                    /* Just some default */

if ((?Addr)->addr = Tube_addr) then
  (                                                            /* for SWI */
  data_value->abort := false;
  if (((?Addr)->Dout and 16#7F) = 4) then (printstr("Stopping\n"); Run := false)
  else
    (
    if (((?Addr)->Dout and 16#7F) = 5) then
      (printstr("{{Time is "); printint(now); printstr("}}\n"))
    else
      ASCII_out((?Addr)->Dout)
    )
  )
else
    (
          if ((mem_addr < 0) || (mem_addr >= 40000)) then/* Memory management*/
            (
            data_value->abort := true
            )
          else
            (
            temp : var(int).

            data_value->abort := false;
            if ((?Addr)->Read) then data_value->data := memory_array[mem_addr];
            if ((?Addr)->Write) then
              (case (?Addr)->Size
                when 4 => memory_array[mem_addr] := (?Addr)->Dout
                when 1 =>
                  (case byte_sel
                    when 0 =>
                      (
                      memory_array[mem_addr] := memory_array[mem_addr] and 16#FFFFFF00;
                      temp                   := (?Addr)->Dout          and 16#000000FF;
                      memory_array[mem_addr] := memory_array[mem_addr] or temp
                      )

                    when 1 =>
                      (
                      memory_array[mem_addr] := memory_array[mem_addr] and 16#FFFF00FF;
                      temp                   := (?Addr)->Dout          and 16#0000FF00;
                      memory_array[mem_addr] := memory_array[mem_addr] or temp
                      )

                    when 2 =>
                      (
                      memory_array[mem_addr] := memory_array[mem_addr] and 16#FF00FFFF;
                      temp                   := (?Addr)->Dout          and 16#00FF0000;
                      memory_array[mem_addr] := memory_array[mem_addr] or temp
                      )

                    when 3 =>
                      (
                      memory_array[mem_addr] := memory_array[mem_addr] and 16#00FFFFFF;
                      temp                   := (?Addr)->Dout          and 16#FF000000;
                      memory_array[mem_addr] := memory_array[mem_addr] or temp
                      )

                  )

                when 2 =>
                  (case (byte_sel and 2)
                    when 0 =>
                      (
                      memory_array[mem_addr] := memory_array[mem_addr] and 16#FFFF0000;
                      temp                   := (?Addr)->Dout          and 16#0000FFFF;
                      memory_array[mem_addr] := memory_array[mem_addr] or temp
                      )
                    when 2 =>
                      (
                      memory_array[mem_addr] := memory_array[mem_addr] and 16#0000FFFF;
                      temp                   := (?Addr)->Dout          and 16#FFFF0000;
                      memory_array[mem_addr] := memory_array[mem_addr] or temp
                      )
                  )

                default
                  (
                  printint((?Addr)->Size);
                  printstr(" is not a legal transfer size\n")
                  )
              )
            );
null);
          Aborted := data_value->abort
          );

        Data!data_value;

        block_lock[block_addr] := false            /* Signal block exited */
        )                                            /* Acknowledge input */
      )
    ).


  printstr("Dual port memory model\n");

  freadfile("programme", memory_array);                 /* Initialisation */
  (i : var(int). for i in 0 to 15 do block_lock[i] := false);    /* Init. */

  (Mem_port(Iaddr, op_code) | Mem_port(Daddr, Dout))
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
