/*------------------------------------------------------------------------*/
/*                                                                        */
/*                AMULET3 Memory/Coprocessor Support Interfaces           */
/*                                                                        */
/*                                             Created: JDG   7/5/97      */
/*                                       Last Modified: JDG   5/6/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/

Memory_support(from_CPU_I    : var(chan(Addr_bus_type)),
               to_CPU_I      : var(chan(Data_bus_type)),
               from_CPU_D    : var(chan(Mem_copro_bus_type)),
               to_CPU_D      : var(chan(Data_bus_type))):
               expr(void) =
  (
  Snoop_Op_code : var(chan(Data_bus_type))./* Returned value from Imemory */
  for_copro     : var(chan(int)). /* Instruction tapped to coprocessor(s) */
  copro_instr_out : var(chan(int)).
  copro_dec_out : var(chan(copro_instr_type)).
  Dmem_addr     : var(chan(Addr_bus_type)).        /* Data memory address */
  Dmem_data     : var(chan(Data_bus_type)).           /* Data memory data */
  to_copro      : var(chan(Data_bus_type)).

/*------------------------------------------------------------------------*/
  /* Monitor the instruction bus for coprocessor instructions             */
  /* Will eventually dispatch these to coprocessor unit                   */

  Snoop_instr(from_mem  : var(chan(Data_bus_type)),
              to_CPU    : var(chan(Data_bus_type)),
              for_copro : var(chan(int))):
              expr(void) =
    (
    forever
      (
      from_mem?
        (
        if (((?from_mem)->extra = 0)
          && ((((?from_mem)->data and 16#0E000000) = 16#0C000000)
           || (((?from_mem)->data and 16#0F000000) = 16#0E000000))) then
          (
/*
          printstr("Coprocessor instruction ");
          printhex((?from_mem)->data);
          printstr(" snooped\n");
*/
          for_copro!((?from_mem)->data)
          );

        to_CPU!(?from_mem)
        )                                    /* ... and acknowledge input */
      )
    ).


/*------------------------------------------------------------------------*/
/* Much simplified coprocessor decoder/execution unit                     */
/* Needs to handshake with CPU decode to accept/reject instruction        */

    Copro_dec(copro_instr_out : var(chan(int)),
              copro_issue     : var(chan(copro_instr_type))):
              expr(void) =
      (
      instr  : var(int).

      forever
        (
        copro_instr_out?(instr := (?copro_instr_out));
        if copro_test(instr) then copro_issue!copro_decode(instr)
        )                           /* Else (if not recognised) evaporate */
      ).

/*------------------------------------------------------------------------*/
/* Data memory support block                                              */

  Data_memory_support(from_CPU        : var(chan(Mem_copro_bus_type)),
                      copro_instr_out : var(chan(copro_instr_type)),
                      to_memory       : var(chan(Addr_bus_type)),
                      from_memory     : var(chan(Data_bus_type)),
                      to_CPU          : var(chan(Data_bus_type)),
                      to_copro        : var(chan(Data_bus_type))):
                      expr(void) =
    (
    copro_ctrl_pipe_type : type = record(copro::bool,    /* Copro. instr. */
                                        colour::int,
                                       invalid::bool,
                                      dest_CPU::bool,
                                    dest_copro::bool,
                                    copro_addr::int).

    ctrl_FIFO_in    : var(chan(copro_ctrl_pipe_type)).
    ctrl_FIFO_out   : var(chan(copro_ctrl_pipe_type)).
    copro_FIFO_in   : var(chan(int)).
    copro_FIFO_out  : var(chan(int)).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
    /* Process data memory/coprocessor outputs from the processor         */
    /* Synchronise or route to units as appropriate                       */
    /* Note LSC/STC should produce a response for the DAI to control the  */
    /* address stream/termination                                         */


    Data_out_int(from_CPU   : var(chan(Mem_copro_bus_type)),
                 from_copro : var(chan(copro_instr_type)),
                 to_ctrl    : var(chan(copro_ctrl_pipe_type)),
                 to_memory  : var(chan(Addr_bus_type)),
                 to_copro   : var(chan(int))):
                 expr(void) =
      (
      ctrl_output  : var(copro_ctrl_pipe_type).
      mem_output   : var(Addr_bus_type).
      copro_output : var(int).

      forever
        (
        from_CPU?
          (                         /* First - set up control information */
          ctrl_output->colour  := (?from_CPU)->Qcol;
          ctrl_output->invalid := (?from_CPU)->invalidate;

          if ((?from_CPU)->copro->use_copro) then
            (                                    /* Coprocessor operation */
            if (((?from_CPU)->copro->instr and 16#02000000) = 0) then
              (                                                    /* LSC */
              ctrl_output->copro      := ((?from_CPU)->invalidate);
                  /* Memory, not copro. pipe - unless operation cancelled */
                  /* NB it is important to avoid sending duff (abortable) */
                     /* addresses when they are not in valid instructions */
              ctrl_output->dest_copro :=
                  (((?from_CPU)->copro->instr and 16#00100000) != 0);/*LDC*/
              ctrl_output->dest_CPU   := true/* Needs to know abort status*/
              )
            else
              (
              ctrl_output->copro := true;       /* Use copro. result pipe */
              if (((?from_CPU)->copro->instr and 16#00100010) = 16#00100010) then
                (ctrl_output->dest_copro := false; ctrl_output->dest_CPU := true)
                                                                   /* MRC */
              else
                (ctrl_output->dest_copro := true;  ctrl_output->dest_CPU := false)
              )                                             /* MCR or CDP */
            )
          else                                    /* CPU memory operation */
            (
            ctrl_output->copro      := false;  /* Memory, not copro. pipe */
            ctrl_output->dest_copro := false;
            ctrl_output->dest_CPU   := true
            );

ctrl_output->dest_CPU := true;   /* HACK SO ALL INSTRUCTIONS RETURN TOKEN */

          ctrl_output->dest_CPU := ctrl_output->dest_CPU && (!ctrl_output->invalid);
                                /* CPU has "skipped" if operation invalid */

          /* Second - rendezvous with coprocessor instruction if necessary*/
          if ((?from_CPU)->copro->use_copro) then
            wait_until(chan_ready(from_copro));

          ctrl_output->copro_addr := (?from_copro)->dest;
                                                  /* Destination (here??) */

          /* Third - set up `data' in relevant pipe                       */
          if (ctrl_output->copro) then
            (
            if (((?from_CPU)->copro->instr and 16#00100010) = 16#00000010) then
              copro_output := (?from_CPU)->Dout                    /* MCR */
            else
              copro_output := (?from_copro)->value          /* MRC or CDP */
            )
          else
            (
            mem_output->addr  := (?from_CPU)->addr;
            mem_output->seq   := (?from_CPU)->seq;
            mem_output->mode  := (?from_CPU)->mode;
            mem_output->Qcol  := (?from_CPU)->Qcol;
            mem_output->Read  := (?from_CPU)->Read;
            mem_output->Write := (?from_CPU)->Write;
            mem_output->Size  := (?from_CPU)->Size;
            if ((?from_CPU)->copro->use_copro) then
              mem_output->Dout := (?from_copro)->value
            else
              mem_output->Dout := (?from_CPU)->Dout;
            mem_output->extra := 0          /* Not used in data interface */
            );


          (                                     /* Last - parallel output */
            if (ctrl_output->dest_CPU || ctrl_output->dest_copro) then
              (             /* NB must be a destination if memory is used */
                to_ctrl!ctrl_output
              |
                if (ctrl_output->copro) then to_copro!copro_output
                else                         to_memory!mem_output
              |
                if (((?from_CPU)->copro->use_copro)    /* Do side effects */
                && (!(?from_CPU)->invalidate)) then copro_execute(?from_copro)
              )
          |
            if ((?from_CPU)->copro->use_copro) then from_copro?null
          )                              /* Acknowledge coprocessor input */

          )                                          /* Acknowledge input */
        )
      ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
    /* Route data from data memory/coprocessor to appropriate destination */

    Data_in_int(from_ctrl   : var(chan(copro_ctrl_pipe_type)),
                from_memory : var(chan(Data_bus_type)),
                from_copro  : var(chan(int)),
                to_CPU      : var(chan(Data_bus_type)),
                to_copro    : var(chan(Data_bus_type))):
                expr(void) =
      (
      Scratch_data : var(Data_bus_type).
      Aborted : var(bool).
      Abt_col : var(int).

      Scratch_data->extra := 0;                           /* Unused field */
      Aborted := false;

      forever
        (
        from_ctrl?
          (
          if ((?from_ctrl)->copro) then
            (
            wait_until(chan_ready(from_copro));
            Scratch_data->data  := (?from_copro);
            Scratch_data->abort := false                       /* Default */
            )
          else
            (
            wait_until(chan_ready(from_memory));
            Scratch_data := (?from_memory)
            );


          if (Aborted && ((?from_ctrl)->colour = Abt_col)) then
            Scratch_data->abort := true            /* Inside abort shadow */
          else
            (
            Abt_col := (?from_ctrl)->colour;       /* Colour from control */
            Aborted := (!(?from_ctrl)->copro) && (?from_memory)->abort
            );


          (                                           /* Parallel output */
            if ((?from_ctrl)->dest_CPU) then to_CPU!Scratch_data
          |
            if ((?from_ctrl)->dest_copro) then
              (
              Scratchier_data : var(Data_bus_type).

              Scratchier_data := Scratch_data;
              Scratchier_data->extra := (?from_ctrl)->copro_addr;
                           /* Naughty reuse of field "because it's there" */

              if ((?from_ctrl)->invalid) then Scratchier_data->abort := true;
              to_copro!Scratchier_data
              )                  /* (Uses abort to signal result invalid) */
          );

          if ((?from_ctrl)->copro) then from_copro?null else from_memory?null
                                            /* Acknowledge input channels */


          )                          /* ... and acknowledge "from_memory" */
        )
      ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

    init_chan(ctrl_FIFO_in, "ctrl_FIFO_in");
    init_chan(ctrl_FIFO_out, "ctrl_FIFO_out");
    init_chan(copro_FIFO_in, "copro_FIFO_in");
    init_chan(copro_FIFO_out, "copro_FIFO_out");

    (
      Data_out_int(from_CPU, copro_instr_out, ctrl_FIFO_in, to_memory, copro_FIFO_in)
    |
      FIFO(ctrl_FIFO_in, ctrl_FIFO_out, 2, 10, false, "Ext Mem Ctrl: ")
    |
      FIFO(copro_FIFO_in, copro_FIFO_out, 2, 10, false, "Copro FIFO: ")
    |
      Data_in_int(ctrl_FIFO_out, from_memory, copro_FIFO_out, to_CPU, to_copro)
    )

    ).

/*------------------------------------------------------------------------*/
/* Memory initialisation                                                  */

  dual_port     := true;                  /* "true" for dual port memory */
  mem_addr_mask := 0;                             /* 2^N-1 for 2^N blocks */

  MEM_delay     := 100;


  init_chan(Snoop_Op_code,   "Snoop Op_code");
  init_chan(for_copro,       "Cp Instr In");
  init_chan(copro_instr_out, "Cp Instr Out");
  init_chan(copro_dec_out,   "from copro");
  init_chan(Dmem_addr,       "Data address");
  init_chan(Dmem_data,       "Data data");
  init_chan(to_copro,        "Copro. result");

  copro_reg_lock := 0;                                  /* Bit of a bodge */


  (
    Memory(from_CPU_I, Snoop_Op_code, Dmem_addr, Dmem_data)
  |
    Snoop_instr(Snoop_Op_code, to_CPU_I, for_copro)
  |
    FIFO(for_copro, copro_instr_out, 3, 10, false, "Copro. prefetch: ")
  |
    Copro_dec(copro_instr_out, copro_dec_out)
                          /* Handshaking with instruction decoder missing */
  |
    Data_memory_support(from_CPU_D, copro_dec_out, Dmem_addr,
                        Dmem_data,  to_CPU_D,      to_copro)
  |
    copro_write(to_copro)
  )

  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
