/*------------------------------------------------------------------------*/
/*                                                                        */
/*                 AMULET3 Instruction Address Interface                  */
/*                                                                        */
/*                                             Created: JVW 13/12/96      */
/*                                       Last Modified: JDG  20/5/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/* Notes: Some state is kept in output bundle to minimise copying.        */
/*        "IRQ" & "FIQ" are global, Boolean variables set by "Environment"*/
/*        "Run" is a global variable set to false to halt execution       */
/*                                                                        */
/* Missing: Halt                                                          */
/*                                                                        */
/* "seq" - sequential access indicator a bit bodged                       */
/*                                                                        */
/* Interrupt handling somewhat bodged                                     */
/*                                                                        */
/* Designed to fetch always as 32 bit words (reflected in BTC design)     */
/*                                                                        */
/*------------------------------------------------------------------------*/

IAI(Branch     : var(chan(Branch_type)),
    to_Mem     : var(chan(Addr_bus_type)),
    to_MemCtrl : var(chan(MemCtrl_bus_type)),
    indirPC    : var(chan(Data_bus_type))):
    expr(void)=
  (

  BTC_CAM   : var(int^20).
  BTC_valid : var(bool^20).
  BTC_PC    : var(int^20).
  BTC_link  : var(bool^20).
  BTC_CC    : var(int^20).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  BTC_insert(Old_PC : val(int),
             New_PC : val(int),
             link   : val(bool),
             CC     : val(int),
             Thumb  : val(bool)):
             expr(void)=
    (
    BTC_ptr : var(int).

    BTC_ptr := rand(BTC_size);

    if Thumb then                       /* Select appropriate half of BTC */
      (if ((Old_PC and 2) = 0) then BTC_ptr := BTC_ptr and 16#FFFFFFFE
       else                         BTC_ptr := BTC_ptr or  16#00000001);

    BTC_CAM[BTC_ptr]   := Old_PC and 16#FFFFFFFC;
    BTC_valid[BTC_ptr] := true;
    BTC_PC[BTC_ptr]    := New_PC;
    BTC_link[BTC_ptr]  := link;
    BTC_CC[BTC_ptr]    := CC
    ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

    BTC_search(PC : val(int), Thumb : val(bool), BTC_index : var(int)):
               expr(int) =
      (
      local_PC   : var(int).
      i          : var(int).
      prediction : var(int).

      prediction := 16#00;          /* Bits 6,5: 00=none, x1=even, 10=odd */
      local_PC := PC and 16#FFFFFFFC;

      if (!Thumb) then
        forseq i in 0 to (BTC_size-1) do               /* ARM mode search */
          (
          if ((BTC_CAM[i] = local_PC) && BTC_valid[i]) then
            (BTC_index := i; prediction := 16#20)      /* ARM mode: found */
          )
      else
        (                                            /* Thumb mode search */
        index_e : var(int).
        index_o : var(int).

        forseq i in 0 to (BTC_size-1) do
          (
          if ((BTC_CAM[i] = local_PC) && BTC_valid[i]) then
            (                                   /* Test for even or odd ? */
            if ((i and 1) = 0) then             /* Thumb mode: found even */
              (index_e := i; prediction := prediction or 16#20)
            else                                /* Thumb mode: found odd  */
              (index_o := i; prediction := prediction or 16#40)
            )
          );

        if ((PC and 2) = 0) then                /* Both halves considered */
          BTC_index  := (if ((prediction and 16#20) != 0) then index_e
                         else index_o)
        else
          (                                   /* Only odd half considered */
          prediction := prediction and 16#5F; /* Ensure "even" bit clear */
          BTC_index  := index_o
          )
        );

      prediction
      ).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

    BTC_what_branch(BTC_index : val(int)) : expr(int) =
      ((if BTC_link[BTC_index] then 16#10 else 16#00)
       or (BTC_CC[BTC_index] and 16#0F)).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

    BTC_read(BTC_index : val(int)) : expr(int) = BTC_PC[BTC_index].

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/**/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

  FIQ_svc : var(bool).
  IRQ_svc : var(bool).

  MemBundle : var(Addr_bus_type).
  MCBundle  : var(MemCtrl_bus_type).
  Loaded_PC : var(Data_bus_type).

  BTC_prediction : var(int).        /* Bits 6,5: 00=none, x1=even, 10=odd */
  BTC_index      : var(int).


/* Initialise static or unused record elements                            */
  MemBundle->Read  := true;
  MemBundle->Write := false;
  MemBundle->Size  := 4;
  MemBundle->Dout  := 0;                                        /* Unused */

/* Initialise BTC                                                         */
  (i : var(int). forseq i in 0 to (BTC_size-1) do BTC_valid[i] := false);

  MCBundle->PC               := 0;                  /* Initialise PC etc. */
  MCBundle->col              := 0;
  MCBundle->PSR_ctrl->IRQ_En := false;             /* Interrupts disabled */
  MCBundle->PSR_ctrl->FIQ_En := false;
  MCBundle->PSR_ctrl->Thumb  := false;                   /* Full ARM mode */
  MCBundle->PSR_ctrl->mode   := 3;          /* Reset into supervisor mode */

  MemBundle->seq             := false;           /* Non-sequential access */

  FIQ_svc := false;
  IRQ_svc := false;


/*forever*/
  while Run do
    (
    watchdog := now;                                     /* Keep time log */
    wait_for(IAI_delay);

    MCBundle->IndirPC    := false;   /* No indirection unless "branch"ing */
    MCBundle->Mem_bypass := false;
    MCBundle->extension  := 0;                       /* No extension code */
    MCBundle->Dabt_svc   := false;

    if chan_ready(Branch) then               /* See if a branch is wanted */
      (Branch?                                             /* Yes - do it */
        (
        MCBundle->col  := (?Branch)->New_col;               /* Regardless */
        MemBundle->seq := false;                 /* Non-sequential access */

        if ((?Branch)->Exception != 0) then
          (                        /* Exception handler - code is address */
            /* Possibilities are Undef (4), SWI (8), Pabt (12), Dabt (16) */
          MCBundle->PC         := (?Branch)->Exception;  /* Branch to ... */
/*
          MCBundle->PSR_ctrl->Thumb := false;
          if ((?Branch)->Exception < 12) then
            (
            if ((?Branch)->Exception = 4) then
              MCBundle->PSR_ctrl->mode := 11            /* Undefined mode */
            else
              MCBundle->PSR_ctrl->mode := 3            /* Supervisor mode */
            )
          else
            (
            MCBundle->PSR_ctrl->mode := 7;                  /* Abort mode */
            if ((?Branch)->Exception = 16) then
              (
              MCBundle->extension := 1;                     /* Data abort */
              MCBundle->Dabt_svc  := true
              )
            )
*/
          MCBundle->PSR_ctrl := (?Branch)->New_PSR_ctrl;

/* Checking routines */
          if (MCBundle->PSR_ctrl->IRQ_En) then
            printstr("Exception entry - IRQ still enabled\n");

          (case ((?Branch)->Exception)
            when  4 => if (MCBundle->PSR_ctrl->mode != 11) then
                         printstr("Undefined entry - wrong mode\n")
            when  8 => if (MCBundle->PSR_ctrl->mode !=  3) then
                         printstr("SWI entry - wrong mode\n")
            when 12 => if (MCBundle->PSR_ctrl->mode !=  7) then
                         printstr("Pabt entry - wrong mode\n")
            when 16 => if (MCBundle->PSR_ctrl->mode !=  7) then
                         printstr("Dabt entry - wrong mode\n")
            default printstr("Bad exception entry\n")
          );                                  /* End of confidence checks */

          if ((?Branch)->Exception = 16) then
            (
            MCBundle->extension := 1;                       /* Data abort */
            MCBundle->Dabt_svc  := true
            )

          )
        else
          (
          MCBundle->PC       := (?Branch)->New_PC;
          MCBundle->IndirPC  := (?Branch)->Indir_req;
          MCBundle->PSR_ctrl := (?Branch)->New_PSR_ctrl;
/* printstr("Branched to ");printhex(MCBundle->PC);printstr("\n"); */

          if ((?Branch)->Cacheable && BTC_Enable) then   /* New BTC entry */
            BTC_insert((?Branch)->Old_PC,
                       (?Branch)->New_PC,
                       (?Branch)->link,
                       (?Branch)->CC,
                       (?Branch)->New_PSR_ctrl->Thumb)
          )
        );
      FIQ_svc := false;   /* Override any interrupt operation in progress */
      IRQ_svc := false
      );         /* NB Branch "ack"ed before address dispatched - WRONG ? */



/*Insert interrupt service address (currently overwites incremented value)*/
/*  Hangover from previous cycle                                          */
    if FIQ_svc then (MCBundle->PC := 28; FIQ_svc := false);
    if IRQ_svc then (MCBundle->PC := 24; IRQ_svc := false);

    if !(MCBundle->IndirPC) then          /* PC load is non-interruptable */
      /* Test for interrupts                                              */
      (
      if (MCBundle->PSR_ctrl->FIQ_En && FIQ) then
        (
        FIQ_svc := true;
        MCBundle->extension        := 2;
        MCBundle->PSR_ctrl->IRQ_En := false;
        MCBundle->PSR_ctrl->FIQ_En := false;
        MCBundle->PSR_ctrl->Thumb  := false;
        MCBundle->PSR_ctrl->mode   := 1
        )
      else
        (
        if (MCBundle->PSR_ctrl->IRQ_En && IRQ) then
          (
          IRQ_svc := true;
          MCBundle->extension        := 3;
          MCBundle->PSR_ctrl->IRQ_En := false;
          MCBundle->PSR_ctrl->Thumb  := false;
          MCBundle->PSR_ctrl->mode   := 2
          )
        )
      );


    MemBundle->addr  := MCBundle->PC and 16#FFFFFFFC;     /* Always words */
    MemBundle->mode  := MCBundle->PSR_ctrl->mode;
    MemBundle->Qcol  := MCBundle->col;    /* This one probably irrelevant */
    MemBundle->extra := (if MCBundle->IndirPC then 1 else 0);

/*  Bypass for any unusual case except "entry" to Dabt                    */
    if (MCBundle->extension > 1) then MCBundle->Mem_bypass := true;

/*
    printstr("IAI: MAR = "); printhex(MCBundle->PC); printstr("  Colour ");
    printint(MCBundle->col); printstr("\n");
*/
    if (MCBundle->IndirPC) then wait_until(STR_semaphore = 0);
                             /* Wait for stores to complete if loading PC */

    if (BTC_Enable && (MCBundle->extension = 0)) then         /* CHECK !!!*/
      BTC_prediction := BTC_search(MCBundle->PC,             /* Check BTC */
                                   MCBundle->PSR_ctrl->Thumb,
                                   BTC_index)
    else
      BTC_prediction := 16#00;

    if (BTC_prediction != 0) then               /* Get branch instruction */
      (
      MCBundle->extension  := BTC_prediction or BTC_what_branch(BTC_index);
      MCBundle->Mem_bypass := ((BTC_prediction and 16#20) != 0)
                         /* Only if "even" prediction - includes ARM mode */
      );

    (                                                  /* Parallel output */
      if !(MCBundle->Mem_bypass) then to_Mem!MemBundle
    |
      to_MemCtrl!MCBundle
    );

    if (MCBundle->IndirPC) then              /* Did we request a PC load? */
      (                                        /* Yes - wait for response */
      indirPC?(Loaded_PC:=(?indirPC));  /* Input from indirection channel */
                                        /* Early ack. - should not matter */
      if (Loaded_PC->abort) then
        (                          
        while (wait_until(chan_ready(Branch));
               !((?Branch)->Exception = 16)) do   /* Data abort = code 16 */
             Branch?null            /* Spurious branch discard and repeat */
                                   /* Shouldn't happen - remove when sure */
        )                                   /* End of indirect load stuff */
      else
        MCBundle->PC := Loaded_PC->data                /* Adopt loaded PC */
      )
    else
      (
      if (BTC_prediction != 0) then
        (
        MemBundle->seq := false;                 /* Non-sequential access */
        MCBundle->PC   := BTC_read(BTC_index) /* Predicted branch address */
        )
      else
        (
        MemBundle->seq := true;                      /* Sequential access */
        MCBundle->PC   := ((MCBundle->PC) + 4) and 16#FFFFFFFC
        )                                  /* PC increment and word align */
      )

  )
).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
