/*------------------------------------------------------------------------*/
/*                                                                        */
/*                           AMULET3 - The All New Queue                  */
/*                                                                        */
/*                                             Created: JDG   8/1/97      */
/*                                       Last Modified: JDG   5/6/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/* Maybe should contain BRpipe also?                                      */
/*                                                                        */
/* NEEDS COLOURED INVALIDATION MECHANISM ADDING (& forwarding colour).    */
/*                                                                        */
/*------------------------------------------------------------------------*/
/* Scoping should probably be tightened */

Queue_entry_type : type = record(data::int,
                                 full::bool,
                              invalid::bool,
                                abort::bool,
                               has_BR::bool,
                                 Fcol::int,
                            bv_colour::int).

Queue_buffer : var(Queue_entry_type^10).

/*------------------------------------------------------------------------*/

Queue_input(input : var(chan(Queue_input_type))):
            expr(void)=
  (
  forever
    (
    input?
      (
      wait_for(Qin_delay);
      (Queue_buffer[(?input)->queue_addr])->data    := (?input)->data;
      (Queue_buffer[(?input)->queue_addr])->invalid := (?input)->invalid;
      (Queue_buffer[(?input)->queue_addr])->abort   := (?input)->abort;
      (Queue_buffer[(?input)->queue_addr])->has_BR  := (?input)->has_BR;
      (Queue_buffer[(?input)->queue_addr])->full    := true;
      (Queue_buffer[(?input)->queue_addr])->Fcol    :=
                           (Queue_buffer[(?input)->queue_addr])->Fcol xor 1;
      (Queue_buffer[(?input)->queue_addr])->bv_colour := (?input)->bv_colour;
      bview_set_colour("queue"@i2s((?input)->queue_addr),
                       bview_colour_list[(?input)->bv_colour])
      )
    )
  ).

/*------------------------------------------------------------------------*/

Queue_output(output    : var(chan(Queue_output_type)),
             BRpipe    : var(chan(BRpipe_type)),
             QTout     : var(chan(void)),
             Dabt_go_d : var(chan(void))):
             expr(void)=
  (
  ptr    : var(int).
  result : var(Queue_output_type).

  ptr := 0;

  forever
    (
    wait_until((Queue_buffer[ptr])->full);           /* Next result ready */

    wait_for(Qout_delay);

    if !((Queue_buffer[ptr])->invalid) then
      (                                                   /* if valid ... */
      result->data     := (Queue_buffer[ptr])->data;    /* ... write back */
      result->bv_colour := (Queue_buffer[ptr])->bv_colour;

      if ((Queue_buffer[ptr])->abort) then                  /* if aborted */
        (
        result->reg_addr := 14;                           /* write LR_abt */
        result->mode     := 7
        )
      else
        (     /* Naughtily peeks what should be local data in "registers" */
        result->reg_addr := (Queue_CAM[ptr])->addr;
        result->mode     := (Queue_CAM[ptr])->mode
        );

      if (result->reg_addr != 15) then output!result /* Don't control R15 */
      );

  /* Complete reading & copying base register before allowing abort to go */
    if ((Queue_buffer[ptr])->has_BR) then
      (
      BRpipe?
        (
        if (((Queue_buffer[ptr])->abort)
       && (!((Queue_buffer[ptr])->invalid))) then
          (                                /* Recover base register value */
          result->reg_addr := (?BRpipe)->reg;
          result->mode     := (?BRpipe)->mode;
          result->data     := (?BRpipe)->base;
          if (result->reg_addr != 15) then output!result   /* Restore reg.*/
          )
        else
          null                           /* No abort - discard base value */
        )                                                /* Ack base pipe */
      );

    if (((Queue_buffer[ptr])->abort)
  && (!((Queue_buffer[ptr])->invalid))) then
      (                                         /* Invalidate whole queue */
      i : var(int).
      forseq i in 0 to 9 do (Queue_buffer[i])->invalid := true;
      Dabt_go_d!null          /*  Signal that abort response may continue */
      );                      /* now that the register bank is up to date */

    (Queue_buffer[ptr])->full := false;               /* Deallocate entry */

    QTout!null;                  /* signal queue location ready for reuse */

    ptr := (ptr + 1) mod Queue_size                  /* Move to next slot */
    )
  ).

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

Queue_forward(input  : var(chan(Forward_req_type)),
              output : var(chan(Forward_data_type))):
              expr(void)=
  (
  in_latch : var(Forward_req_type).
  reg_out  : var(Forward_data_type).

  i    : var(int).
  slot : var(int).
  Fcol : var(int).

  forever               /* MORE COMPLEX DELAYS REALLY */
    (
    input?(in_latch := (?input));          /* NOT a pipeline stage though */
                       /* This is to allow decoder to proceed in parallel */

    i    := 0;
    Fcol := in_latch->Fcol;                /* Start with requested colour */
    slot := in_latch->start;                  /* First candidate position */
    reg_out->found := false;
    while ((!reg_out->found) && (i < Queue_size)) do
      (
      if (in_latch->vector[slot]) then
        (                                           /* Possible candidate */
        wait_until(Fcol = (Queue_buffer[slot])->Fcol);
        wait_for(Qfwd_cand_delay);            /* Delay for examining slot */
        reg_out->found := (!((Queue_buffer[slot])->invalid))
        )
      else
        wait_for(Qfwd_byp_delay);             /* Delay for bypassing slot */

      if !(reg_out->found) then 
        (
        i    := i + 1;                           /* Count through entries */
        slot := slot - 1;                      /* Move to _previous_ slot */
        if (slot < 0) then
          (
          slot := slot + Queue_size;                   /* Wrap around ... */
          Fcol := Fcol xor 1              /* ... thus expect other colour */
          )
        )
      );                                                     /* End while */

    if (reg_out->found) then reg_out->data := (Queue_buffer[slot])->data
    else                     reg_out->data := 0;   /* Just for appearance */

    wait_for(Qfwd_delay);                 /* Real timing is quite complex */

    output!reg_out
    )                                                 /* End of "forever" */
  ).

/*------------------------------------------------------------------------*/
/**/
/*------------------------------------------------------------------------*/

Queue(from_exec : var(chan(Queue_input_type)),
      from_Dmem : var(chan(Queue_input_type)),
      skip_Dmem : var(chan(Queue_input_type)),
      reg_out   : var(chan(Queue_output_type)),
      BRpipe    : var(chan(BRpipe_type)),
      QTout     : var(chan(void)),
      Dabt_go   : var(chan(void)),
      QA_req    : var(chan(Forward_req_type)),
      QB_req    : var(chan(Forward_req_type)),
      QC1_req   : var(chan(Forward_req_type)),
      QC2_req   : var(chan(Forward_req_type)),
      QA_data   : var(chan(Forward_data_type)),
      QB_data   : var(chan(Forward_data_type)),
      QC1_data  : var(chan(Forward_data_type)),
      QC2_data  : var(chan(Forward_data_type))):
      expr(void)=
  (
/* "Dabt_go_decouple" frees "Queue_output" task to continue draining pipe */
  Dabt_go_decouple : var(chan(void)).
  i : var(int).

  init_chan(Dabt_go_decouple, "Dabt_go_latch");

  forseq i in 0 to 9 do
    (                                                      /* Reset queue */
    (Queue_buffer[i])->full := false;
    (Queue_buffer[i])->Fcol := 0

    );

  forseq i in 1 to Queue_size do QTout!null;          /* Prime token pipe */

  (
  Queue_input(from_exec) |
  Queue_input(from_Dmem) |
  Queue_input(skip_Dmem) |
  Queue_output(reg_out, BRpipe, QTout, Dabt_go_decouple) |
  Queue_forward(QA_req, QA_data) |
  Queue_forward(QB_req, QB_data) |
  Queue_forward(QC1_req, QC1_data) |   /* NB really a single process with */
  Queue_forward(QC2_req, QC2_data) |   /*   steering control later.       */

  Latch(Dabt_go_decouple, Dabt_go, 0, false, "Dabt_go_latch")
                                                   /* Spawn latch process */
  )

  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
