/*------------------------------------------------------------------------*/
/*                                                                        */
/*                         AMULET3 - Register bank                        */
/*                                                                        */
/*                                             Created: JDG   9/1/97      */
/*                                       Last Modified: JDG  29/4/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/*                                                                        */
/*                                                                        */
/*------------------------------------------------------------------------*/
/* SCOPING? */

Register_bank    : var(int^42).
Register_bank_PC : var(int).

/*------------------------------------------------------------------------*/
/* Map banked registers to unique address                                 */

Reg_no(reg : val(int), mode : val(int)) : expr(int) =
  (
  if ((mode = 0) || (mode = 15) || (reg < 8)) then reg
  else
    (
    if (mode = 1) then (reg + 7)                                   /* FIQ */
    else
      (
      if (reg > 12) then (reg + (2 * mode) + 5) /* Other banked registers */
      else reg
      )
    )
  ).

/*------------------------------------------------------------------------*/

Register_write(input : var(chan(Queue_output_type))):
               expr(void)=
  (
  forever
    (
    input?
      (
      active_block("reg_write",(?input)->bv_colour,(
      wait_for(REG_wr_delay);
      Register_bank[Reg_no((?input)->reg_addr, (?input)->mode)] := (?input)->data
      ))
      )                                      /* ... and acknowledge input */
    )
  ).

/*------------------------------------------------------------------------*/

Register_read(reg_addr    : var(chan(Reg_addr_type)),
              queue_value : var(chan(Forward_data_type)),
              reg_value   : var(chan(int))):
              expr(void)=
  (

  forever
    (
    reg_addr?
      (
      reg_value!
        (
        wait_for(REG_rd_delay);                     /* Register read time */
        queue_value?
          (
          if ((?reg_addr)->addr = 15) then               /* MUX delay ??? */
            Register_bank_PC     /* PC set earlier with incremented value */
          else
            (
            if ((?queue_value)->found) then
              (?queue_value)-> data  /* Select default or forwarded value */
            else
              Register_bank[Reg_no((?reg_addr)->addr, (?reg_addr)->mode)]
            )
          )                                    /* Acknowledge queue input */
        )                                                /* Output result */
      )                                      /* Acknowledge address input */
    )                                                          /* forever */
  ).

/*------------------------------------------------------------------------*/

Registers(A_addr    : var(chan(Reg_addr_type)),
          QA_data   : var(chan(Forward_data_type)),
          Areg      : var(chan(int)),
          B_addr    : var(chan(Reg_addr_type)),
          QB_data   : var(chan(Forward_data_type)),
          Breg      : var(chan(int)),
          C1_addr   : var(chan(Reg_addr_type)),
          QC1_data  : var(chan(Forward_data_type)),
          C1reg     : var(chan(int)),
          C2_addr   : var(chan(Reg_addr_type)),
          QC2_data  : var(chan(Forward_data_type)),
          C2reg     : var(chan(int)),
          writeback : var(chan(Queue_output_type))):
          expr(void)=
  (
  /* Any old initialisation                                               */
  (i : var(int). forseq i in 0 to 14 do Register_bank[i] := 2*i);

  (
  Register_read(A_addr,  QA_data,  Areg) |
  Register_read(B_addr,  QB_data,  Breg) |
  Register_read(C1_addr, QC1_data, C1reg) |  /* NB there is really only   */
  Register_read(C2_addr, QC2_data, C2reg) |  /* one C_read process        */
  Register_write(writeback)
  )

  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
/*
  ------------------------------------------------------------------------
Register mapping table:

	No	Reg	No	Reg	No	Reg
	 0	 0	15	 8_fiq	30	-
	 1	 1	16	 9_fiq	31	-
	 2	 2	17	10_fiq	32	13_abt
	 3	 3	18	11_fiq	33	14_abt
	 4	 4	19	12_fiq	34	-
	 5	 5	20	13_fiq	35	-
	 6	 6	21	14_fiq	36	-
	 7	 7	22	13_irq	37	-
	 8	 8	23	14_irq	38	-
	 9	 9	24	13_svc	39	-
	10	10	25	14_svc	40	13_und
	11	11	26	-	41	13_und
	12	12	27	-	
	13	13	28	-	
	14	14	29	-	
*/
/*------------------------------------------------------------------------*/
