/*------------------------------------------------------------------------*/
/*                                                                        */
/*                        AMULET3 Utility routines                        */
/*                                                                        */
/*                                             Created: JVW 13/12/96      */
/*                                            Modified: JDG 16/12/96      */
/*                                            Modified: JVW 17/12/96      */
/*                                       Last Modified: JDG  20/5/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/* Dummy block - simulates logic delay */

Dummy(In  : var(chan(T:type)),
      Out : var(chan(T)),
      Delay: val(int),
      Print: val(bool),
      Name: val(string)):
      expr(void)=
  (
  forever
    (
    In?
      (
      if (Print) then (printstr(Name@" at time "); printint(now); printstr("\n"));
      wait_for(Delay);
      Out!(?In)
      )
    )
  ).

/*------------------------------------------------------------------------*/
/* Single latch */

Latch(In   : var(chan(T:type)),
      Out  : var(chan(T)),
      Delay: val(int),
      Print: val(bool),
      Name : val(string)):
      expr(void)=
  (
  latch : var(T).

  forever
    (
    In?(latch:=(?In));
    if (Print) then (printstr(Name@" at time "); printint(now); printstr("\n"));
    wait_for(Delay);
    Out!latch
    )
  ).

/*------------------------------------------------------------------------*/
/* Channel data sink */
/* ADD DELAYS */

Sink(Z    : var(chan(T:type)),
     Print: val(bool),
     Name : val(string)):
     expr(void)=
  (
  forever
    (
    Z?null;
    if (Print) then (printstr(Name@" at time "); printint(now); printstr("\n"))
    )
  ).

/*------------------------------------------------------------------------*/
/* Channel data source - data sourced has undefined value(s) */
/* ADD DELAYS */

Source(Z    : var(chan(T:type)),
       Print: val(bool),
       Name : val(string)):
       expr(void)=
  (
  undef : var(T).

  forever
    (
    if (Print) then (printstr(Name@" at time "); printint(now); printstr("\n"));
    Z!undef
    )
  ).

/*------------------------------------------------------------------------*/

Delayed_write(Variable : var(T:type), Value : val(T), delay : val(int)) :
              expr(void) =
  (
  sink(spawn(wait_for(delay); Variable := Value, forparnum))
  ).

/*------------------------------------------------------------------------*/

/*    This is a generic FIFO                                              */

FIFO(Input  : var(chan(T:type)),
     Output : var(chan(T)),
     Size   : val(int),
     Delay  : val(int),
     Print  : val(bool),
     Name   : val(string)):
     expr(void)=
  (

/* "Delay" is currently not finalised */

    Fifo:var(T^10).
    W_ptr:var(int).
    R_ptr:var(int).
    F_full:var(bool).
    F_empty:var(bool).

    W_ptr:=0;
    R_ptr:=0;
    F_full:=false;
    F_empty:=true;

    if (Size = 0) then
      Dummy(Input, Output, Delay, Print, Name)
    else
      (
      forever
        (
        wait_until(!F_full);
        Input?(Fifo[W_ptr]:=(?Input));
        W_ptr:= (W_ptr+1) mod Size;
        F_empty:=false;
        if (W_ptr = R_ptr) then F_full:=true;
        if (Print) then
          (printstr(Name); printint((((W_ptr+Size-1)-R_ptr) mod Size)+1);
           printstr(" locations filled at time "); printint(now);
           printstr("\n"))
        )  /* forever */

      |

      forever
        (
        wait_until(!F_empty);
        wait_for(Delay);
        Output!(Fifo[R_ptr]);
        R_ptr:= (R_ptr+1)mod Size;
        F_full:=false;
        if (R_ptr = W_ptr) then F_empty:=true;
        if (Print) then
          (printstr(Name); printint(((W_ptr+Size)-R_ptr) mod Size);
           printstr(" locations filled at time "); printint(now);
           printstr("\n"))
        )  /* forever */
      )    /* if ... then ... else */

  ).

/*------------------------------------------------------------------------*/

ASCII_out(x : val(int)) : expr(void) =
  (
  y : var(int).

  y := x and 16#7F;
/*
if (y=16#5C) then y:=32;                 /* HORRIBLE "\" BODGE FOR LARD 6 */
*/
  if (y < 16#20) then
    (
    if ((y = 16#0D) || (y = 16#0A)) then printstr("\n")
    else (printstr("["); printint(y); printstr("]"))
    )
  else
    print(cast(char, y))
  ).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
