/*------------------------------------------------------------------------*/
/*                                                                        */
/*                           AMULET3 Thumb decoder                        */
/*                                                                        */
/*                                             Created: JVW  22/1/97      */
/*                                                                        */
/*------------------------------------------------------------------------*/
/* By the pricking of our Thumbs !!                                       */

(Thumb(In      : var(chan(Instr_pkt_type)),
       Out     : var(chan(Instr_pkt_type)),
       Dabt_go : var(chan(void))):
       expr(void)=
  (

  latch  : var(Instr_pkt_type).
  ThumbI : var(int).
  ArmI   : var(int).
  InstExt : var(int).

/* The following is a Field Move procedure.  */
(FM(Mask   : val(int),
    Shift_L : val(bool),
    Shift_dist : val(int)):
    expr(void)=
 (
  Al_reg : var(int).

  Al_reg := ThumbI and Mask;
    if (Shift_L) then
       (
        Al_reg := Al_reg << Shift_dist
       )
    else
       (
        Al_reg := Al_reg >> Shift_dist;
        ArmI := ArmI or Al_reg
       )
 )
).        /* End of Field Move definition */


/* The following is a Field Move procedure.  */
/* But this one inverts the selected field   */
(FMI(Mask   : val(int),
    Shift_L : val(bool),
    Shift_dist : val(int)):
    expr(void)=
 (
  Al_reg : var(int).

  Al_reg := ThumbI and Mask;
    if (Shift_L) then
      (
       Al_reg := Al_reg << Shift_dist
       )
    else
       (
        Al_reg := Al_reg >> Shift_dist
       );
    Al_reg := Al_reg xor Mask;
    ArmI := ArmI or Al_reg
 )
).
/* End of Inverting Field Move definition */

/* The following is a Field Insert  procedure.  */
/* This one inserts supplied (Mask) field   */
(FI(In_field : val(int),
    Shift_L : val(bool),
    Shift_dist : val(int)):
    expr(void)=
 (
  Al_reg : var(int).
  Al_reg := In_field;
    if (Shift_L) then
       (Al_reg := Al_reg << Shift_dist)
    else
       (Al_reg := Al_reg >> Shift_dist);
    ArmI := ArmI or Al_reg
 )
).
/* End of Field Insert definition */

/* The following expands Thumb encoding to Arm format  */
/* Assumption here is that Thumb instruction is in ls 16 bits of input field */
/*                                                                          */
(Thumb_Ex(ThumbI : var(int),
          ArmI   : var(int)):
          expr(void)=

(
 M : var(int).  /* mask */
 L : var(bool). /* Left Shift */
 D : var(int).  /* Shift Distance */


(F1orF2Exp: expr(void)=
 (if ((ThumbI and 16#00001800) != 16#00001800) then  /* F1 Move shifted reg. */
    (
      ArmI := 16#E1B00000;
      M := 16#00001800;     /* Op field */
      L := false;
      D := 10#6;
      FM(M,L,D);
      M := 16#000007C0;      /* Offset5  */
      L := true;
      D := 10#1;
      FM(M,L,D);
      M := 16#00000038;      /* Rs  */
      L := false;
      D := 10#3;
      FM(M,L,D);
      M := 16#00000003;      /* Rd  */
      L := true;
      D := 10#12;
      FM(M,L,D)
    )
                /* End of Format 1 expansion  */
 else                                          /* F2 Add/Sub*/
    (
      ArmI := 16#E0100000;
      M := 16#00000400;     /* Imm Bit */
      L := true;
      D := 10#15;
      FM(M,L,D);
      M := 16#00000200;     /* Op field */
      L := true;
      D := 10#13;
      FM(M,L,D);            /* Bit 22  */
      D := 10#14;
      FMI(M,L,D);            /* Bit 23  */
      M := 16#000001C0;      /* Offset3/Rn  */
      L := false;
      D := 10#6;
      FM(M,L,D);
      M := 16#00000038;      /* Rs  */
      L := true;
      D := 10#13;
      FM(M,L,D);
      M := 16#00000003;      /* Rd  */
      L := true;
      D := 10#12;
      FM(M,L,D)
    )             /* End of Format 2 expansion  */
 )
                   /* End of F1 or F2 exp.  */
).                      /* End of F1orF2Exp*/


/*=========================================================================*/
                /* F3 Mov/comp/add/sub imm. */
(F3Exp: expr(void)=
 (
     ArmI := 16#E0100000;
     L := true;
     D := 10#1;
     if ((ThumbI and 16#00001800) = 16#00000000) then  /* F3(00) - MOV  */
          (
           M := 16#00D00000;
           FI(M,L,D)
          )
     else
       (
       if ((ThumbI and 16#00001800) = 16#00000800) then /* F3(01) - CMP  */
            (
             M := 16#00A00000;
             FI(M,L,D)
            )
       else
         (
         if ((ThumbI and 16#00001800) = 16#00001000) then  /* F3(10) - ADD  */
              (
               M := 16#00400000;
               FI(M,L,D)
              )
         else
           (
           if ((ThumbI and 16#00001800) = 16#00001800) then  /* F3(11) - SUB */
                (
                 M := 16#00200000;
                 FI(M,L,D)
                 )
           )
         )
       );

     M := 16#00000700;      /* Rd  */
     L := true;
     D := 10#8;
     FM(M,L,D);              /* As First Op.*/
     D := 10#4;
     FM(M,L,D);              /* As Dest.*/
     M := 16#000000FF;      /* Offset8  */
     L := true;
     D := 10#0;
     FM(M,L,D)
 )
).                  /*  End of F3Exp   */
/*=========================================================================*/

/* >>>>>>>>>>>>>>>>>>> F4 EXPANSION <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
(F4Exp: expr(void)=
 (
   ArmI := 16#E0000000;

   if(
      ((ThumbI and 16#000003C0) = 16#00000000) ||  /* AND */
      ((ThumbI and 16#000003C0) = 16#00000040) ||  /* EOR */
      ((ThumbI and 16#000003C0) = 16#00000140) ||  /* ADC */
      ((ThumbI and 16#000003C0) = 16#00000180) ||  /* SBC */
      ((ThumbI and 16#000003C0) = 16#00000200) ||  /* TST */
      ((ThumbI and 16#000003C0) = 16#00000280) ||  /* CMP */
      ((ThumbI and 16#000003C0) = 16#000002C0) ||  /* CMN */
      ((ThumbI and 16#000003C0) = 16#00000300) ||  /* ORR */
      ((ThumbI and 16#000003C0) = 16#00000380) ||  /* BIC */
      ((ThumbI and 16#000003C0) = 16#000003C0)     /* MVN */
      )
   then
      (
        (
         L := true;     
         D := 10#12;
         M := 16#00000007;     /* Rd field  */
         FM(M,L,D);            /* Rd as dest */
         D := 10#16;
         FM(M,L,D);            /* Rd as Op-1 */
         M := 16#00000038;     /* Rs field  */
         L := false;     
         D := 10#3;
         FM(M,L,D)             /* Rs */
        );

         if ((ThumbI and 16#000003C0) = 16#00000000) then  /* AND */
          (
           M := 16#00000000;     /* ARM-Op = 16#0  */
           L := true;     
           D := 10#1;            /* SL 1 Req. */
           FI(M,L,D)             /* Insert Op-code in 24-21 */
          )
         else
          (if ((ThumbI and 16#000003C0) = 16#00000040) then  /* EOR */
           (
            M := 16#00100000;     /* ARM-Op = 16#1  */
            L := true;     
            D := 10#1;            /* SL 1 Req. */
            FI(M,L,D)             /* Insert Op-code in 24-21 */
           )
          else
           (if ((ThumbI and 16#000003C0) = 16#00000140)  then  /* ADC */
            (
             M := 16#00500000;     /* ARM-Op = 16#5  */
             L := true;     
             D := 10#1;            /* SL 1 Req. */
             FI(M,L,D)             /* Insert Op-code in 24-21 */
            )
           else
            (if ((ThumbI and 16#000003C0) = 16#00000180) then  /* SBC */
             (
              M := 16#00600000;     /* ARM-Op = 16#6  */
              L := true;     
              D := 10#1;            /* SL 1 Req. */
              FI(M,L,D)             /* Insert Op-code in 24-21 */
             )
            else
             (if ((ThumbI and 16#000003C0) = 16#00000200) then  /* TST */
              (
               M := 16#00800000;     /* ARM-Op = 16#8  */
               L := true;     
               D := 10#1;            /* SL 1 Req. */
               FI(M,L,D)             /* Insert Op-code in 24-21 */
              )
             else
              (if ((ThumbI and 16#000003C0) = 16#00000280) then  /* CMP */
               (
                M := 16#00A00000;     /* ARM-Op = 16#A  */
                L := true;     
                D := 10#1;            /* SL 1 Req. */
                FI(M,L,D)             /* Insert Op-code in 24-21 */
               )
              else
               (if ((ThumbI and 16#000003C0) = 16#000002C0) then  /* CMN */
                (
                 M := 16#00B00000;     /* ARM-Op = 16#B  */
                 L := true;     
                 D := 10#1;            /* SL 1 Req. */
                 FI(M,L,D)             /* Insert Op-code in 24-21 */
                )
               else
                (if ((ThumbI and 16#000003C0) = 16#00000300) then  /* ORR */
                 (
                  M := 16#00C00000;     /* ARM-Op = 16#C  */
                  L := true;     
                  D := 10#1;            /* SL 1 Req. */
                  FI(M,L,D)             /* Insert Op-code in 24-21 */
                 )
                else
                 (if ((ThumbI and 16#000003C0) = 16#00000380 )then  /* BIC */
                  (
                   M := 16#00E00000;     /* ARM-Op = 16#E  */
                   L := true;     
                   D := 10#1;            /* SL 1 Req. */
                   FI(M,L,D)             /* Insert Op-code in 24-21 */
                  )
                 else
                  (if ((ThumbI and 16#000003C0) = 16#000003C0) then /* MVN */
                   (
                    M := 16#00F00000;     /* ARM-Op = 16#F  */
                    L := true;     
                    D := 10#1;            /* SL 1 Req. */
                    FI(M,L,D)             /* Insert Op-code in 24-21 */
                   )
                  )
                 )
                )
               )
              )
             )
            )
           )
          )

      )      /*End of AND/EOR/ADC/SBC/TST/CMP/CMN/ORR/BIC/MVN sub-group*/
     else    /* It's the others.....*/
      (
       if (
          ((ThumbI and 16#000003C0) = 16#00000080) ||  /* LSL */
          ((ThumbI and 16#000003C0) = 16#000000C0) ||  /* LSR */
          ((ThumbI and 16#000003C0) = 16#00000100) ||  /* ASR */
          ((ThumbI and 16#000003C0) = 16#000001C0)) then     /* ROR */
          (   /* The Shift Group  */
            L := true;     
            D := 10#0;
            M := 16#00000007;     /* Rd field  */
            FM(M,L,D);            /* Rd as dest */
            D := 10#16;
            FM(M,L,D);            /* Rd as Op-1 */
            M := 16#00000038;     /* Rs field  */
            L := true;     
            D := 10#5;
            FM(M,L,D);             /* Rs */
           if ((ThumbI and 16#000003C0) = 16#00000080) then  /* LSL */
            (
             M := 16#00D00000;
             L := true;
             D := 10#1;
             FI(M,L,D)         /* F4 LSL -> MOVS  */
            )
           else
           (if ((ThumbI and 16#000003C0) = 16#000000C0) then  /* LSR */
             (
              M := 16#00D00000;
              L := true;
              D := 10#1;
              FI(M,L,D)         /* F4 LSR -> MOVS  */
             )
            else
             (if ((ThumbI and 16#000003C0) = 16#00000100) then  /* ASR */
              (
               M := 16#00D00000;
               L := true;
               D := 10#1;
               FI(M,L,D)         /* F4 ASR -> MOVS  */
              )
             else
              (if ((ThumbI and 16#000003C0) = 16#000001C0) then  /* ROR */
               (
                M := 16#00D00000;
                L := true;
                D := 10#1;
                FI(M,L,D)         /* F4 ROR -> MOVS  */
               )
              )
             )
            )
          )  /*     End of the Shift Group*/

         else
          (if ((ThumbI and 16#000003C0) = 16#00000240) then  /* NEG */
           (
            M := 16#02000000;     /* Imm bit */
            L := true;
            D := 10#0;
            FI(M,L,D);        /* F4 NEG -> RSBS Rd, Rs, #0  */
            M := 16#00300000; /* RSBS OpCode */
            L := true;
            D := 10#1;
            FI(M,L,D);
            M := 16#00000003; /* Rd field source */
            L := true;
            D := 10#12;
            FM(M,L,D);        /* F4 NEG Rd field  */
            M := 16#00000038; /* Rs field source*/
            L := true;
            D := 10#13;
            FM(M,L,D)         /* Rs field moved  */
           )     /* End of NEG expansion  */
          else
           (if ((ThumbI and 16#000003C0) = 16#00000340) then  /* MUL */
            (
             M := 16#00100000;     /* Op. =0 A=0 S=1 */
             L := true;
             D := 10#0;
             FI(M,L,D);
             M := 16#00000003; /* Rd field */
             L := true;
             D := 10#16;
             FM(M,L,D);        /* F4 MUL Rd as Dest. */
             L := true;
             D := 10#8;
             FM(M,L,D);        /* F4 MUL Rd as Source */
             M := 16#00000038; /* Rs field source */
             L := false;
             D := 10#3;
             FM(M,L,D);        /* F4 MUL Rs as Source. */
             M := 16#00000090; /* Bits 7->4 = 1001 for MUL */
             L := true;
             D := 10#0;
             FI(M,L,D) 
            )             /* MUL is Rd := Rs * Rd in ARM code */
           else
            (printstr("Failed Format 4 Decode"))
           )
          )
      )
 )
).             /* End of F4Exp*/

/*===========================================================================*/

/*      F5 Expansion    */

(F5Exp: expr(void)=
 (
  if ((ThumbI and 16#00000300) != 16#00000300) then /* NOT the "BX" case */
   (
    ArmI := 16#E000000;
    M := 16#00000080;
    L := true;
    D := 10#12;
    FM(M,L,D);              /* H1 bit */
    D := 10#8;
    FM(M,L,D);              /* H1 bit */
    M := 16#00000040;
    L := false;
    D := 10#3;
    FM(M,L,D);              /* H2 bit */
    M := 16#00000038;
    L := false;
    D := 10#3;
    FM(M,L,D);              /* Rs field */
    M := 16#00000007;
    L := true;
    D := 10#12;
    FM(M,L,D);              /* Rd field - Dest. */
    D := 10#16;
    FM(M,L,D);              /* Rd field - First Op. */

    if ((ThumbI and 16#00000300)= 16#00000000) then   /* ADD */
     (
      M := 16#00400000;    /* Op Code 0100 - ADD */
      L := true;
      D := 10#1;
      FI(M,L,D) 
     )
    else
     (if ((ThumbI and 16#00000300)= 16#00000100) then   /* CMP */
      (
       M := 16#00A80000;    /*  OP Code 1010 - CMP- Set CCs */
       L := true;
       D := 10#1;
       FI(M,L,D) 
       )
      else

       (if ((ThumbI and 16#00000300)= 16#00000200) then   /* MOV */
        (
         M := 16#00D00000;    /*  OP Code 1101 - MOV  */
         L := true;
         D := 10#1;
         FI(M,L,D) 
         )
        else
         (printstr("Failed Format 5 Decode"))
       )
      )
   )         /* End of NON-BX cases*/
  else       /*  This is the BX case */
   (
    ThumbI := 16#E12FFF10;
    M := 16#00000040;
    L := false;
    D := 10#3;
    FM(M,L,D);              /* H2 bit */
    M := 16#00000038;
    L := false;
    D := 10#3;
    FM(M,L,D)               /* Rs field */
   )                       /* End of BX case */
 )
).         /* End of F5Exp */

/*=========================================================================*/
/*      F6 Expansion    */

(F6Exp: expr(void)=
 (
  ArmI := 16#E59F0000;
  M := 16#00000700;
  L := true;
  D := 10#4;
  FM(M,L,D);            /* Rd field */
  M := 16#000000FF;
  L := true;
  D := 10#2;
  FM(M,L,D)             /* Word8 field */
 )
).         /* End of F6 expansion */

/*=========================================================================*/
/*      F7 Expansion    */

(F7Exp: expr(void)=
 (
  ArmI := 16#E7800000;
  M := 16#00000800;
  L := true;
  D := 10#9;
  FM(M,L,D);            /* L bit */
  M := 16#00000400;
  L := true;
  D := 10#12;
  FM(M,L,D);            /* B bit */
  M := 16#000008C0;
  L := false;
  D := 10#6;
  FM(M,L,D);            /* Ro - Offset reg */
  M := 16#00000038;
  L := true;
  D := 10#13;
  FM(M,L,D);            /* Rb field */
  M := 16#00000007;
  L := true;
  D := 10#12;
  FM(M,L,D)            /* Rd field */
 )
).         /* End of F7 expansion */

/*=========================================================================*/
/*      F8 Expansion    */

(F8Exp: expr(void)=
 (
  ArmI := 16#E1800090;
  if ((ThumbI and 16#00000C00) = 16#00000000) then
   ( M := 16#00000020;           /* L=0 S=0 H=1 */   
     L := true;
     D := 10#0;
     FI(M,L,D) 
   )
  else
   (if ((ThumbI and 16#00000C00) = 16#00000800) then
    ( M := 16#00100020;           /* L=1 S=0 H=1 */   
      L := true;
      D := 10#0;
      FI(M,L,D) 
    )
   else
    (if ((ThumbI and 16#00000C00) = 16#00000400) then
     ( M := 16#00100040;           /* L=1 S=1 H=0 */   
       L := true;
       D := 10#0;
       FI(M,L,D) 
     )
    else
     (if ((ThumbI and 16#00000C00) = 16#00000C00) then
      ( M := 16#00100060;           /* L=1 S=1 H=1 */   
        L := true;
        D := 10#0;
        FI(M,L,D) 
      )
     else
      (printstr("Failed Format 8 Decode"))
     )
    )
   );
  M := 16#000008C0;
  L := false;
  D := 10#6;
  FM(M,L,D);            /* Ro - Offset reg */
  M := 16#00000038;
  L := true;
  D := 10#13;
  FM(M,L,D);            /* Rb field */
  M := 16#00000007;
  L := true;
  D := 10#12;
  FM(M,L,D)             /* Rd field */
 )
).         /* End of F8 expansion */

/*=========================================================================*/
/*      F9 Expansion    */
(F9Exp: expr(void)=
 (
  ArmI := 16#E5800000;
  M := 16#00000800;
  L := true;
  D := 10#9;
  FM(M,L,D);            /* L bit */
  M := 16#00001000;
  L := true;
  D := 10#10;
  FM(M,L,D);            /* B bit */
  M := 16#000007C0;
  L := false;
  if ((ThumbI and 16#00001000) = 16#00000000) then /* I.E. B=0 */
    (D := 10#2)
  else                                             /* I.E. B=1 */
    (D := 10#0);
  FM(M,L,D);            /*  Offset5 */
  M := 16#00000038;
  L := true;
  D := 10#13;
  FM(M,L,D);            /* Rb field */
  M := 16#00000007;
  L := true;
  D := 10#12;
  FM(M,L,D)             /* Rd field */
 )
).         /* End of F9 expansion */

/*=========================================================================*/
/*      F10 Expansion    */

(F10Exp: expr(void)=
 (
  ArmI := 16#E1C000D0;
  M := 16#00000800;
  L := true;
  D := 10#9;
  FM(M,L,D);            /* L bit */
  M := 16#00000400;
  L := false;
  D := 10#2;
  FM(M,L,D);            /* m.s. Offset5 bit */
  M := 16#000003C0;
  L := false;
  D := 10#6;
  FM(M,L,D);            /* l.s. Offset5 bits */
  M := 16#00000038;
  L := true;
  D := 10#13;
  FM(M,L,D);            /* Rb field */
  M := 16#00000007;
  L := true;
  D := 10#12;
  FM(M,L,D)             /* Rd field */
 )
).         /* End of F10 expansion */

/*=========================================================================*/
/*      F11 Expansion    */

(F11Exp: expr(void)=
 (
  ArmI := 16#E58D0000;
  M := 16#00000800;
  L := true;
  D := 10#9;
  FM(M,L,D);            /* L bit */
  M := 16#00000700;
  L := true;
  D := 10#4;
  FM(M,L,D);            /* Rd field */
  M := 16#000000FF;
  L := true;
  D := 10#2;
  FM(M,L,D)             /* Word8 <<2 */
 )
).         /* End of F11 expansion */

/*=========================================================================*/
/*      F12 Expansion    */

(F12Exp: expr(void)=
 (

  ArmI := 16#E28D0000;
  M := 16#00000800;
  L := true;
  D := 10#6;
  FMI(M,L,D);            /* SP bit */
  M := 16#00000700;
  L := true;
  D := 10#4;
  FM(M,L,D);            /* Rd field */
  M := 16#000000FF;
  L := true;
  D := 10#2;
  FM(M,L,D)             /* Word8 <<2 */
 )
).         /* End of F12 expansion */

/*=========================================================================*/

/*      F13 Expansion    */

(F13Exp: expr(void)=
 (
  ArmI := 16#E20DDF00;

  if ((ThumbI and 16#00000080) = 16#00000000) then /* I.E. S=0 */
   (
    M := 16#00400000
   )
  else
   (
    M := 16#00200000
   );                             /* I.E. S=1 */
  L := true;
  D := 10#1;
  FI(M,L,D);            /* Op Field field */
  M := 16#0000007F;
  L := true;
  D := 10#0;
  FM(M,L,D)             /* SWord7 field */
 )
).         /* End of F13 expansion */

/*=========================================================================*/
/*      F14 Expansion    */

(F14Exp: expr(void)=
 (
  ArmI := 16#E00DDF00;
  if ((ThumbI and 16#00000900) = 16#00000000) then /* L=0 R=0 */
    (M := 16#01200000)
  else
    (if ((ThumbI and 16#00000900) = 16#00000100) then /* L=0 R=1 */
      (M := 16#01600000)
    else
      (if ((ThumbI and 16#00000900) = 16#00000800) then /* L=1 R=0 */
        (M := 16#00D00000)
      else
        (if ((ThumbI and 16#00000900) = 16#00000900) then /* L=1 R=1 */
          (M := 16#00F00000)
        else
          (printstr("Failed Format 14 Decode"))
        )
      )
    );
  L := true;
  D := 10#0;
  FI(M,L,D);            /* Op field */
  M := 16#000000FF;
  L := true;
  D := 10#0;
  FM(M,L,D)             /* Rlist field */
 )
).         /* End of F14 expansion */

/*=========================================================================*/
/*      F15 Expansion (Multple Load/Store)   */

(F15Exp: expr(void)=
 (
  ArmI := 16#E8A00000;
  M := 16#00000800;                             /* L-bit  */
  L := true;
  D := 10#9;
  FM(M,L,D);
  M := 16#00000700;                             /* Rb field */
  L := true;
  D := 10#8;
  FM(M,L,D);
  M := 16#000000FF;
  L := true;
  D := 10#0;
  FM(M,L,D)                                   /* Rlist field */
 )
).         /* End of F15 expansion */

/*=========================================================================*/
/*      F16/17 Expansion    */
(F1617Exp: expr(void)=
 (
  if ((ThumbI and 16#00000F00) != 16#00000F00) then  /* F16 -  NOT SWI Case */
   (
    ArmI := 16#0A000000;
    M := 16#0F000000;
    L := true;
    D := 10#4;
    FM(M,L,D);            /* Cond. Code  Field */
    M := 16#000000FF;
    L := false;
    D := 10#1;
    FM(M,L,D);            /* SOffset field  >>1 */
    InstExt := 0;
    if ((ThumbI and 16#00000001) = 16#00000001) then
      (InstExt := 1);
    (
     if ((ThumbI and 16#00000080) = 16#00000080) then /* Neg - sign ext */
      (
       M := 16#00FFFF00;    /* Neg sign ext */
       L := true;
       D := 10#0;
       FI(M,L,D)      
      )
    )
   )
  else                      /*  SWI case - F17*/
   (
    ArmI := 16#EF000000;
    M := 16#000000FF;
    L := true;
    D := 10#0;
    FM(M,L,D)             /* SOffset field  */
   )
 )
).         /* End of F16/F17 expansion */
/*=========================================================================*/
/*      F18 Expansion    */
(F18Exp: expr(void)=
 (
  ArmI := 16#EA000000;
  M := 16#000000FE;
  L := false;
  D := 10#1;
  FM(M,L,D);            /* SOffset field  >>1 */
  InstExt := 0;
  if ((ThumbI and 16#00000001) = 16#00000001) then
     (InstExt := 1);                   /* Half word alignment */
   (
    if ((ThumbI and 16#00000400) = 16#00000400) then /* Neg - sign ext */
    (
     M := 16#00FFFC00;    /* Neg sign ext */
     L := true;
     D := 10#0;
     FI(M,L,D)      
    )
   )
 )
).         /* End of F18 expansion */
/*=========================================================================*/
/*                       F19 Expansion outstanding                         */
(F19Exp: expr(void)=
 (
   printstr(" Format 19 Unimplemented ");
     M := 16#FFFFFFFF;    /* ILLEGAL ?? */
     L := true;
     D := 10#0;
     FI(M,L,D)      

 )
).         /* End of F19 expansion */
/*=========================================================================*/


 if ((ThumbI and 16#0000E000) = 16#00000000) then F1orF2Exp
 else
  (if ((ThumbI and 16#0000E000) = 16#00002000) then F3Exp
   else
    (if ((ThumbI and 16#0000FC00) = 16#00004000) then F4Exp    
     else
      (if ((ThumbI and 16#0000FC00) = 16#00004400) then F5Exp    
       else 
        (if ((ThumbI and 16#0000F600) = 16#00004800) then F6Exp
         else
          (if ((ThumbI and 16#0000F200) = 16#00005000) then F7Exp
           else
            (if ((ThumbI and 16#0000F200) = 16#00005200) then F8Exp
             else
              (if ((ThumbI and 16#0000E000) = 16#00006000) then F9Exp
               else
               (if ((ThumbI and 16#0000F000) = 16#00008000) then F10Exp
   else
    (if ((ThumbI and 16#0000F000) = 16#00009000) then F11Exp
     else
      (if ((ThumbI and 16#0000F000) = 16#0000A000) then F12Exp
       else
        (if ((ThumbI and 16#0000FF00) = 16#0000B000) then F13Exp
         else
          (if ((ThumbI and 16#0000F600) = 16#0000B400) then F14Exp
           else
            (if ((ThumbI and 16#0000F000) = 16#0000C000) then F15Exp
             else
              (if ((ThumbI and 16#0000F000) = 16#0000D000) then F1617Exp
               else
                (if ((ThumbI and 16#0000F800) = 16#0000E000) then F18Exp
                 else
                  (if ((ThumbI and 16#0000F000) = 16#0000F000) then F19Exp
                   else
                    (printstr("Failed Format Group Decode"))
                  )
                )
              )
            )
          )
        )
      )
    )
 
                )
              )
            )
          )
        )
      )
    )
  )



)
).

/* End of Thumb expansion definition  */
/*=========================================================================*/
/*=========================================================================*/
/*=========================================================================*/

  forever
    (
    In?(latch:=(?In));

    if (latch->PSR_ctrl->Thumb) then
      wait_for(Thumb_on_delay)
    else
      wait_for(Thumb_off_delay);

    if (latch->Dabt_svc) then Dabt_go?null;        /* Input from abort_go */


    Out!latch
    )
  )
).

/*------------------------------------------------------------------------*/

null

/*------------------------------------------------------------------------*/
