
/* instruction decoder */

decode(instr_bundle:var(chan(instr_bundle_t)),
       decoded_instr_bundle:var(chan(decoded_instr_bundle_t)),
       alu_result:var(chan(int)),
       mem_result:var(chan(int)):expr(void)=(
  
  do_decode(instr:val(int)):expr(decoded_instr_t)=(

    /* this is the function that actually does the decoding */

    bits(from_bit:val(int),to_bit:val(int)):expr(int)=(
      (instr and ((2<<to_bit)-1))>>from_bit
    ).

    bit(bitnum:val(int)):expr(bool)=(
      (instr and (1<<bitnum))!=0
    ).

    sext_bits(from_bit:val(int),to_bit:val(int)):expr(int)=(
      if bit(to_bit) then (
        bits(from_bit,to_bit-1)-(1<<(to_bit-from_bit))
      ) else (
        bits(from_bit,to_bit-1)
      )
    ).

    dec_instr:var(decoded_instr_t).

    dec_instr->ccode:=(case bits(8,11)
                         when 2#0000 => AL
                         when 2#0001 => NV
                         when 2#0010 => HI
                         when 2#0011 => LS
                         when 2#0100 => CC
                         when 2#0101 => CS
                         when 2#0110 => NE
                         when 2#0111 => EQ
                         when 2#1000 => VC
                         when 2#1001 => VS
                         when 2#1010 => PL
                         when 2#1011 => MI
                         when 2#1100 => GE
                         when 2#1101 => LT
                         when 2#1110 => GT
                         when 2#1111 => LE);
                       
    dec_instr->itype:=(case bits(13,15)
                         when 2#000 => ADD
                         when 2#001 => ADC
                         when 2#010 => SUB
                         when 2#011 => SBC
                         when 2#100 => AND
                         when 2#101 => OR
                         when 2#110 => if bit(11) then STORE else LOAD
                         when 2#111 => BRANCH);
                       
    dec_instr->shift:=(case bits(0,1)
                         when 2#00 => no_shift
                         when 2#01 => ASR
                         when 2#10 => ROR
                         when 2#11 => RRC);

    dec_instr->format:=(if bit(12) then onesource else twosource);
    dec_instr->srcA:=bits(5,7);
    dec_instr->srcB:=bits(2,4);
    dec_instr->immediate:=sext_bits(0,4);
    dec_instr->branch_offset:=sext_bits(0,7);
    dec_instr->dest:=bits(8,10);
    dec_instr->set_cc:=bit(11);
    dec_instr
  ).
  
  /* the decoder process */
  /* receive an instruction, along with any result from the ALU from
     the previous instruction, and any result from MEM for the instruction
     prior to that */
  
  instr_bundle_v:var(instr_bundle_t).
  decoded_instr_bundle_v:var(decoded_instr_bundle_t).
  
  get_reg(regnum:val(int)):expr(int)=(
    case regnum
      when 0 => 0
      when 7 => instr_bundle_v->pc+1
      default   regs[regnum]
  ).

  
  
  prev_instr->alu_result:=false;
  prev_instr->mem_result:=false;
  prev_prev_instr->alu_result:=false;
  prev_prev_instr->mem_result:=false;
  
  forever(
    (
      instr_bundle?(instr_bundle_v:=?instr_bundle) |
      if (prev_instr->alu_result) then (
        alu_result?(
          if (?alu_result->executed) then (
            regs[prev_instr->dest_reg]:=?alu_result->data
          )
        )
      ) |
      if (prev_prev_instr->mem_result) then (
        mem_result?(
          if (?mem_result->executed) then (
            regs[prev_prev_instr->dest_reg]:=?mem_result->data
          )
        )
      )
    );
    this_instr:=do_decode(instr_bundle_v->instr);
    dec_alu_bundle_v->colour:=instr_bundle_v->colour;
    dec_alu_bundle_v->itype:=this_instr->itype;
    dec_alu_bundle_v->shift:=this_instr->shift;
    dec_alu_bundle_v->ccode:=this_instr->ccode;
    dec_alu_bundle_v->srcA_data:=get_reg(this_instr->srcA);
    dec_alu_bundle_v->srcB_data:=(
      case this_instr->format
        when onesource => this_instr->immediate
        when twosource => get_reg(this_instr->srcB)
    );
    dec_alu_bundle_v->stdata:=get_reg(this_instr->dest);
    dec_alu_bundle!dec_alu_bundle_v;
    prev_prev_instr:=prev_instr;
    prev_instr:=this_instr
  )
).

null
