
/* stump memory model */
/* this version models a single-ported memory where instruction fetch
   and data accesses arbitrate for access to the memory */
/* 3 addresses have special behaviour when written to:
   FFFF (-1) : the value written is interpreted as a character and sent
               to standard output
   FFFE (-2) : the value written is interpreted as an integer and sent
               to standard output
   FFFD (-3) : the simulation is terminated.  */
  

/* variable to store the memory data */

mem:var(int^65536).
mem_available:var(bool).


/* instruction memory port */
/* read only */

imem(addr:var(chan(int)),
     data:var(chan(int))):expr(void)=(

  forever(
    addr?(
      wait_until(mem_available);
      mem_available:=false;
      wait_for(50);
      data!(mem[?addr]);
      mem_available:=true
    )
  )
).


/* data memory port */
/* read/write */

dmem(addrctrlstdata:var(chan(daddrctrlstdata_t)),
     lddata:var(chan(int))):expr(void)=(

  forever(
    addrctrlstdata?(
      wait_until(mem_available);
      mem_available:=false;
      wait_for(50);
      if (?addrctrlstdata->store) then (
        case (?addrctrlstdata->addr)
          when 16#FFFF => print(cast(char,(?addrctrlstdata->stdata)))
          when 16#FFFE => print(?addrctrlstdata->stdata)
          when 16#FFFD => ( print("terminated\n"); exit(0) )
          default         mem[?addrctrlstdata->addr]:=?addrctrlstdata->stdata
      ) else (
        lddata!(mem[?addrctrlstdata->addr])
      );
      mem_available:=true
    )
  )
).


/* memory initialisation from file */

meminit(fn:val(string)):expr(void)=(
  
  f:var(tfile).
  a:var(int).
  
  newstringvalparam(fn);
  
  f:=topenin(fn);
  set_default_input_base(2);
  a:=0;
  while (!eof(f)) do (
    if (a>=65536) then (
      fatal("Memory initialisation file "@fn
            @" contains more than 2^16 entries")
    );
    mem[a]:=freadint(f);
    a:=a+1
  );
  close(f);
  
  endstringvalparam(fn)
).


mem_available:=true;

meminit("progs/bubsort.bin")
